; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"global bits for ARCADEmini firmware"



;===========================================================================================
; global bits
;===========================================================================================

; these bits are locates in the bytes BITSx

; for a bit there are several defines:
; - the bit: BIT_<bitname>
; - the byte the bit is located in: BIT_<bitname>_BYTE
; - the mask to access the bit in its byte: BIT_<bitname>_MASK

; indicator which output buffer is owned by interrupt
; 0: O_FRAME0 is owned by interrupt
; 1: O_FRAME1 is owned by interrupt
; - bit only altered by interrupt
#define		BIT_O_INT_BUF		BITS0,0
#define		BIT_O_INT_BUF_BYTE	BITS0
#define		BIT_O_INT_BUF_MASK	0x01

; indicator that buffer switch is requested
; 1: buffer switch requested
; - bit set by non-interrupt software
; - bit reset by interrupt when buffer change is done
#define		BIT_O_BUF_SW		BITS0,1
#define		BIT_O_BUF_SW_BYTE	BITS0
#define		BIT_O_BUF_SW_MASK	0x02

; indicator bit that compact flash is busy
; 1: compact flash machinery is busy
#define		BIT_CF_BUSY		BITS0,2
#define		BIT_CF_BUSY_BYTE	BITS0
#define		BIT_CF_BUSY_MASK	0x04

; indicator bit that a compact flash error occured
; 1: compact flash error
#define		BIT_CF_ERROR		BITS0,3
#define		BIT_CF_ERROR_BYTE	BITS0
#define		BIT_CF_ERROR_MASK	0x08

; indicator bit if commands are executed from compact flash or internal flash 
; 0: executing commands from internal flash
; 1: executing commands from compact flash
; set by MAIN to signal execution of commands from compact flash
; cleared by COMMANDS to signal compact flash error
#define		BIT_CMD_CF		BITS0,4
#define		BIT_CMD_CF_BYTE		BITS0
#define		BIT_CMD_CF_MASK		0x10

; indicator bit that delay is active
; 0: delay is not active
; 1: delay is active
#define		BIT_DELAY		BITS0,5
#define		BIT_DELAY_BYTE		BITS0
#define		BIT_DELAY_MASK		0x20

; indicator bit if to wait for delay
; 0: do not wait for delay
; 1: wait for delay
#define		BIT_DELAY_WAIT		BITS0,6
#define		BIT_DELAY_WAIT_BYTE	BITS0
#define		BIT_DELAY_WAIT_MASK	0x40

; indicator that some taks should be executed
; 1: it's time to do some tasks
; - set by 1 ms interrupt to signal it's time for execution of
;   some tasks that need periodic processing but cannot be
;   done in interrupt
#define		BIT_TASKS_EXEC		BITS0,7
#define		BIT_TASKS_EXEC_BYTE	BITS0
#define		BIT_TASKS_EXEC_MASK	0x80

; indicator that usb power is allowed
; 1: usb power may be used
; - written by usb
; - read by output
#define		BIT_USB_PWR_OK		BITS1,0
#define		BIT_USB_PWR_OK_BYTE	BITS1
#define		BIT_USB_PWR_OK_MASK	0x01

; indicator that usb stopped command execution
; 1: usb has stopped command execution
; - written by usb
; - read by commands
#define		BIT_USB_CMD_STOP	BITS1,1
#define		BIT_USB_CMD_STOP_BYTE	BITS1
#define		BIT_USB_CMD_STOP_MASK	0x02

; indicator that command execution shall be stopped after next command
; 1: stop command execution after next command
; - set by usb
; - read and reset by commands
#define		BIT_USB_CMD_ST_N	BITS1,2
#define		BIT_USB_CMD_ST_N_BYTE	BITS1
#define		BIT_USB_CMD_ST_N_MASK	0x04

; indicator that command execution stop after next command has been done
; 1: command excution stop after next command has been done
; - set by commands
; - read and reset by usb
#define		BIT_USB_CMD_ST_D	BITS1,3
#define		BIT_USB_CMD_ST_D_BYTE	BITS1
#define		BIT_USB_CMD_ST_D_MASK	0x08
