; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"ARCADEmini firmware boot code"

	include		"P18F8520.INC"

	include		"IoPins.inc"



;===========================================================================================
; imported procedures
;===========================================================================================
	extern		ENTRY, HIGH_INTERRUPT, LOW_INTERRUPT
	extern		B_CF_IDENTIFY, B_CF_READ_SECTOR
	extern		B_TEA2_DEC



;===========================================================================================
; imported variables
;===========================================================================================
	extern		A1_REG, B1_REG, C1_REG, A2_REG, A4_REG
	extern		BUFFER, BUFFER_OV
	extern		CMD_SECTOR



;===========================================================================================
b_entry		CODE			; boot code entry point
;===========================================================================================
B_ENTRY:
	nop				; used by ICD
	nop
	goto	B_INIT

;===========================================================================================
b_int_hi	CODE			; high priority interrupts (boot code)
;===========================================================================================
B_HIGH_INTERRUPT:
	goto	HIGH_INTERRUPT		; jump to real interrupt code

;===========================================================================================
b_int_lo	CODE			; low priority interrupts (boot code)
;===========================================================================================
B_LOW_INTERRUPT:
	goto	LOW_INTERRUPT		; jump to real interrupt code



;===========================================================================================
b_prog		CODE			; boot code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; initialization
;-------------------------------------------------------------------------------------------
B_INIT:
; set default values for output pins
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_SH_nWE		; clear write enable
	bsf	PIN_CF_nCE		; clear CF chip enable
	bcf	PIN_CF_RESET		; clear CF reset
	bsf	PIN_USB_nCS		; clear USB chip select
	bsf	PIN_USB_nRESET		; clear USB chip reset
	clrf	PORT_COL0		; turn off all LEDs
	clrf	PORT_COL1
	clrf	PORT_COL2
	clrf	PORT_ROW2_COL3
	clrf	PORT_ROW1
	clrf	PORT_ROW0
; disable all analog inputs
	movlw	b'00001111'
	movwf	ADCON1
; disable comparator inputs
	movlw	b'00000111'
	movwf	CMCON
; define inputs/outputs
	clrf	TRISA
	clrf	TRISB
	clrf	TRISC
	clrf	TRISD
	clrf	TRISE
	clrf	TRISF
	movlw	b'00011111'
	movwf	TRISG
	setf	TRISH
	clrf	TRISJ
; disable timer 0
	bcf	T0CON,7
; disable interrupts
	bcf	RCON,7			; disable interrupt priority levels
	bcf	INTCON,7		; disable interrupts
	bcf	INTCON,6
; jump to boot code
	bra	B_MAIN

;-------------------------------------------------------------------------------------------
; compact flash boot id string (32 bytes)
; - firmware update
;-------------------------------------------------------------------------------------------
B_COMPACT_FLASH_ID:
	db	"ARCADEmini 26x20/16 fw-up 1.0   "

;-------------------------------------------------------------------------------------------
; firmware update key (16 bytes)
; (B5BAEEF99D2CFE3DF561B07780BDA661)
;-------------------------------------------------------------------------------------------
B_FW_UP_KEY:
	db	0xB5, 0xBA, 0xEE, 0xF9, 0x9D, 0x2C, 0xFE, 0x3D
	db	0xF5, 0x61, 0xB0, 0x77, 0x80, 0xBD, 0xA6, 0x61

;-------------------------------------------------------------------------------------------
; main boot code
;-------------------------------------------------------------------------------------------
B_MAIN:
; wait a few milliseconds to give compact flash time to boot
	setf	A2_REG
	setf	A2_REG+1
B_MAIN_WAIT:
	clrwdt
	decfsz	A2_REG
	bra	B_MAIN_WAIT
	decfsz	A2_REG+1
	bra	B_MAIN_WAIT

; check if a compact flash card is inserted
; and reset it if one is found

; check if compact flash is detected
	btfsc	PIN_CF_nCD1		; jump to end if no compact flash is detected
	bra	B_MAIN_END
; reset compact flash card
	bsf	PIN_CF_RESET		; set compact flash reset bit
	setf	A2_REG			; wait about 5 milliseconds
	movlw	0x50
	movwf	A2_REG+1
B_MAIN_RST_1:
	clrwdt
	decfsz	A2_REG
	bra	B_MAIN_RST_1
	decfsz	A2_REG+1
	bra	B_MAIN_RST_1
	bcf	PIN_CF_RESET		; clear compact flash reset bit
	movlw	0x10			; wait for compact flash busy bit
	movwf	B1_REG
	setf	C1_REG
B_MAIN_RST_WAIT:
	setf	A2_REG			; wait about a millisecond
	movlw	0x10
	movwf	A2_REG+1
B_MAIN_RST_0:
	clrwdt
	decfsz	A2_REG
	bra	B_MAIN_RST_0
	decfsz	A2_REG+1
	bra	B_MAIN_RST_0
	btfsc	PIN_CF_RDY		; if compact flash ready, go on
	bra	B_MAIN_RST_END
	decfsz	C1_REG
	bra	B_MAIN_RST_WAIT
	decfsz	B1_REG			; wait loop
	bra	B_MAIN_RST_WAIT
	bra	B_MAIN_END		; timeout
B_MAIN_RST_END:

; check if compact flash is working
; read first sector and check for firmware update identification

; identify compact flash
	call	B_CF_IDENTIFY
	btfsc	STATUS,C		; exit in case of error
	bra	B_MAIN_END
; read first sector of compact flash
	clrf	A4_REG			; sector 0
	clrf	A4_REG+1
	clrf	A4_REG+2
	clrf	A4_REG+3
	lfsr	FSR0,BUFFER		; address of buffer (512 bytes)
	call	B_CF_READ_SECTOR
	btfsc	STATUS,C		; exit in case of error
	bra	B_MAIN_END
; check id string (first 32 bytes)
	movlw	32			; 32 bytes
	movwf	A1_REG
	lfsr	FSR0,BUFFER		; address of buffer with CF data
	movlw	0x80			; configure for code read
	movwf	EECON1
	movlw	upper B_COMPACT_FLASH_ID	; address of id string in code
	movwf	TBLPTRU
	movlw	high B_COMPACT_FLASH_ID
	movwf	TBLPTRH
	movlw	low B_COMPACT_FLASH_ID
	movwf	TBLPTRL	
B_MAIN_CF_ID_CHECK:
	movf	POSTINC0,W		; read byte of CF data
	tblrd*+				; read byte of id string in code
	cpfseq	TABLAT			; compare and skip if match
	bra	B_MAIN_END		; mismatch, so exit
	decfsz	A1_REG
	bra	B_MAIN_CF_ID_CHECK

; erase the program memory from 0x0800 to 0x7BFF
; - this is done first to ensure that nobody can read the code by overwriting only a part

; set TBLPTR to 0x0800
	clrf	TBLPTRU
	movlw	0x08
	movwf	TBLPTRH
	clrf	TBLPTRL
B_MAIN_ERASE_LOOP:
	clrwdt
; erase the next 0x40 bytes
	movlw	0x94			; configure for code erase
	movwf	EECON1
	movlw	0x55			; code erase sequence
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR
	nop
	movlw	0x80			; configure for code read (better be on the safe side)
; advance TBLPTR (TBLPTRU stays 0 all the time)
	movlw	0x40
	addwf	TBLPTRL
	movlw	0
	addwfc	TBLPTRH
; finished when past address 0x7BFF
	movlw	0x7B
	cpfsgt	TBLPTRH
	bra	B_MAIN_ERASE_LOOP

; initialization for the TEA 2.0 algorithm

; load firmware update key
	movlw	16			; 16 bytes
	movwf	A1_REG
	lfsr	FSR0,BUFFER_OV		; write key into BUFFER_OV+0..15
	movlw	0x80			; configure for code read
	movwf	EECON1
	movlw	upper B_FW_UP_KEY	; address of id string in code
	movwf	TBLPTRU
	movlw	high B_FW_UP_KEY
	movwf	TBLPTRH
	movlw	low B_FW_UP_KEY
	movwf	TBLPTRL	
B_MAIN_FW_UP_KEY:
	tblrd*+				; read byte of key
	movff	TABLAT,POSTINC0		; write byte into buffer
	decfsz	A1_REG
	bra	B_MAIN_FW_UP_KEY
; clear previous encrypted block
	lfsr	FSR0,BUFFER_OV+16
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0

; read and decrypt the first four sectors
; - addresses 0x0000..0x07FF (address range of boot code)
; - sectors 1..4 (sector 0 is identification)
; - boot code is not written to program memory
; - decryption is done to get previous encrypted block right
;   (we must start decryptiong at the beginning)
; - sectors are only counted in low byte of CMD_SECTOR
;   - we have far less than 256 sectors)

; - these four sectors are not really needed, but we do it that way for these reasons
;   - we can just copy an encrypted binary image of the entire program meory
;     behind the identification sector and write it to the compact flash
;   - if we use random data for the first four sectors (which are not used)
;     we have completely different data in every firmware update
;     even if the code (or the beginning of the code) is the same

; start in sector 1 (sector 0 is identification)
	movlw	1
	movwf	CMD_SECTOR
B_MAIN_PRE_WRITE_LOOP:
	clrwdt
; read next sector
	movff	CMD_SECTOR,A4_REG
	clrf	A4_REG+1
	clrf	A4_REG+2
	clrf	A4_REG+3
	lfsr	FSR0,BUFFER
	call	B_CF_READ_SECTOR
	btfsc	STATUS,C		; exit in case of error
	bra	B_MAIN_END
; decrypt sector
	lfsr	FSR1,BUFFER
	movlw	64			; 64 blocks = 512 bytes
	movwf	C1_REG
	call	B_TEA2_DEC
; next sector
	incf	CMD_SECTOR
	movlw	4			; sector 4 is address 0x0600 (sector 0 is identification)
	cpfsgt	CMD_SECTOR
	bra	B_MAIN_PRE_WRITE_LOOP

; start writing new code into program memory at address 0x0800
; - we are now at sector 5 which is address 0x0800 (sector 0 is identification)
; - sectors are only counted in low byte of CMD_SECTOR
;   - we have far less than 256 sectors)

; set TBLPTR to 0x0800
	clrf	TBLPTRU
	movlw	0x08
	movwf	TBLPTRH
	clrf	TBLPTRL

; write new code into program memory
; - addresses 0x0800..0x7BFF (address range of real firmware)
; - sectors 5..62 (sector 0 is identification)

B_MAIN_WRITE_LOOP:
	clrwdt
; read next sector
	movff	CMD_SECTOR,A4_REG
	clrf	A4_REG+1
	clrf	A4_REG+2
	clrf	A4_REG+3
	lfsr	FSR0,BUFFER
	call	B_CF_READ_SECTOR
	btfsc	STATUS,C		; exit in case of error
	bra	B_MAIN_END
; decrypt sector
	lfsr	FSR1,BUFFER
	movlw	64			; 64 blocks = 512 bytes
	movwf	C1_REG
	call	B_TEA2_DEC

; write sector into program memory
	lfsr	FSR0,BUFFER		; fetch data from buffer
	movlw	64			; 64 times 8 bytes = 512 bytes
	movwf	A1_REG
B_MAIN_WRITE_SECTOR_LOOP:
	movff	POSTINC0,TABLAT		; 1st byte into holding register
	tblwt*+
	movff	POSTINC0,TABLAT		; 2nd byte into holding register
	tblwt*+
	movff	POSTINC0,TABLAT		; 3rd byte into holding register
	tblwt*+
	movff	POSTINC0,TABLAT		; 4th byte into holding register
	tblwt*+
	movff	POSTINC0,TABLAT		; 5th byte into holding register
	tblwt*+
	movff	POSTINC0,TABLAT		; 6th byte into holding register
	tblwt*+
	movff	POSTINC0,TABLAT		; 7th byte into holding register
	tblwt*+
	movff	POSTINC0,TABLAT		; 8th byte into holding register
	tblwt*				; (no + here to stay within desired write address range)
	movlw	0x84			; configure for code write
	movwf	EECON1
	movlw	0x55			; code write sequence
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR
	nop
	tblrd*+				; dummy read to advance to next address (+ was left out some lines ago)
	movlw	0x80			; configure for code read (better be on the safe side)
; next 8 bytes 
	decfsz	A1_REG
	bra	B_MAIN_WRITE_SECTOR_LOOP

; next sector
	incf	CMD_SECTOR
	movlw	62			; sector 62 is address 0x7A00 (sector 0 is identification)
	cpfsgt	CMD_SECTOR
	bra	B_MAIN_WRITE_LOOP

; now we are done and can jump to our new firmware

; the bootloader finished
; - no compact flash is inserted
; - inserted compact flash does not have an firmware update identification
; - firmware update is completed
; now we can start the real code

B_MAIN_END:
; start normal code
	goto	ENTRY



	end
