; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"eeprom procedures for ARCADEmini firmware"

	include		"P18F8520.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; EEPROM layout
;===========================================================================================
	#define		EE_ADDR_ID		0x00	; ID in EEPROM (needs 32 bytes)
	#define		EE_ADDR_OUT_STYLE	0x20	; output style flag (needs 2 bytes)
	#define		EE_ADDR_OUT_TIMING	0x22	; output timing (needs 30 bytes)
	#define		EE_ADDR_OUT_GREYMAP	0x40	; output greyscale mapping (needs 30 bytes)
	#define		EE_ADDR_END		0x5E	; unused from here



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0, BITS1
	extern		A1_REG, B1_REG, C1_REG, D1_REG
	extern		A2_REG, B2_REG, C2_REG, D2_REG
	extern		A4_REG
	extern		O_STYLE, O_TIMING, O_GREY_MAP



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

; private bit variables
EE_BITS		res	1
#define		EE_BUSY			EE_BITS,0	; writing something to EEPROM
#define		EE_WR_ID_ACT		EE_BITS,1	; writing ID from RAM to EEPROM
#define		EE_WR_OUT_STYLE_ACT	EE_BITS,2	; writing output style from RAM to EEPROM
#define		EE_WR_OUT_TIMING_ACT	EE_BITS,3	; writing output timing from RAM to EEPROM
#define		EE_WR_OUT_GREYMAP_ACT	EE_BITS,4	; writing output greyscale mapping from RAM to EEPROM

; next byte to write to EEPROM
EE_WR_ADDR	res	1



;===========================================================================================
ram_ee		UDATA			; normal RAM for EEPROM
;===========================================================================================

; ID to write to EEPROM
EE_BUFFER	res	16



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; EEPROM initialization
;-------------------------------------------------------------------------------------------
EE_INIT:
; not writing anything to EEPROM
	bcf	EE_BUSY
	bcf	EE_WR_ID_ACT
	bcf	EE_WR_OUT_STYLE_ACT
	bcf	EE_WR_OUT_TIMING_ACT
	bcf	EE_WR_OUT_GREYMAP_ACT
; done
	return

;-------------------------------------------------------------------------------------------
; EEPROM configuration
; - called after intialization
; - reads settings from EEPROM and writes it into the appropriate positions
; changes:	A1_REG, FSR0, EE*
;-------------------------------------------------------------------------------------------
EE_CONFIG:
; get output style
	clrf	EECON1			; configurate EEPROM for reading
	movlw	EE_ADDR_OUT_STYLE	; address of output style in EEPROM
	movwf	EEADR
	clrf	EEADRH
	bsf	EECON1,RD		; read EEPROM
	movff	EEDATA,A1_REG
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	bsf	EECON1,RD		; read EEPROM
	comf	EEDATA,W		; if value is not complement of other value, ...
	cpfseq	A1_REG
	clrf	A1_REG			; ..., use default style 0x00
; set output style flag of output procedure (atomically)
	movff	A1_REG,O_STYLE
; get output timing
	lfsr	FSR0,O_TIMING		; address of buffer for timing data
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	EECON1			; configurate EEPROM for reading
	movlw	EE_ADDR_OUT_TIMING	; start address of output timing in EEPROM
	movwf	EEADR
	clrf	EEADRH
EE_CONFIG_OUT_TIMING_LOOP:
	bsf	EECON1,RD		; read EEPROM
	movff	EEDATA,INDF0
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	bsf	EECON1,RD		; read EEPROM
	comf	EEDATA,W		; if value is not complement of other value, ...
	cpfseq	INDF0
	bra	EE_CONFIG_OUT_TIMING_ERR ; ..., error
	movlw	50+1			; if value > 50, ...
	cpfslt	INDF0
	bra	EE_CONFIG_OUT_TIMING_ERR ; ..., error
	movlw	49			; value range 0..50 -> 49..99
	addwf	POSTINC0
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	decfsz	A1_REG			; loop
	bra	EE_CONFIG_OUT_TIMING_LOOP
	bra	EE_CONFIG_OUT_TIMING_END
; set output timing to default
EE_CONFIG_OUT_TIMING_ERR:
	lfsr	FSR0,O_TIMING		; address of buffer for timing data
	movlw	15			; loop 15 times
	movwf	A1_REG
EE_CONFIG_OUT_TIMING_ERR_LOOP:
	movlw	74			; default value
	movwf	POSTINC0
	decfsz	A1_REG			; loop
	bra	EE_CONFIG_OUT_TIMING_ERR_LOOP
EE_CONFIG_OUT_TIMING_END:
; get output greyscale mapping
	lfsr	FSR0,O_GREY_MAP		; address of buffer for greyscale mapping data
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	EECON1			; configurate EEPROM for reading
	movlw	EE_ADDR_OUT_GREYMAP	; start address of output greyscale mapping in EEPROM
	movwf	EEADR
	clrf	EEADRH
EE_CONFIG_OUT_GREYMAP_LOOP:
	bsf	EECON1,RD		; read EEPROM
	movff	EEDATA,INDF0
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	bsf	EECON1,RD		; read EEPROM
	comf	EEDATA,W		; if value is not complement of other value, ...
	cpfseq	INDF0
	bra	EE_CONFIG_OUT_GREYMAP_ERR ; ..., error
	movf	INDF0			; if value == 0, ...
	btfsc	STATUS,Z
	bra	EE_CONFIG_OUT_GREYMAP_ERR ; ..., error
	movlw	15+1			; if value > 15, ...
	cpfslt	INDF0
	bra	EE_CONFIG_OUT_GREYMAP_ERR ; ..., error
	swapf	POSTINC0		; convert value (swap it)
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	decfsz	A1_REG			; loop
	bra	EE_CONFIG_OUT_GREYMAP_LOOP
	bra	EE_CONFIG_OUT_GREYMAP_END
; set output greyscale mapping to default
EE_CONFIG_OUT_GREYMAP_ERR:
	lfsr	FSR0,O_GREY_MAP		; address of buffer for greyscale mapping data
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	B1_REG
EE_CONFIG_OUT_GREYMAP_ERR_LOOP:
	incf	B1_REG
	swapf	B1_REG,W		; default value
	movwf	POSTINC0
	decfsz	A1_REG			; loop
	bra	EE_CONFIG_OUT_GREYMAP_ERR_LOOP
EE_CONFIG_OUT_GREYMAP_END:
; done
	return

;-------------------------------------------------------------------------------------------
; EEPROM task
; - called periodically
; changes:	A1_REG, FSR0, EE*
;-------------------------------------------------------------------------------------------
EE_TASK:
; nothing needs to be done
	btfss	EE_BUSY
	return
; EEPROM still busy
	btfsc	EECON1,WR
	return
; writing ID from RAM to EEPROM
	btfss	EE_WR_ID_ACT
	bra	EE_TASK_NOT_ID
; get next byte from RAM
	lfsr	FSR0,EE_BUFFER		; EEDATA = *(&EE_BUFFER+(EE_WR_ADDR>>1))
	bcf	STATUS,C
	rrcf	EE_WR_ADDR,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	INDF0,EEDATA
	btfsc	EE_WR_ADDR,0		; complement EEDATA if (EE_WR_ADDR&1)!=0
	comf	EEDATA
; write byte to EEPROM
	movlw	EE_ADDR_ID		; get address of byte in EEPROM
	addwf	EE_WR_ADDR,W
	movwf	EEADR
	clrf	EEADRH
	movlw	0x04			; configurate EEPROM for writing
	movwf	EECON1
	bcf	INTCON,GIEL		; disable interrupts
	bcf	INTCON,GIEH
	movlw	0x55			; write sequence
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR
	bsf	INTCON,GIEH		; enable interrupts
	bsf	INTCON,GIEL
; next byte
	incf	EE_WR_ADDR
	movlw	0x1F			; if not finished, ...
	cpfsgt	EE_WR_ADDR
	return				; ..., simply exit
; clear write flag for ID and busy flag
	bcf	EE_WR_ID_ACT
	bcf	EE_BUSY
; done
	return
EE_TASK_NOT_ID:
; writing output style from RAM to EEPROM
	btfss	EE_WR_OUT_STYLE_ACT
	bra	EE_TASK_NOT_OUT_STYLE
; get flag from RAM
	movff	EE_BUFFER,EEDATA
	btfsc	EE_WR_ADDR,0		; complement EEDATA if (EE_WR_ADDR&1)!=0
	comf	EEDATA
; write byte to EEPROM
	movlw	EE_ADDR_OUT_STYLE	; get address of byte in EEPROM
	addwf	EE_WR_ADDR,W
	movwf	EEADR
	clrf	EEADRH
	movlw	0x04			; configurate EEPROM for writing
	movwf	EECON1
	bcf	INTCON,GIEL		; disable interrupts
	bcf	INTCON,GIEH
	movlw	0x55			; write sequence
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR
	bsf	INTCON,GIEH		; enable interrupts
	bsf	INTCON,GIEL
; next byte
	incf	EE_WR_ADDR
	movlw	0x01			; if not finished, ...
	cpfsgt	EE_WR_ADDR
	return				; ..., simply exit
; clear write flag for output style and busy flag
	bcf	EE_WR_OUT_STYLE_ACT
	bcf	EE_BUSY
; done
	return
EE_TASK_NOT_OUT_STYLE:
; writing output timing from RAM to EEPROM
	btfss	EE_WR_OUT_TIMING_ACT
	bra	EE_TASK_NOT_OUT_TIMING
; get next byte from RAM
	lfsr	FSR0,EE_BUFFER		; EEDATA = *(&EE_BUFFER+(EE_WR_ADDR>>1))
	bcf	STATUS,C
	rrcf	EE_WR_ADDR,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	INDF0,EEDATA
	btfsc	EE_WR_ADDR,0		; complement EEDATA if (EE_WR_ADDR&1)!=0
	comf	EEDATA
; write byte to EEPROM
	movlw	EE_ADDR_OUT_TIMING	; get address of byte in EEPROM
	addwf	EE_WR_ADDR,W
	movwf	EEADR
	clrf	EEADRH
	movlw	0x04			; configurate EEPROM for writing
	movwf	EECON1
	bcf	INTCON,GIEL		; disable interrupts
	bcf	INTCON,GIEH
	movlw	0x55			; write sequence
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR
	bsf	INTCON,GIEH		; enable interrupts
	bsf	INTCON,GIEL
; next byte
	incf	EE_WR_ADDR
	movlw	0x1D			; if not finished, ...
	cpfsgt	EE_WR_ADDR
	return				; ..., simply exit
; clear write flag for output timing and busy flag
	bcf	EE_WR_OUT_TIMING_ACT
	bcf	EE_BUSY
; done
	return
EE_TASK_NOT_OUT_TIMING:
; writing output greyscale mapping from RAM to EEPROM
	btfss	EE_WR_OUT_GREYMAP_ACT
	bra	EE_TASK_NOT_OUT_GREYMAP
; get next byte from RAM
	lfsr	FSR0,EE_BUFFER		; EEDATA = *(&EE_BUFFER+(EE_WR_ADDR>>1))
	bcf	STATUS,C
	rrcf	EE_WR_ADDR,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	INDF0,EEDATA
	btfsc	EE_WR_ADDR,0		; complement EEDATA if (EE_WR_ADDR&1)!=0
	comf	EEDATA
; write byte to EEPROM
	movlw	EE_ADDR_OUT_GREYMAP	; get address of byte in EEPROM
	addwf	EE_WR_ADDR,W
	movwf	EEADR
	clrf	EEADRH
	movlw	0x04			; configurate EEPROM for writing
	movwf	EECON1
	bcf	INTCON,GIEL		; disable interrupts
	bcf	INTCON,GIEH
	movlw	0x55			; write sequence
	movwf	EECON2
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR
	bsf	INTCON,GIEH		; enable interrupts
	bsf	INTCON,GIEL
; next byte
	incf	EE_WR_ADDR
	movlw	0x1D			; if not finished, ...
	cpfsgt	EE_WR_ADDR
	return				; ..., simply exit
; clear write flag for output greyscale mapping and busy flag
	bcf	EE_WR_OUT_GREYMAP_ACT
	bcf	EE_BUSY
; done
	return
EE_TASK_NOT_OUT_GREYMAP:
; clear busy flag
	bcf	EE_BUSY
; done
	return

;-------------------------------------------------------------------------------------------
; get ID from EEPROM
; input:	FSR0		pointer to 16 byte buffer to write ID to
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, FSR0, A1_REG, B1_REG, EE*
;-------------------------------------------------------------------------------------------
EE_GET_ID:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_GET_ID_GO
; clear buffer
	movlw	16			; loop 16 times
	movwf	A1_REG
EE_GET_ID_BUSY_CLR_LOOP:
	clrf	POSTINC0		; clear next byte
	decfsz	A1_REG			; loop
	bra	EE_GET_ID_BUSY_CLR_LOOP
; return error
	retlw	1
; read ID from EEPROM and check it
EE_GET_ID_GO:
	movlw	16			; loop 16 times
	movwf	A1_REG
	clrf	EECON1			; configurate EEPROM for reading
	movlw	EE_ADDR_ID		; start address of ID in EEPROM
	movwf	EEADR
	clrf	EEADRH
	bcf	B1_REG,0		; no error yet
EE_GET_ID_LOOP:
	bsf	EECON1,RD		; read EEPROM
	movff	EEDATA,INDF0
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	bsf	EECON1,RD		; read EEPROM
	comf	EEDATA,W		; if value is not complement of other value, ...
	cpfseq	POSTINC0
	bsf	B1_REG,0		; ..., remember error
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	decfsz	A1_REG			; loop
	bra	EE_GET_ID_LOOP
; return success if no error
	btfss	B1_REG,0
	retlw	0
; return to start of buffer
	movlw	16
	subwf	FSR0L
	movlw	0
	subwfb	FSR0H
; clear ID
	movlw	16			; loop 16 times
	movwf	A1_REG
EE_GET_ID_CLR_LOOP:
	clrf	POSTINC0		; clear next byte
	decfsz	A1_REG			; loop
	bra	EE_GET_ID_CLR_LOOP
; return success (wrong data in EEPROM is not a real error)
	retlw	0

;-------------------------------------------------------------------------------------------
; set ID in EEPROM
; input:	FSR0		pointer to 16 byte buffer with ID
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, FSR0, A1_REG, A2_REG, EE*
; schedules the ID to be written into EEPROM
;-------------------------------------------------------------------------------------------
EE_SET_ID:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_SET_ID_GO
; skip buffer content
	movlw	16
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
; return error
	retlw	1
; save ID to RAM
EE_SET_ID_GO:
	lfsr	FSR1,EE_BUFFER		; address of buffer
	movlw	16			; loop 16 times
	movwf	A1_REG
EE_SET_ID_LOOP:
	movff	POSTINC0,POSTINC1	; copy next byte
	decfsz	A1_REG			; loop
	bra	EE_SET_ID_LOOP
; schedule writing ID to EEPROM
	bsf	EE_BUSY
	bsf	EE_WR_ID_ACT
	clrf	EE_WR_ADDR
; return success
	retlw	0

;-------------------------------------------------------------------------------------------
; get output style flag from EEPROM
; input:	-
; output:	WREG		error code (0 for success)
;		A1_REG		output style
; changes:	WREG, STATUS, A1_REG, EE*
;-------------------------------------------------------------------------------------------
EE_GET_OUT_STYLE:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_GET_OUT_STYLE_GO
; clear flag
	clrf	A1_REG
; return error
	retlw	1
; read output flag from EEPROM and check it
EE_GET_OUT_STYLE_GO:
	clrf	EECON1			; configurate EEPROM for reading
	movlw	EE_ADDR_OUT_STYLE	; address of output style in EEPROM
	movwf	EEADR
	clrf	EEADRH
	bsf	EECON1,RD		; read EEPROM
	movff	EEDATA,A1_REG
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	bsf	EECON1,RD		; read EEPROM
	comf	EEDATA,W		; if value is not complement of other value, ...
	cpfseq	A1_REG
	bra	EE_GET_OUT_FLAG_INVAL_DATA
; return success
	retlw	0
; clear flag
EE_GET_OUT_FLAG_INVAL_DATA:
	clrf	A1_REG
; return success (wrong data in EEPROM is not a real error)
	retlw	0

;-------------------------------------------------------------------------------------------
; set output style flag in EEPROM
; input:	A1_REG		new output style
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, A1_REG, EE*
; schedules the output style flag to be written into EEPROM
; set output style flag of output procedure
;-------------------------------------------------------------------------------------------
EE_SET_OUT_STYLE:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_SET_OUT_FLAG_GO
; return error
	retlw	1
; save output style to RAM
EE_SET_OUT_FLAG_GO:
	movff	A1_REG,EE_BUFFER	; get flag to EEPROM buffer
	movlw	0x01+1
	cpfslt	EE_BUFFER
	clrf	EE_BUFFER
; schedule writing output style to EEPROM
	bsf	EE_BUSY
	bsf	EE_WR_OUT_STYLE_ACT
	clrf	EE_WR_ADDR
; set output style flag of output procedure (atomically)
	movff	A1_REG,O_STYLE
; return success
	retlw	0

;-------------------------------------------------------------------------------------------
; get output timing from EEPROM
; input:	FSR0		pointer to 15 byte buffer to write output timing to
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, FSR0, A1_REG, B1_REG, EE*
;-------------------------------------------------------------------------------------------
EE_GET_OUT_TIMING:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_GET_OUT_TIMING_GO
; set buffer to standard values
	movlw	15			; loop 15 times
	movwf	A1_REG
EE_GET_OUT_TIMING_BUSY_CLR_LOOP:
	movlw	25			; set next byte to standard value
	movwf	POSTINC0
	decfsz	A1_REG			; loop
	bra	EE_GET_OUT_TIMING_BUSY_CLR_LOOP
; return error
	retlw	1
; read output timing from EEPROM and check it
EE_GET_OUT_TIMING_GO:
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	EECON1			; configurate EEPROM for reading
	movlw	EE_ADDR_OUT_TIMING	; start address of output timing in EEPROM
	movwf	EEADR
	clrf	EEADRH
	bcf	B1_REG,0		; no error yet
EE_GET_OUT_TIMING_LOOP:
	bsf	EECON1,RD		; read EEPROM
	movff	EEDATA,INDF0
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	bsf	EECON1,RD		; read EEPROM
	comf	EEDATA,W		; if value is not complement of other value, ...
	cpfseq	INDF0
	bsf	B1_REG,0		; ..., remember error
	movlw	50+1			; if value > 50, ...
	cpfslt	POSTINC0
	bsf	B1_REG,0		; ..., remember error
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	decfsz	A1_REG			; loop
	bra	EE_GET_OUT_TIMING_LOOP
; return success if no error
	btfss	B1_REG,0
	retlw	0
; return to start of buffer
	movlw	15
	subwf	FSR0L
	movlw	0
	subwfb	FSR0H
; set buffer to standard values
	movlw	15			; loop 15 times
	movwf	A1_REG
EE_GET_OUT_TIMING_CLR_LOOP:
	movlw	25			; set next byte to standard value
	movwf	POSTINC0
	decfsz	A1_REG			; loop
	bra	EE_GET_OUT_TIMING_CLR_LOOP
; return success (wrong data in EEPROM is not a real error)
	retlw	0

;-------------------------------------------------------------------------------------------
; set output timing in EEPROM
; input:	FSR0		pointer to 15 byte buffer with output timing
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, FSR0, A1_REG, B1_REG, A2_REG, B2_REG, EE*
; schedules the output timing to be written into EEPROM
; sets output timing information of output procedure
;-------------------------------------------------------------------------------------------
EE_SET_OUT_TIMING:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_SET_OUT_TIMING_GO
; skip buffer content
	movlw	15
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
; return error
	retlw	1
; save output timing to RAM
EE_SET_OUT_TIMING_GO:
	lfsr	FSR1,EE_BUFFER		; address of buffer
	movlw	15			; loop 15 times
	movwf	A1_REG
	movlw	50+1
	movwf	B1_REG
EE_SET_OUT_TIMING_LOOP:
	movf	POSTINC0,W		; check, correct and copy next byte
	cpfsgt	B1_REG
	movlw	25
	movwf	POSTINC1
	decfsz	A1_REG			; loop
	bra	EE_SET_OUT_TIMING_LOOP
; schedule writing output timing to EEPROM
	bsf	EE_BUSY
	bsf	EE_WR_OUT_TIMING_ACT
	clrf	EE_WR_ADDR
; set timing information (of output procedure)
	movff	FSR0L,A2_REG+0		; save FSR0
	movff	FSR0H,A2_REG+1
	lfsr	FSR0,EE_BUFFER		; address of buffers
	lfsr	FSR1,O_TIMING
	movlw	15			; loop 15 times
	movwf	A1_REG
EE_SET_OUT_TIMING_LOOP_2:
	movf	POSTINC0,W		; transfer next byte
	addlw	49			; add 49 (value range 0..50 -> 49..99)
	movwf	POSTINC1
	decfsz	A1_REG			; loop
	bra	EE_SET_OUT_TIMING_LOOP_2
	movff	A2_REG+0,FSR0L		; restore FSR0
	movff	A2_REG+1,FSR0H
; return success
	retlw	0

;-------------------------------------------------------------------------------------------
; reset output timing in EEPROM
; input:	-		
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, FSR0, A1_REG, B1_REG, A2_REG, B2_REG, EE*
; schedules the output timing to be written into EEPROM
; sets output timing information of output procedure
;-------------------------------------------------------------------------------------------
EE_RESET_OUT_TIMING:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_RESET_OUT_TIMING_GO
; return error
	retlw	1
; save output timing to RAM
EE_RESET_OUT_TIMING_GO:
	lfsr	FSR0,EE_BUFFER		; address of buffer
	movlw	15			; loop 15 times
	movwf	A1_REG
EE_RESET_OUT_TIMING_LOOP:
	movlw	25
	movwf	POSTINC0
	decfsz	A1_REG			; loop
	bra	EE_RESET_OUT_TIMING_LOOP
; schedule writing output timing to EEPROM
	bsf	EE_BUSY
	bsf	EE_WR_OUT_TIMING_ACT
	clrf	EE_WR_ADDR
; set timing information (of output procedure)
	lfsr	FSR0,O_TIMING		; address of buffer
	movlw	15			; loop 15 times
	movwf	A1_REG
EE_RESET_OUT_TIMING_LOOP_2:
	movlw	74
	movwf	POSTINC0
	decfsz	A1_REG			; loop
	bra	EE_RESET_OUT_TIMING_LOOP_2
; return success
	retlw	0

;-------------------------------------------------------------------------------------------
; get output greyscale mapping from EEPROM
; input:	FSR0		pointer to 15 byte buffer to write output greyscale mapping to
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, FSR0, A1_REG, B1_REG, EE*
;-------------------------------------------------------------------------------------------
EE_GET_OUT_GREYMAP:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_GET_OUT_GREYMAP_GO
; set buffer to standard values
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	B1_REG
EE_GET_OUT_GREYMAP_BUSY_CLR_LOOP:
	incf	B1_REG
	movff	B1_REG,POSTINC0		; set next byte to standard value
	decfsz	A1_REG			; loop
	bra	EE_GET_OUT_GREYMAP_BUSY_CLR_LOOP
; return error
	retlw	1
; read output greyscale mapping from EEPROM and check it
EE_GET_OUT_GREYMAP_GO:
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	EECON1			; configurate EEPROM for reading
	movlw	EE_ADDR_OUT_GREYMAP	; start address of output greyscale mapping in EEPROM
	movwf	EEADR
	clrf	EEADRH
	bcf	B1_REG,0		; no error yet
EE_GET_OUT_GREYMAP_LOOP:
	bsf	EECON1,RD		; read EEPROM
	movff	EEDATA,INDF0
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	bsf	EECON1,RD		; read EEPROM
	comf	EEDATA,W		; if value is not complement of other value, ...
	cpfseq	INDF0
	bsf	B1_REG,0		; ..., remember error
	movf	INDF0,W			; if value == 0, ...
	btfsc	STATUS,Z
	bsf	B1_REG,0		; ..., remember error
	movlw	15+1			; if value > 15, ...
	cpfslt	POSTINC0
	bsf	B1_REG,0		; ..., remember error
	incf	EEADR			; next address in EEPROM (no overflow of EEADR here)
	decfsz	A1_REG			; loop
	bra	EE_GET_OUT_GREYMAP_LOOP
; return success if no error
	btfss	B1_REG,0
	retlw	0
; return to start of buffer
	movlw	15
	subwf	FSR0L
	movlw	0
	subwfb	FSR0H
; set buffer to standard values
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	B1_REG
EE_GET_OUT_GREYMAP_CLR_LOOP:
	incf	B1_REG
	movff	B1_REG,POSTINC0		; set next byte to standard value
	decfsz	A1_REG			; loop
	bra	EE_GET_OUT_GREYMAP_CLR_LOOP
; return success (wrong data in EEPROM is not a real error)
	retlw	0

;-------------------------------------------------------------------------------------------
; set output greyscale mapping in EEPROM
; input:	FSR0		pointer to 15 byte buffer with output greyscale mapping
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, FSR0, A1_REG, B1_REG, C1_REG, A2_REG, B2_REG, EE*
; schedules the output greyscale mapping to be written into EEPROM
; sets output greyscale mapping of output procedure
;-------------------------------------------------------------------------------------------
EE_SET_OUT_GREYMAP:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_SET_OUT_GREYMAP_GO
; skip buffer content
	movlw	15
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
; return error
	retlw	1
; save output greyscale mapping to RAM
EE_SET_OUT_GREYMAP_GO:
	lfsr	FSR1,EE_BUFFER		; address of buffer
	movlw	15			; loop 15 times
	movwf	A1_REG
	movlw	15+1
	movwf	B1_REG
	movlw	1
	movwf	C1_REG
EE_SET_OUT_GREYMAP_LOOP:
	movf	POSTINC0,W		; check, correct and copy next byte
	cpfsgt	B1_REG
	movf	C1_REG,W
	movwf	POSTINC1
	incf	C1_REG
	decfsz	A1_REG			; loop
	bra	EE_SET_OUT_GREYMAP_LOOP
; schedule writing output greyscale mapping to EEPROM
	bsf	EE_BUSY
	bsf	EE_WR_OUT_GREYMAP_ACT
	clrf	EE_WR_ADDR
; set greyscale mapping information (of output procedure)
	movff	FSR0L,A2_REG+0		; save FSR0
	movff	FSR0H,A2_REG+1
	lfsr	FSR0,EE_BUFFER		; address of buffers
	lfsr	FSR1,O_GREY_MAP
	movlw	15			; loop 15 times
	movwf	A1_REG
EE_SET_OUT_GREYMAP_LOOP_2:
	swapf	POSTINC0,W		; transfer next byte (swap it to convert value)
	movwf	POSTINC1
	decfsz	A1_REG			; loop
	bra	EE_SET_OUT_GREYMAP_LOOP_2
	movff	A2_REG+0,FSR0L		; restore FSR0
	movff	A2_REG+1,FSR0H
; return success
	retlw	0

;-------------------------------------------------------------------------------------------
; reset output greyscale mapping in EEPROM
; input:	-		
; output:	WREG		error code (0 for success)
; changes:	WREG, STATUS, FSR0, A1_REG, B1_REG, A2_REG, B2_REG, EE*
; schedules the output greyscale mapping to be written into EEPROM
; sets output greyscale mapping information of output procedure
;-------------------------------------------------------------------------------------------
EE_RESET_OUT_GREYMAP:
; if EEPROM is busy
	btfss	EE_BUSY
	bra	EE_RESET_OUT_GREYMAP_GO
; return error
	retlw	1
; save output greyscale mapping to RAM
EE_RESET_OUT_GREYMAP_GO:
	lfsr	FSR0,EE_BUFFER		; address of buffer
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	B1_REG
EE_RESET_OUT_GREYMAP_LOOP:
	incf	B1_REG
	movff	B1_REG,POSTINC0
	decfsz	A1_REG			; loop
	bra	EE_RESET_OUT_GREYMAP_LOOP
; schedule writing output greyscale mapping to EEPROM
	bsf	EE_BUSY
	bsf	EE_WR_OUT_GREYMAP_ACT
	clrf	EE_WR_ADDR
; set greyscale mapping (of output procedure)
	lfsr	FSR0,O_GREY_MAP		; address of buffer
	movlw	15			; loop 15 times
	movwf	A1_REG
	clrf	B1_REG
EE_RESET_OUT_GREYMAP_LOOP_2:
	incf	B1_REG
	swapf	B1_REG,W		; default value
	movwf	POSTINC0
	decfsz	A1_REG			; loop
	bra	EE_RESET_OUT_GREYMAP_LOOP_2
; return success
	retlw	0



;===========================================================================================
; global variables
;===========================================================================================



;===========================================================================================
; global procedures
;===========================================================================================

	global		EE_INIT, EE_CONFIG, EE_TASK
	global		EE_GET_ID, EE_SET_ID
	global		EE_GET_OUT_STYLE, EE_SET_OUT_STYLE
	global		EE_GET_OUT_TIMING, EE_SET_OUT_TIMING, EE_RESET_OUT_TIMING
	global		EE_GET_OUT_GREYMAP, EE_SET_OUT_GREYMAP, EE_RESET_OUT_GREYMAP



	end

