; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"internal flash access procedure for ARCADEmini firmware"

	include		"P18F8520.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0, BITS1
	extern		A1_REG, A2_REG, A4_REG



;===========================================================================================
; imported lables
;===========================================================================================
	extern		IF_DATA_START, IF_DATA_END



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; get number of internal flash sectors
; output:	A4_REG		number of internal flash sectors
; changes:	WREG
;-------------------------------------------------------------------------------------------
IF_SECTOR_COUNT:
; get number of bytes in internal flash
	movlw	low IF_DATA_END		; end address
	movwf	A4_REG
	movlw	high IF_DATA_END
	movwf	A4_REG+1
	movlw	low IF_DATA_START	; subtract start address
	subwf	A4_REG
	movlw	high IF_DATA_START
	subwfb	A4_REG+1
; convert to number of sectors
	bcf	STATUS,C		; shift right 9 bit
	rrcf	A4_REG+1,W
	movwf	A4_REG
	clrf	A4_REG+1		; rest of bytes must be 0
	clrf	A4_REG+2		; (processor flash is small)
	clrf	A4_REG+3
	return

;-------------------------------------------------------------------------------------------
; read a "sector" of internal flash data
; input:	A4_REG		number of sector to read (1..x for compatibility with CF)
;		FSR0		address of memory where to store data (512 bytes)
; output:	STATUS,C	set on error (sector number too big)
; changes:	FSR0, TBLPTRU, TBLPTRH, TBLPTRL, TABLAT, A1_REG, WREG
;-------------------------------------------------------------------------------------------
IF_READ_SECTOR:
; check sector number
	movf	A4_REG+1,W		; byte 1, 2 or 3 != 0x00 -> too big 
	iorwf	A4_REG+2,W
	iorwf	A4_REG+3,W
	bnz	IF_READ_SECTOR_TOO_BIG
; subtract one from sector number to realize range 1..x
	decf	A4_REG+0
; convert to offset (shift left 9 bits)
	clrf	A4_REG+2
	movff	A4_REG+0,A4_REG+1
	clrf	A4_REG+0
	bcf	STATUS,C
	rlcf	A4_REG+0
	rlcf	A4_REG+1
	rlcf	A4_REG+2
; convert to address (add IF_DATA_START)
	movlw	low IF_DATA_START
	addwf	A4_REG
	movlw	high IF_DATA_START
	addwfc	A4_REG+1
	movlw	upper IF_DATA_START
	addwfc	A4_REG+2
; save this address in TBLPTR
	movff	A4_REG,TBLPTRL
	movff	A4_REG+1,TBLPTRH
	movff	A4_REG+2,TBLPTRU
; check if address is within valid range (address + 512 < IF_END)
	movlw	2			; add 512 to address
	addwf	A4_REG+1
	movlw	0
	addwfc	A4_REG+2
	movlw	upper IF_DATA_END	; check if within range
	cpfsgt	A4_REG+2
	bra	IF_READ_SECTOR_CHECK0
	bra	IF_READ_SECTOR_TOO_BIG
IF_READ_SECTOR_CHECK0:
	movlw	high IF_DATA_END
	cpfsgt	A4_REG+1
	bra	IF_READ_SECTOR_CHECK1
	bra	IF_READ_SECTOR_TOO_BIG
IF_READ_SECTOR_CHECK1:
	movlw	low IF_DATA_END
	cpfsgt	A4_REG+0
	bra	IF_READ_SECTOR_OK
	bra	IF_READ_SECTOR_TOO_BIG
; sector number too big
IF_READ_SECTOR_TOO_BIG:
; return error
	bsf	STATUS,C
	return
; sector number is okay
; (address is already in TBLPTR)
IF_READ_SECTOR_OK:
; read a "sector" (512 bytes)
	movlw	0x80			; configure for code read
	movwf	EECON1
	movlw	64			; read 64 times 8 byte
	movwf	A1_REG
IF_READ_SECTOR_LOOP:
	tblrd*+				; byte 0
	movff	TABLAT,POSTINC0
	tblrd*+				; byte 1
	movff	TABLAT,POSTINC0
	tblrd*+				; byte 2
	movff	TABLAT,POSTINC0
	tblrd*+				; byte 3
	movff	TABLAT,POSTINC0
	tblrd*+				; byte 4
	movff	TABLAT,POSTINC0
	tblrd*+				; byte 5
	movff	TABLAT,POSTINC0
	tblrd*+				; byte 6
	movff	TABLAT,POSTINC0
	tblrd*+				; byte 7
	movff	TABLAT,POSTINC0
	decfsz	A1_REG			; loop
	bra	IF_READ_SECTOR_LOOP
; return success
	bcf	STATUS,C
	return



;===========================================================================================
; global procedures
;===========================================================================================
	global		IF_SECTOR_COUNT, IF_READ_SECTOR



	end
