; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"USB procedures for ARCADEmini firmware"

	include		"P18F8520.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0, BITS1
	extern		A1_REG, B1_REG, C1_REG, D1_REG
	extern		A2_REG, B2_REG, C2_REG, D2_REG
	extern		A4_REG
	extern		O_FRAME0, O_FRAME1
	extern		CMD_SECTOR, CMD_OFFSET, CMD_STACK_PTR
	extern		CMD_OVERLAY, CMD_VARS, CMD_STACK
	extern		BUFFER
	extern		CF_SECTOR_CNT
	extern		TIME_SEC, TIME_MIN, TIME_HOUR
	extern		KEYS, KEYS_PRESSED, KEYS_RELEASED
	extern		SNES0, SNES1, SNES0_PRESSED, SNES1_PRESSED, SNES0_RELEASED, SNES1_RELEASED



;===========================================================================================
; imported procedures
;===========================================================================================

	extern		IF_SECTOR_COUNT, IF_READ_SECTOR
	extern		CF_READ_SECTOR, CF_WRITE_SECTOR
	extern		EE_GET_ID, EE_SET_ID
	extern		EE_GET_OUT_STYLE, EE_SET_OUT_STYLE
	extern		EE_GET_OUT_TIMING, EE_SET_OUT_TIMING
	extern		EE_GET_OUT_GREYMAP, EE_SET_OUT_GREYMAP



;===========================================================================================
					; constants
;===========================================================================================

; addresses of usb chip registers (USBN9604)
USB_MCNTRL	equ	h'00'
USB_CCONF	equ	h'01'
USB_FAR		equ	h'04'
USB_NFSR	equ	h'05'
USB_ALTEV	equ	h'08'
USB_NAKEV	equ	h'0E'
USB_NAKMSK	equ	h'0F'
USB_EPC0	equ	h'20'
USB_TXD0	equ	h'21'
USB_TXS0	equ	h'22'
USB_TXC0	equ	h'23'
USB_RXD0	equ	h'25'
USB_RXS0	equ	h'26'
USB_RXC0	equ	h'27'
USB_EPC1	equ	h'28'
USB_TXD1	equ	h'29'
USB_TXS1	equ	h'2A'
USB_TXC1	equ	h'2B'
USB_EPC2	equ	h'2C'
USB_RXD1	equ	h'2D'
USB_RXS1	equ	h'2E'
USB_RXC1	equ	h'2F'



;===========================================================================================
ram_def		UDATA		; default RAM (this page is selected by default)
;===========================================================================================

USB_BITS	res		1		; bits for usb
#define		USB_IS_INIT	USB_BITS,0	; if usb is initialized
#define		USB_S0_SEND	USB_BITS,1	; currently sending a message on endpoint 0
#define		USB_S0_PACKET	USB_BITS,2	; currently sending a packet of a message on endpoint 0
#define		USB_0_TOGGLE	USB_BITS,3	; toggle bit of endpoint 0
#define		USB_CONF	USB_BITS,4	; if we are configured
#define		USB_S1_TOGGLE	USB_BITS,5	; toggle bit of endpoint 1
#define		USB_S1_REQ	USB_BITS,6	; sending on endpoint 1 is requested
#define		USB_S1_ACT	USB_BITS,7	; currently sending on endpoint 1

USB_INIT_CNT	res		1		; initialization counter
#define		USB_INIT_TIME	40		; number of USB ticks for intialization of 3.3 volts

USB_R0_CNT	res		1		; counter for received bytes of EP0

USB_S0_LEN	res		1		; length of rest of message to send
USB_S0_LEN_OLD	res		1		; backup of above (needed when sending a packet again)
USB_S0_PTR	res		3		; pointer to rest of message to send (little endian)
USB_S0_PTR_OLD	res		3		; backup of above (needed when sending a packet again)

; buffer to receive messages from EP0 in
USB_R0_BUF	res		0		; subdivided into 2 1-byte regs and 3 2-byte regs
USB_BM_REQ_TYPE	res		1		; bmRequestType
USB_B_REQUEST	res		1		; bRequest
USB_W_VALUE	res		2		; wValue (little endian)
USB_W_INDEX	res		2		; wIndex (little endian)
USB_W_LENGTH	res		2		; wLength (little endian)

; length of buffer to send and receive messages on data endpoint
USB_CNT		res		1		; number of received bytes

; variable for wait mode for CF
USB_WAIT_CF	res		1
USB_WAIT_CF_NOT		equ	0x00		; not wainting for CF
USB_WAIT_CF_PRE_WRITE	equ	0x01		; wait for possibility to write
USB_WAIT_CF_WRITE	equ	0x02		; wait for write to finish
USB_WAIT_CF_PRE_READ	equ	0x03		; wait for possibility to read
USB_WAIT_CF_READ	equ	0x04		; wait for read to finish



;===========================================================================================
ram_usb		UDATA			; normal RAM for usb
;===========================================================================================

; buffer to send and receive messages on data endpoint
USB_BUF		res		64		; receive buffer



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; usb initialization
;-------------------------------------------------------------------------------------------
USB_INIT:
; reset usb chip
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_USB_nRESET		; set reset
	bsf	PIN_USB_nCS		; clear chip select
	bsf	PIN_USB_nRESET		; clear reset
; initialize variables
	clrf	USB_INIT_CNT		; reset init counter
	bcf	USB_IS_INIT		; usb not yet initialized
	bcf	BIT_USB_PWR_OK		; power may not be taken from usb
	bcf	BIT_USB_CMD_STOP	; do not stop command execution
	bcf	BIT_USB_CMD_ST_N	; do not stop command execution after next command
	bcf	BIT_USB_CMD_ST_D        ; command excution stop after next command has not been done
	movlw	USB_WAIT_CF_NOT		; not wainting for CF
	movwf	USB_WAIT_CF
; done
	return

;-------------------------------------------------------------------------------------------
; read a usb chip register
; input:	W		register address (0..63)
; output:	A1_REG		value read from register
; changes:	-
;-------------------------------------------------------------------------------------------
USB_READ_REG:
; disable interrupts (data port is shared)
	bcf	INTCON,6
	bcf	INTCON,7
; read register
	movwf	PORT_SH_D		; write register address to data port
	movlw	PORT_SH_D_TR_O		; set data port to output
	movwf	PORT_SH_D_TRIS
	setf	PORT_SH_A		; output address select bit (A0=1)
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_SH_nWE		; set write enable
	bsf	PIN_SH_nWE		; clear write enable
	bsf	PIN_USB_nCS		; clear chip select
	movlw	PORT_SH_D_TR_I		; set data port back to input
	movwf	PORT_SH_D_TRIS
	clrf	PORT_SH_A		; output data select bit (A0=0)
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_SH_nOE		; set output enable
	movff	PORT_SH_D,A1_REG	; read data
	bsf	PIN_SH_nOE		; clear output enable
	bsf	PIN_USB_nCS		; clear chip select
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; enable interrupts
	bsf	INTCON,7
	bsf	INTCON,6
; done
	return

;-------------------------------------------------------------------------------------------
; write a usb chip register
; input:	W		register address (0..63)
;		A1_REG		value to write to register
; changes:	-
;-------------------------------------------------------------------------------------------
USB_WRITE_REG:
; disable interrupts (data port is shared)
	bcf	INTCON,6
	bcf	INTCON,7
; write register
	movwf	PORT_SH_D		; write register address to data port
	movlw	PORT_SH_D_TR_O		; set data port to output
	movwf	PORT_SH_D_TRIS
	setf	PORT_SH_A		; output address select bit (A0=1)
	movlw	PORT_SH_A_TR_O		; set address port to output
	movwf	PORT_SH_A_TRIS
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_SH_nWE		; set write enable
	bsf	PIN_SH_nWE		; clear write enable
	bsf	PIN_USB_nCS		; clear chip select
	movff	A1_REG,PORT_SH_D	; write data
	clrf	PORT_SH_A		; output data select bit (A0=0)
	bcf	PIN_USB_nCS		; set chip select
	bcf	PIN_SH_nWE		; set write enable
	bsf	PIN_SH_nWE		; clear write enable
	bsf	PIN_USB_nCS		; clear chip select
	movlw	PORT_SH_D_TR_I		; set data port back to input
	movwf	PORT_SH_D_TRIS
	movlw	PORT_SH_A_TR_I		; set address port back to input
	movwf	PORT_SH_A_TRIS
; enable interrupts
	bsf	INTCON,7
	bsf	INTCON,6
; done
	return

;-------------------------------------------------------------------------------------------
; usb task
; - called periodically
; changes:	PRODH, PRODL, A1_REG, B1_REG, C1_REG, TBLPTR, TABLAT, WREG
;-------------------------------------------------------------------------------------------
USB_TASK:
; check if usb is initialized
	btfsc	USB_IS_INIT		; if usb is initialized ...
	bra	USB_PROCESS		; ... jump to processing
	bra	USB_INITIALIZE		; jump to initialization

;-------------------------------------------------------------------------------------------
; usb initial state
;-------------------------------------------------------------------------------------------
USB_INITIALIZE:
; check if already waiting for 3.3 volts to come up
	movf	USB_INIT_CNT
	bnz	USB_INIWAIT
; setup usb chip
	movlw	b'00000100'		; 3.3 volts on, no node attach, no interrupt output
	movwf	A1_REG
	movlw	USB_MCNTRL
	call	USB_WRITE_REG
	movlw	b'10001011'		; clock output off, divisor to 4 Mhz
	movwf	A1_REG
	movlw	USB_CCONF
	call	USB_WRITE_REG
; finished first step of initialization
	incf	USB_INIT_CNT
	return

; wait a few steps to give 3.3 volts time to come up
USB_INIWAIT:
	incf	USB_INIT_CNT
	movlw	USB_INIT_TIME
	cpfsgt	USB_INIT_CNT
	return
; go on with initializing usb chip
	movlw	b'00001100'	; 3.3 volts still on, but node attach on, no interrupt output
	movwf	A1_REG
	movlw	USB_MCNTRL
	call	USB_WRITE_REG
; now, usb is initialized
	bsf	USB_IS_INIT
; enable NAK events
	movlw	b'11111111'	; NAKMSK: NAK_*=1
	movwf	A1_REG
	movlw	USB_NAKMSK
	call	USB_WRITE_REG
; read ALTEV to clear node reset bit
	movlw	USB_ALTEV
	call	USB_READ_REG
; done - continue with node reset
	bra	USB_NODE_RESET

;-------------------------------------------------------------------------------------------
; usb node reset processing
;-------------------------------------------------------------------------------------------
USB_NODE_RESET:
; set node functional state to node reset
	movlw	b'00000000'	; NFSR: NFS=NodeReset
	movwf	A1_REG
	movlw	USB_NFSR
	call	USB_WRITE_REG
; done - continue with start
	bra	USB_START

;-------------------------------------------------------------------------------------------
; usb start (after initialization or node reset)
;-------------------------------------------------------------------------------------------
USB_START:
; set node address to 0
	movlw	b'10000000'	; FAR: AD_EN=1, AD=0
	movwf	A1_REG
	movlw	USB_FAR
	call	USB_WRITE_REG
; initialize enpoint 0
	movlw	b'00000000'	; EPC0: STALL=0, DEF=0, EP=0
	movwf	A1_REG
	movlw	USB_EPC0
	call	USB_WRITE_REG
	movlw	b'00001000'	; TXC0: IGN_IN=0, FLUSH=1, TOGGLE=0, TX_EN=0
	movwf	A1_REG		; FLUSH setzen und TX_EN rcksetzen
	movlw	USB_TXC0
	call	USB_WRITE_REG
; initialize variables
	bcf	USB_0_TOGGLE	; expect TOGGLE=0 for EP0
	bcf	USB_S0_SEND	; not sending a message on endpoint 0
	bcf	USB_S0_PACKET	; not sending a packet on endpoint 0
	clrf	USB_R0_CNT	; nothing received so far
	bcf	USB_CONF	; we are not configured
	call	USB_CONF_OFF
	bcf	BIT_USB_PWR_OK	; power may not be taken from usb
	bcf	USB_S1_TOGGLE	; TOGGLE=0 for first packet sent on endpoint 1
	bcf	USB_S1_REQ	; no request to send a message on endpoint 1
	bcf	USB_S1_ACT	; not sending a message on endpoint 1
; allow reception
	movlw	b'00000001'	; RXC0: FLUSH=0, IGN_SETUP=0, IGN_OUT=0, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
; set node functional state to operational
	movlw	b'00000010'	; NFSR: NFS=NodeOperational
	movwf	A1_REG
	movlw	USB_NFSR
	call	USB_WRITE_REG
; done
	return

;-------------------------------------------------------------------------------------------
; usb processing
;-------------------------------------------------------------------------------------------
USB_PROCESS:
; check for node reset
	movlw	USB_ALTEV		; read ALTEV
	call	USB_READ_REG
	btfsc	A1_REG,6		; ALTEV: RESET=1 ===> node reset was detected
	bra	USB_NODE_RESET

; sending on EP0?
	btfsc	USB_S0_PACKET		; sending a packet
	bra	USB_CONFIG_SEND_PACKET
	btfsc	USB_S0_SEND		; data to send is available
	bra	USB_CONFIG_SEND

; receive messages
; get receive status
	movlw	USB_RXS0		; read RXS0
	call	USB_READ_REG
	movf	A1_REG,W
	movwf	C1_REG			; remember RXS0 in C1_REG
	andlw	b'00001111'		; RXS0: RCOUNT - put number of received bytes into B1_REG
	movwf	B1_REG
	bz	USB_CONFIG_RECV_READ_DONE	; no bytes to read
; get recevied bytes
USB_CONFIG_RECV_READ:
	movf	USB_R0_CNT,W		; get offset into receive buffer
	andlw	b'00000111'		;   - use only last 3 bits because buffer length is 8
					;   - longest expected message is 8 bytes in length
	lfsr	FSR0,USB_R0_BUF		; get address within buffer
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	USB_RXD0		; RXD0 - read a data byte from fifo
	call	USB_READ_REG
	movff	A1_REG,INDF0
	incf	USB_R0_CNT		; count received bytes
	decfsz	B1_REG			; count down bytes in fifo
	bra	USB_CONFIG_RECV_READ
USB_CONFIG_RECV_READ_DONE:
; check if message complete
	btfss	C1_REG,4		; if packet was not last one (RXS0: LAST=0) ...
	bra	USB_CONFIG_RECV_END	; ... jump to end of reception
; check if message is a valid setup message
	bcf	USB_0_TOGGLE		; store inverted toggle bit (used later for sending)
	btfss	C1_REG,5		; (RXS0: TOGGLE)
	bsf	USB_0_TOGGLE
	btfss	C1_REG,6		; if not a setup message (RXS0: SETUP=0) ...
	bra	USB_CONFIG_RECV_DONE	; ... ignore this message
	movf	USB_R0_CNT,W		; if not 8 bytes in length ...
	xorlw	8
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ	; ... treat this message as unknown request
; interpret setup message and prepare to send answer
	movlw	0x00			; bRequest=0x00?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_00
	movlw	0x01			; bRequest=0x01?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_01
	movlw	0x05			; bRequest=0x05?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_05
	movlw	0x06			; bRequest=0x06?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_06
	movlw	0x08			; bRequest=0x08?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_08
	movlw	0x09			; bRequest=0x09?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_09
	movlw	0x0A			; bRequest=0x0A?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_0A
	movlw	0x0B			; bRequest=0x0B?
	xorwf	USB_B_REQUEST,W
	bz	USB_CONFIG_REQ_0B
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x00
USB_CONFIG_REQ_00:
	movlw	0x80			; bmRequestType=0x80?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_STATUS_DEVICE
	movlw	0x81			; bmRequestType=0x81?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_STATUS_INTERFACE
	movlw	0x82			; bmRequestType=0x82?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_STATUS_ENDPOINT
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x01
USB_CONFIG_REQ_01:
	movlw	0x00			; bmRequestType=0x00?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_CLEAR_FEAT_DEVICE
	movlw	0x01			; bmRequestType=0x01?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_CLEAR_FEAT_INTERFACE
	movlw	0x02			; bmRequestType=0x02?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_CLEAR_FEAT_ENDPOINT
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x05
USB_CONFIG_REQ_05:
	movlw	0x00			; bmRequestType=0x00?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_SET_ADDRESS
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x06
USB_CONFIG_REQ_06:
	movlw	0x00			; bmRequestType=0x00/0x80?
	xorwf	USB_BM_REQ_TYPE,W
	andlw	0x7F
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_DESCR
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x08
USB_CONFIG_REQ_08:
	movlw	0x80			; bmRequestType=0x80?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_CONFIG
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x09
USB_CONFIG_REQ_09:
	movlw	0x00			; bmRequestType=0x00?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_SET_CONFIG
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x0A
USB_CONFIG_REQ_0A:
	movlw	0x81			; bmRequestType=0x81?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_GET_INTERFACE
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; interpret bRequest=0x0B
USB_CONFIG_REQ_0B:
	movlw	0x01			; bmRequestType=0x01?
	xorwf	USB_BM_REQ_TYPE,W
	btfsc	STATUS,Z
	bra	USB_CONFIG_SET_INTERFACE
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; request is "get status device/interface/endpoint"
USB_CONFIG_GET_STATUS_DEVICE:
USB_CONFIG_GET_STATUS_INTERFACE:
USB_CONFIG_GET_STATUS_ENDPOINT:
	movwf	USB_W_VALUE+0,W		; wValue=0x0000, wIndex=0x0000, wLength=0x0002?
	iorwf	USB_W_VALUE+1,W
	iorwf	USB_W_INDEX+0,W
	iorwf	USB_W_INDEX+1,W
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	0x02
	xorwf	USB_W_LENGTH+0,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	low USB_ANSWER_00_00	; send 0x00, 0x00
	movwf	TBLPTRL
	movlw	high USB_ANSWER_00_00
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_00_00
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "get status device/interface/endpoint"
USB_CONFIG_CLEAR_FEAT_DEVICE:
	movwf	USB_W_INDEX+0,W		; wIndex=x0000?
	iorwf	USB_W_INDEX+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
USB_CONFIG_CLEAR_FEAT_INTERFACE:
USB_CONFIG_CLEAR_FEAT_ENDPOINT:
	movwf	USB_W_LENGTH+0,W	; wLength=x0000?
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "set address"
USB_CONFIG_SET_ADDRESS:
	movlw	b'01000000'	; EPC0: STALL=0, DEF=1, EP=0
	movwf	A1_REG
	movlw	USB_EPC0
	call	USB_WRITE_REG
	movlw	b'10000000'	; FAR: AD_EN=1, AD=address
	iorwf	USB_W_VALUE+0,W
	movwf	A1_REG
	movlw	USB_FAR		
	call	USB_WRITE_REG
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "get descriptor"
USB_CONFIG_GET_DESCR:
	movlw	0x01			; wValue=0x01__?
	xorwf	USB_W_VALUE+1,W
	bz	USB_CONFIG_GET_DESCR_DEV
	movlw	0x02			; wValue=0x02__?
	xorwf	USB_W_VALUE+1,W
	bz	USB_CONFIG_GET_DESCR_CONF
	movlw	0x03			; wValue=0x03__?
	xorwf	USB_W_VALUE+1,W
	bz	USB_CONFIG_GET_DESCR_STR
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; request is "get device descriptor"
USB_CONFIG_GET_DESCR_DEV:
	movlw	low USB_DESCR_DEV	; send device desriptor
	movwf	TBLPTRL
	movlw	high USB_DESCR_DEV
	movwf	TBLPTRH
	movlw	upper USB_DESCR_DEV
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get configuration descriptor"
USB_CONFIG_GET_DESCR_CONF:
	movlw	low USB_DESCR_CONF	; send configuration desriptor (with subdescriptors)
	movwf	TBLPTRL
	movlw	high USB_DESCR_CONF
	movwf	TBLPTRH
	movlw	upper USB_DESCR_CONF
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get string descriptor"
USB_CONFIG_GET_DESCR_STR:
	movlw	0x00			; wValue=0x__00?
	xorwf	USB_W_VALUE,W
	bz	USB_CONFIG_GET_DESCR_STR0
	movlw	0x01			; wValue=0x__01?
	xorwf	USB_W_VALUE,W
	bz	USB_CONFIG_GET_DESCR_STR1
	movlw	0x02			; wValue=0x__02?
	xorwf	USB_W_VALUE,W
	bz	USB_CONFIG_GET_DESCR_STR2
	bra	USB_CONFIG_UNKNOWN_REQ	; unknown request
; request is "get string descriptor 0"
USB_CONFIG_GET_DESCR_STR0:
	movlw	low USB_DESCR_STR0	; send string desriptor 0
	movwf	TBLPTRL
	movlw	high USB_DESCR_STR0
	movwf	TBLPTRH
	movlw	upper USB_DESCR_STR0
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get string descriptor 1"
USB_CONFIG_GET_DESCR_STR1:
	movlw	low USB_DESCR_STR1	; send string desriptor 1
	movwf	TBLPTRL
	movlw	high USB_DESCR_STR1
	movwf	TBLPTRH
	movlw	upper USB_DESCR_STR1
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get string descriptor 2"
USB_CONFIG_GET_DESCR_STR2:
	movlw	low USB_DESCR_STR2	; send string desriptor 2
	movwf	TBLPTRL
	movlw	high USB_DESCR_STR2
	movwf	TBLPTRH
	movlw	upper USB_DESCR_STR2
	movwf	TBLPTRU
	movff	USB_W_LENGTH,A1_REG	; put maximum length into A1_REG
	tstfsz	USB_W_LENGTH+1
	setf	A1_REG
	bra	USB_CONFIG_ANSWER_LEN
; request is "get configuration"
USB_CONFIG_GET_CONFIG:
	movwf	USB_W_VALUE+0,W		; wValue=0x0000, wIndex=0x0000, wLength=0x0001?
	iorwf	USB_W_VALUE+1,W
	iorwf	USB_W_INDEX+0,W
	iorwf	USB_W_INDEX+1,W
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	0x01
	xorwf	USB_W_LENGTH+0,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	btfsc	USB_CONF		; are we configured?
	bra	USB_CONFIG_GET_CONFIG_YES
	movlw	low USB_ANSWER_00	; send 0x00
	movwf	TBLPTRL
	movlw	high USB_ANSWER_00
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_00
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
USB_CONFIG_GET_CONFIG_YES:
	movlw	low USB_ANSWER_01	; send 0x01
	movwf	TBLPTRL
	movlw	high USB_ANSWER_01
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_01
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "set configuration"
USB_CONFIG_SET_CONFIG:
	movwf	USB_W_INDEX+0,W		; wIndex=0x0000, wLength=0x0001?
	iorwf	USB_W_INDEX+1,W
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movf	USB_W_VALUE,W		; configuration or deconfiguration?
	bnz	USB_CONFIG_SET_CONFIG_YES
	bcf	USB_CONF		; deconfiguration
	call	USB_CONF_OFF
	bcf	BIT_USB_PWR_OK		; power may not be taken from usb
	bra	USB_CONFIG_SET_CONFIG_ANSWER
USB_CONFIG_SET_CONFIG_YES:
	bsf	USB_CONF		; configuration
	call	USB_CONF_ON
	bsf	BIT_USB_PWR_OK		; power may be taken from usb
USB_CONFIG_SET_CONFIG_ANSWER:
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "get interface"
USB_CONFIG_GET_INTERFACE:
	movwf	USB_W_VALUE+0,W		; wValue=0x0000, wLength=0x0001?
	iorwf	USB_W_VALUE+1,W
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	0x01
	xorwf	USB_W_LENGTH+0,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	low USB_ANSWER_00	; send 0x00
	movwf	TBLPTRL
	movlw	high USB_ANSWER_00
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_00
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; request is "set interface"
USB_CONFIG_SET_INTERFACE:
	movwf	USB_W_LENGTH+0,W	; wLength=0x0001?
	iorwf	USB_W_LENGTH+1,W
	btfss	STATUS,Z
	bra	USB_CONFIG_UNKNOWN_REQ
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
;unknown request
USB_CONFIG_UNKNOWN_REQ:
	movlw	low USB_ANSWER_EMPTY	; send empty answer
	movwf	TBLPTRL
	movlw	high USB_ANSWER_EMPTY
	movwf	TBLPTRH
	movlw	upper USB_ANSWER_EMPTY
	movwf	TBLPTRU
	bra	USB_CONFIG_ANSWER
; received message is processed
USB_CONFIG_RECV_DONE:
; enable reception of new message on EP0
	clrf	USB_R0_CNT		; empty EP0 receive buffer
	movlw	b'00001000'		; RXC0: FLUSH=1, IGN_SETUP=0, IGN_OUT=0, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	movlw	b'00000001'		; RXC0: FLUSH=0, IGN_SETUP=0, IGN_OUT=0, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
; end of reception branch of configuration stuff
USB_CONFIG_RECV_END:
	bra	USB_CONFIG_END

; start sending answer at TBLPTR
USB_CONFIG_ANSWER:
	setf	A1_REG			; any length is allowed
; start sending answer at TBLPTR with max length A1_REG
USB_CONFIG_ANSWER_LEN:
	movlw	0x80			; configure for code read
	movwf	EECON1
	tblrd*+				; read length of table
	movff	TABLAT,USB_S0_LEN
	tblrd*+				; increment pointer
	movf	A1_REG,W		; max. length from A1_REG
	cpfslt	USB_S0_LEN
	movwf	USB_S0_LEN
	movff	TBLPTRL,USB_S0_PTR+0	; save pointer
	movff	TBLPTRH,USB_S0_PTR+1
	movff	TBLPTRU,USB_S0_PTR+2
	bsf	USB_S0_SEND		; data to send is now available
	bcf	USB_S0_PACKET		; no packet is being sent at the moment
					; - packet transmission is aborted if active
	bra	USB_CONFIG_SEND

; sending a packet
USB_CONFIG_SEND_PACKET:
; get send status
	movlw	USB_TXS0		; read TXS0
	call	USB_READ_REG		; and store its contents in A1_REG
; check if packet was sent
	btfss	A1_REG,5		; jump if packet was not send (TXS0: TX_DONE=0)
	bra	USB_CONFIG_SEND_END
; check if packet was successfully sent
	btfsc	A1_REG,6		; jump if packet was successfully sent (TXS0: ACK_STAT=1)
	bra	USB_CONFIG_SEND_OK
; restore old state, packet needs to be sent again
	movff	USB_S0_LEN_OLD,USB_S0_LEN
	movff	USB_S0_PTR_OLD+0,USB_S0_PTR+0
	movff	USB_S0_PTR_OLD+1,USB_S0_PTR+1
	movff	USB_S0_PTR_OLD+2,USB_S0_PTR+2
	bsf	USB_S0_SEND		; there is data to be sent
USB_CONFIG_SEND_OK:
	bcf	USB_S0_PACKET		; no packet is being sent
	btg	USB_0_TOGGLE		; invert toggle bit
; check if there is more data to send
	btfsc	USB_S0_SEND		; jump if more data to send is available
	bra	USB_CONFIG_SEND
; no more data to send available
; flush FIFO and enable reception on EP0
	clrf	USB_R0_CNT		; empty EP0 receive buffer
	movlw	b'00001000'		; RXC0: FLUSH=1, IGN_SETUP=0, IGN_OUT=0, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	movlw	b'00000001'		; RXC0: FLUSH=0, IGN_SETUP=0, IGN_OUT=0, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	bra	USB_CONFIG_SEND_END

; data to send is available - load the next packet to send
; if we get here, USB_S0_SEND is always set and data to send is available
USB_CONFIG_SEND:
; disable reception and clear buffer on EP0
	movlw	b'00001000'		; RXC0: FLUSH=1, IGN_SETUP=0, IGN_OUT=0, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	movlw	b'00001000'		; TXC0: IGN_IN=0, FLUSH=1, TOGGLE=0, TX_EN=0
	movwf	A1_REG
	movlw	USB_TXC0
	call	USB_WRITE_REG
; save state for the case that packet fails and needs to be sent again
	movff	USB_S0_LEN,USB_S0_LEN_OLD
	movff	USB_S0_PTR+0,USB_S0_PTR_OLD+0
	movff	USB_S0_PTR+1,USB_S0_PTR_OLD+1
	movff	USB_S0_PTR+2,USB_S0_PTR_OLD+2
; check if to send 8 bytes or more
	movlw	8			; subtract 8 from length
	subwf	USB_S0_LEN,W		; and store result in WREG
	btfss	STATUS,C
	bra	USB_CONFIG_SEND_LAST_PACKET	; jump if packet is less than 8 bytes
	movwf	USB_S0_LEN		; save new length (now 8 bytes less)
	movlw	8			; 8 bytes to load into fifo
	movwf	B1_REG
	bra	USB_CONFIG_SEND_LOAD_FIFO
; prepare for last packet
USB_CONFIG_SEND_LAST_PACKET:
	movff	USB_S0_LEN,B1_REG	; remember number of bytes to load into fifo
	clrf	USB_S0_LEN		; entire message has been sent
	bcf	USB_S0_SEND
	movf	B1_REG,W		; check if empty packet
	bz	USB_CONFIG_SEND_START	; nothing needs to be loaded into fifo for an empty packet
; load a packet data into fifo
USB_CONFIG_SEND_LOAD_FIFO:
	movff	USB_S0_PTR+0,TBLPTRL	; get pointer to rest of packet
	movff	USB_S0_PTR+1,TBLPTRH
	movff	USB_S0_PTR+2,TBLPTRU
	movlw	0x80			; configure for code read
	movwf	EECON1
USB_CONFIG_SEND_LOAD_FIFO_LOOP:
	tblrd*+				; get next byte
	movff	TABLAT,A1_REG		; write byte into fifo
	movlw	USB_TXD0
	call	USB_WRITE_REG
	decfsz	B1_REG			; count down bytes to send
	bra	USB_CONFIG_SEND_LOAD_FIFO_LOOP
	movff	TBLPTRL,USB_S0_PTR+0	; save new pointer to rest of packet
	movff	TBLPTRH,USB_S0_PTR+1
	movff	TBLPTRU,USB_S0_PTR+2
; start sending the packet
USB_CONFIG_SEND_START:
	bsf	USB_S0_PACKET		; packet is being sent now
	movlw	b'00000001'		; TXC0: IGN_IN=0, FLUSH=0, TOGGLE=0, TX_EN=1
	btfsc	USB_0_TOGGLE		; write right toggle bit
	iorlw	b'00000100'
	movwf	A1_REG
	movlw	USB_TXC0
	call	USB_WRITE_REG
	movlw	USB_NAKEV		; read NAKEV to clear NAK events
	call	USB_READ_REG		; (it might be possible that there is an old NAK event pending)
; end of configuration send stuff
USB_CONFIG_SEND_END:

; check for EP0 BULK OUT NAK events
; - i.e. if host tries to send and we are not listening
	movlw	USB_NAKEV		; read NAKEV
	call	USB_READ_REG		; and store its contents in A1_REG
	btfss	A1_REG,4		; check if NAK bit for BULK OUT on EP0 is set
	bra	USB_CONFIG_NAK_END	; not set
; clear send bits for EP0
	bcf	USB_S0_SEND
	bcf	USB_S0_PACKET
; cancel transmission on EP0
	movlw	b'00001000'		; TXC0: IGN_IN=0, FLUSH=1, TOGGLE=0, TX_EN=0
	movwf	A1_REG
	movlw	USB_TXC0
	call	USB_WRITE_REG
; clear FIFO and enable reception on EP0
	clrf	USB_R0_CNT		; empty EP0 receive buffer
	movlw	b'00001000'		; RXC0: FLUSH=1, IGN_SETUP=0, IGN_OUT=0, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	movlw	b'00000001'		; RXC0: FLUSH=0, IGN_SETUP=0, IGN_OUT=0, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC0
	call	USB_WRITE_REG
	bra	USB_CONFIG_SEND_END
; end of configuration NAK stuff
USB_CONFIG_NAK_END:

; end of configuration stuff
USB_CONFIG_END:

; return if not configured
	btfss	USB_CONF
	return

; send answer if command execution stop after next command has been done
	btfsc	USB_S1_REQ		; jump if buffer in use
	bra	USB_CMD_ST_D_END
	btfss	BIT_USB_CMD_ST_D	; jump if not done
	bra	USB_CMD_ST_D_END
; clear flag bit for "command excution stop after next command has been done"
	bcf	BIT_USB_CMD_ST_D
; send answer 0x02: a single command was processed
	movlw	0x01
	movwf	USB_CNT
	movlw	0x02
	movff	WREG,USB_BUF
	bsf	USB_S1_REQ
USB_CMD_ST_D_END:

; check if waiting for compact flash
	movf	USB_WAIT_CF,W
	xorlw	USB_WAIT_CF_NOT
	bz	USB_WAIT_CF_END
; exit if CF is still busy
	btfsc	BIT_CF_BUSY
	return

; we are waiting for a possibility to write
	movlw	USB_WAIT_CF_PRE_WRITE
	cpfseq	USB_WAIT_CF
	bra	USB_WAIT_CF_PRE_WRITE_END
; no more waiting
	movlw	USB_WAIT_CF_NOT
	movwf	USB_WAIT_CF
; now CF is ready, so do the write
	lfsr	FSR0,USB_BUF		; initialize pointer to USB message containing sector number, etc.
	bra	USB_RECV_42_DO_IT
USB_WAIT_CF_PRE_WRITE_END:

; we are waiting for a write to finish
	movlw	USB_WAIT_CF_WRITE
	cpfseq	USB_WAIT_CF
	bra	USB_WAIT_CF_WRITE_END
; no more waiting
	movlw	USB_WAIT_CF_NOT
	movwf	USB_WAIT_CF
; now CF is ready, so put error code into WREG and finish write
	movlw	0x00			; success
	btfsc	BIT_CF_ERROR
	movlw	0x01			; error
	bra	USB_RECV_42_DONE
USB_WAIT_CF_WRITE_END:

; we are waiting for a possibility to read
	movlw	USB_WAIT_CF_PRE_READ
	cpfseq	USB_WAIT_CF
	bra	USB_WAIT_CF_PRE_READ_END
; no more waiting
	movlw	USB_WAIT_CF_NOT
	movwf	USB_WAIT_CF
; now CF is ready, so do the write
	lfsr	FSR0,USB_BUF		; initialize pointer to USB message containing sector number, etc.
	bra	USB_RECV_43_DO_IT
USB_WAIT_CF_PRE_READ_END:

; we are waiting for a read to finish
	movlw	USB_WAIT_CF_READ
	cpfseq	USB_WAIT_CF
	bra	USB_WAIT_CF_READ_END
; no more waiting
	movlw	USB_WAIT_CF_NOT
	movwf	USB_WAIT_CF
; now CF is ready, so put error code into WREG and finish read
	movlw	0x00			; success
	btfsc	BIT_CF_ERROR
	movlw	0x01			; error
	bra	USB_RECV_43_DONE
USB_WAIT_CF_READ_END:

USB_WAIT_CF_END:

; do not receive messages if a message to send is still in the buffer
	btfsc	USB_S1_REQ		; jump if buffer in use
	goto	USB_RECV_END
; receive packets on data endpoint 1
	movlw	USB_RXS1		; get receive state from RXS1
	call	USB_READ_REG
	btfss	A1_REG,7		; jump if no error (RX_ERR=0)
	bra	USB_RECV_NO_ERR
	movlw	b'00001100'		; RXC1: FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC1 
	call	USB_WRITE_REG
	movlw	b'00000101'		; RXC1: FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC1 
	call	USB_WRITE_REG
	goto	USB_RECV_END
USB_RECV_NO_ERR:
	btfss	A1_REG,4		; jump to end if packet not complete (RX_LAST=0)
	goto	USB_RECV_END
	clrf	USB_CNT			; clear number of received bytes
	movf	A1_REG,W
	andlw	b'00001111'
	movwf	B1_REG			; put number of received bytes into B1_REG
	btfsc	STATUS,Z		; jump if empty packet
	bra	USB_RECV_DO_IT
; read FIFO
	lfsr	FSR0,USB_BUF		; get address of receive buffer
USB_RECV_LOOP:
	movlw	USB_RXD1		; read a byte from FIFO
	call 	USB_READ_REG
	movff	A1_REG,POSTINC0
	incf	USB_CNT			; count number of received bytes
	decfsz	B1_REG
	bra	USB_RECV_LOOP	
	movlw	USB_RXS1		; get receive status
	call	USB_READ_REG
	movf	A1_REG,W
	andlw	b'00001111'
	movwf	B1_REG			; put number of received bytes into B1_REG
	btfss	STATUS,Z
	bra	USB_RECV_LOOP

; we received a message
USB_RECV_DO_IT:
	movf	USB_CNT,W		; ignore empty messages
	btfsc	STATUS,Z
	goto	USB_RECV_GO_ON
	lfsr	FSR0,USB_BUF
	movlw	0x0F
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_0
	movlw	0x1F 
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_1
	movlw	0x2F
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_2
	movlw	0x3F
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_3
	movlw	0x4F
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_4
	movlw	0x5F
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_5
	movlw	0x6F
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_6
	movlw	0x7F
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_7
	movlw	0x8F
	cpfsgt	INDF0
	bra	USB_RECV_DO_IT_8
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x0_
USB_RECV_DO_IT_0:
	movlw	0x00
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_00
	movlw	0x01
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_01
	movlw	0x02
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_02
	movlw	0x08
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_08
	movlw	0x09
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_09
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x1_
USB_RECV_DO_IT_1:
	movlw	0x10
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_10
	movlw	0x11
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_11
	movlw	0x12
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_12
	movlw	0x13
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_13
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x2_
USB_RECV_DO_IT_2:
	movlw	0x20
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_20
	movlw	0x21
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_21
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x3_
USB_RECV_DO_IT_3:
	movlw	0x30
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_30
	movlw	0x31
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_31
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x4_
USB_RECV_DO_IT_4:
	movlw	0x40
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_40
	movlw	0x42
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_42
	movlw	0x43
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_43
	movlw	0x44
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_44
	movlw	0x47
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_47
	movlw	0x48
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_48
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x5_
USB_RECV_DO_IT_5:
	movlw	0x50
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_50
	movlw	0x51
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_51
	movlw	0x54
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_54
	movlw	0x55
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_55
	movlw	0x56
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_56
	movlw	0x57
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_57
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x6_
USB_RECV_DO_IT_6:
	movlw	0x60
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_60
	movlw	0x62
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_62
	movlw	0x63
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_63
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x7_
USB_RECV_DO_IT_7:
	movlw	0x70
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_70
	movlw	0x71
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_71
	goto	USB_RECV_GO_ON		; unknown message, ignore
; message received is 0x8_
USB_RECV_DO_IT_8:
	movlw	0x80
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_80
	movlw	0x84
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_84
	movlw	0x85
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_85
	movlw	0x88
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_88
	movlw	0x89
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_89
	movlw	0x8A
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_8A
	movlw	0x8B
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_8B
	movlw	0x8C
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_8C
	movlw	0x8D
	xorwf	INDF0,W
	btfsc	STATUS,Z
	bra	USB_RECV_8D
	goto	USB_RECV_GO_ON		; unknown message, ignore

; message 0x00: turn off command processing
USB_RECV_00:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; stop command execution
	bsf	BIT_USB_CMD_STOP
; send answer 0x00: command processing turned off
	movlw	0x01
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x01: turn on command processing
USB_RECV_01:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; enable command execution
	bcf	BIT_USB_CMD_STOP
; send answer 0x01: command processing turned on
	movlw	0x01
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x02: process a single command
USB_RECV_02:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; execute a single command
	bcf	BIT_USB_CMD_STOP	; do not stop command execution
	bsf	BIT_USB_CMD_ST_N	; stop command execution after next command
; message processed
	bra	USB_RECV_GO_ON

; message 0x08 <sector (4 byte, big endian)> <offset (2 byte, big endian)>: set command pointer
USB_RECV_08:
; check length
	movf	USB_CNT,W
	xorlw	0x07
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; set sector
	movff	PREINC0,CMD_SECTOR+3
	movff	PREINC0,CMD_SECTOR+2
	movff	PREINC0,CMD_SECTOR+1
	movff	PREINC0,CMD_SECTOR+0
; set offset
	movf	PREINC0,W
	andlw	0x01			; AND high-byte with 0x01 (command pointer has 9 bit)
	movwf	INDF0			; write ANDed value back (for sending answer)
	movwf	CMD_OFFSET+1
	movff	PREINC0,CMD_OFFSET+0
; send answer 0x08 <sector (4 byte, big endian)> <offset (2 byte, big endian)>: command pointer set
	movlw	0x07
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x09: get command pointer
USB_RECV_09:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x09 <sector (4 byte, big endian)> <offset (2 byte, big endian)>: value of command pointer
	movlw	0x07
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x09
	movwf	POSTINC0
	movff	CMD_SECTOR+3,POSTINC0	; sector
	movff	CMD_SECTOR+2,POSTINC0
	movff	CMD_SECTOR+1,POSTINC0
	movff	CMD_SECTOR+0,POSTINC0
	movff	CMD_OFFSET+1,POSTINC0	; offset
	movff	CMD_OFFSET+0,POSTINC0
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x10 <0x00..0x04> <52 bytes>: load 1/5 of pixel buffer
USB_RECV_10:
; check length
	movf	USB_CNT,W
	xorlw	0x36
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x04) of parameter
	movf	PREINC0,W
	sublw	0x04
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; get pointer to buffer not owned by interrupt into FSR1
	lfsr	FSR1,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR1,O_FRAME1
; continue with common code for message 0x10 and 0x12
	bra	USB_RECV_10_12

; message 0x11 <0x00..0x04>: read 1/5 of pixel buffer
USB_RECV_11:
; check length
	movf	USB_CNT,W
	xorlw	0x02
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x04) of parameter
	movf	PREINC0,W
	sublw	0x04
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; get pointer to buffer not owned by interrupt into FSR1
	lfsr	FSR1,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR1,O_FRAME1
; continue with common code for message 0x11 and 0x13
	bra	USB_RECV_11_13

; message 0x12 <0x00..0x04> <52 bytes>: load 1/5 of overlay buffer
USB_RECV_12:
; check length
	movf	USB_CNT,W
	xorlw	0x36
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x04) of parameter
	movf	PREINC0,W
	sublw	0x04
	btfss	STATUS,C
	return				; invalid parameter
; get pointer to overlay buffer not owned by interrupt into FSR1
	lfsr	FSR1,CMD_OVERLAY
; continue with common code for message 0x10 and 0x12
	bra	USB_RECV_10_12

; message 0x13 <0x00..0x04>: read 1/5 of overlay buffer
USB_RECV_13:
; check length
	movf	USB_CNT,W
	xorlw	0x02
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x04) of parameter
	movf	PREINC0,W
	sublw	0x04
	btfss	STATUS,C
	return				; invalid parameter
; get pointer to overlay buffer not owned by interrupt into FSR1
	lfsr	FSR1,CMD_OVERLAY
; continue with common code for message 0x11 and 0x13
	bra	USB_RECV_11_13

; common code for message 0x10 and 0x12
USB_RECV_10_12:
; get offset of first byte into PRODH:PRODL
	movlw	52
	mulwf	POSTINC0
; get address of first byte
	movf	PRODL,W
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; copy 52 bytes of data
	movlw	52
	movwf	A1_REG
USB_RECV_10_12_LOOP:
	movff	POSTINC0,POSTINC1
	decfsz	A1_REG
	bra	USB_RECV_10_12_LOOP
; send answer 0x10 <0x00..0x04>: 1/5 of pixel buffer loaded
;          or 0x12 <0x00..0x04>: 1/5 of overlay buffer loaded
	movlw	0x02
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; common code for message 0x11 and 0x13
USB_RECV_11_13:
; get offset of first byte into PRODH:PRODL
	movlw	52
	mulwf	POSTINC0
; get address of first byte
	movf	PRODL,W
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; first two bytes of USB buffer contain the correct bytes
; FSR0 points to third byte of USB buffer
; copy 52 bytes of data
	movlw	52
	movwf	A1_REG
USB_RECV_11_13_LOOP:
	movff	POSTINC1,POSTINC0
	decfsz	A1_REG
	bra	USB_RECV_11_13_LOOP
; send answer 0x11 <0x00..0x04> <52 bytes>: data read from 1/5 of pixel buffer
;          or 0x13 <0x00..0x04> <52 bytes>: data read from 1/5 of overlay buffer
	movlw	0x36
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x20: swap display buffers
USB_RECV_20:
; check length
	movf	USB_CNT,W
	xorlw	0x1
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; request output buffer switch
	bsf	BIT_O_BUF_SW
; wait for buffer switch to complete
USB_RECV_20_WAIT:
	btfsc	BIT_O_BUF_SW
	bra	USB_RECV_20_WAIT
; send answer 0x20: display buffers swapped
	movlw	0x01
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x21: swap display buffers and copy image to new background buffer
USB_RECV_21:
; check length
	movf	USB_CNT,W
	xorlw	0x1
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; request output buffer switch
	bsf	BIT_O_BUF_SW
; wait for buffer switch to complete
USB_RECV_21_WAIT:
	btfsc	BIT_O_BUF_SW
	bra	USB_RECV_21_WAIT
; get pointer to buffer owned by interrupt into FSR0
	lfsr	FSR0,O_FRAME0
	btfsc	BIT_O_INT_BUF		; skip if buffer 0 is owned by interrupt
	lfsr	FSR0,O_FRAME1
; get pointer to buffer not owned by interrupt into FSR1
	lfsr	FSR1,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR1,O_FRAME1
; copy buffer owned by interrupt into other one
	movlw	65			; copy 65 times 4 bytes
	movwf	A1_REG
USB_RECV_21_LOOP:
	movff	POSTINC0,POSTINC1	; copy 4 bytes
	movff	POSTINC0,POSTINC1
	movff	POSTINC0,POSTINC1
	movff	POSTINC0,POSTINC1
	decfsz	A1_REG			; loop
	bra	USB_RECV_21_LOOP
; send answer 0x21: display buffers swapped and image copied to new background buffer
	movlw	0x01
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x30 <0x00..0x0F> <32 bytes>: load 1/16 of sector buffer
USB_RECV_30:
; check length
	movf	USB_CNT,W
	xorlw	0x22
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x0F) of parameter
	movf	PREINC0,W
	sublw	0x0F
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; get pointer to sector buffer
	lfsr	FSR1,BUFFER
; get offset of first byte into PRODH:PRODL
	movlw	32
	mulwf	POSTINC0
; get address of first byte
	movf	PRODL,W
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; copy 32 bytes of data
	movlw	32
	movwf	A1_REG
USB_RECV_30_LOOP:
	movff	POSTINC0,POSTINC1
	decfsz	A1_REG
	bra	USB_RECV_30_LOOP
; send answer 0x30 <0x00..0x0F>: 1/16 of sector buffer set
	movlw	0x02
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x31 <0x00..0x0F>: read 1/16 of sector buffer
USB_RECV_31:
; check length
	movf	USB_CNT,W
	xorlw	0x02
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x0F) of parameter
	movf	PREINC0,W
	sublw	0x0F
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; get pointer to sector buffer
	lfsr	FSR1,BUFFER
; get offset of first byte into PRODH:PRODL
	movlw	32
	mulwf	POSTINC0
; get address of first byte
	movf	PRODL,W
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; first two bytes of USB buffer contain the correct bytes
; FSR0 points to third byte of USB buffer
; copy 32 bytes of data
	movlw	32
	movwf	A1_REG
USB_RECV_31_LOOP:
	movff	POSTINC1,POSTINC0
	decfsz	A1_REG
	bra	USB_RECV_31_LOOP
; send answer 0x31 <0x00..0x0F> <32 bytes>: data read from 1/16 of sector buffer
	movlw	0x22
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x40: get number of compact flash sectors
USB_RECV_40:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x40 <sector count (4 byte, big endian)>: number of compact flash sectors
	movlw	0x05
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x40
	movwf	POSTINC0
	movff	CF_SECTOR_CNT+3,POSTINC0
	movff	CF_SECTOR_CNT+2,POSTINC0
	movff	CF_SECTOR_CNT+1,POSTINC0
	movff	CF_SECTOR_CNT+0,POSTINC0
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x42 <sector number (4 byte, big endian)>: write compact flash sector
USB_RECV_42:
; check length
	movf	USB_CNT,W
	xorlw	0x05
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; if CF is busy, start to wait for CF
	btfss	BIT_CF_BUSY
	bra	USB_RECV_42_DO_IT
	movlw	USB_WAIT_CF_PRE_WRITE	; remember what we are waiting for
	movwf	USB_WAIT_CF
	return				; do nothing else
USB_RECV_42_DO_IT:
; write sector
	movff	PREINC0,A4_REG+3	; sector number
	movff	PREINC0,A4_REG+2
	movff	PREINC0,A4_REG+1
	movff	PREINC0,A4_REG+0
	lfsr	FSR0,BUFFER		; address of buffer (512 bytes)
	call	CF_WRITE_SECTOR
	btfsc	STATUS,C		; jump in case of error
	bra	USB_RECV_42_ERR
	movlw	USB_WAIT_CF_WRITE	; remember that we are waiting for the write to finish
	movwf	USB_WAIT_CF
	return				; do nothing else
USB_RECV_42_ERR:
	movlw	0x01			; report error
USB_RECV_42_DONE:
	movff	WREG,USB_BUF+5		; store error number in USB buffer
; send answer 0x42 <sector number (4 byte, big endian)> <error (1 byte)>: compact flash sector written (error = 0 for success)
; data in USB buffer is set already
	movlw	0x06
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x43 <sector number (4 byte, big endian)>: read compact flash sector
USB_RECV_43:
; check length
	movf	USB_CNT,W
	xorlw	0x05
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; if CF is busy, start to wait for CF
	btfss	BIT_CF_BUSY
	bra	USB_RECV_43_DO_IT
	movlw	USB_WAIT_CF_PRE_READ	; remember waht we are waiting for
	movwf	USB_WAIT_CF
	return				; do nothing else
USB_RECV_43_DO_IT:
; read sector
	movff	PREINC0,A4_REG+3	; sector number
	movff	PREINC0,A4_REG+2
	movff	PREINC0,A4_REG+1
	movff	PREINC0,A4_REG+0
	lfsr	FSR0,BUFFER		; address of buffer (512 bytes)
	call	CF_READ_SECTOR
	btfsc	STATUS,C		; jump in case of error
	bra	USB_RECV_43_ERR
	movlw	USB_WAIT_CF_READ	; remember that we are waiting for the read to finish
	movwf	USB_WAIT_CF
	return				; do nothing else
USB_RECV_43_ERR:
	movlw	0x01			; report error
USB_RECV_43_DONE:
	movff	WREG,USB_BUF+5		; store error number in USB buffer
; send answer 0x43 <sector number (4 byte, big endian)> <error (1 byte)>: compact flash sector read (error = 0 for success)
; data in USB buffer is set already
	movlw	0x06
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x44: get number of internal flash sectors
USB_RECV_44:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; get number of sectors in internal flash
	call	IF_SECTOR_COUNT		; result in A4_REG
; send answer 0x44 <sector count (4 byte, big endian)>: number of internal flash sectors
	movlw	0x05
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x44
	movwf	POSTINC0
	movff	A4_REG+3,POSTINC0
	movff	A4_REG+2,POSTINC0
	movff	A4_REG+1,POSTINC0
	movff	A4_REG+0,POSTINC0
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x47 <sector number (4 byte, big endian)>: read internal flash sector
USB_RECV_47:
; check length
	movf	USB_CNT,W
	xorlw	0x05
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; read sector
	movff	PREINC0,A4_REG+3	; sector number
	movff	PREINC0,A4_REG+2
	movff	PREINC0,A4_REG+1
	movff	PREINC0,A4_REG+0
	lfsr	FSR0,BUFFER		; address of buffer (512 bytes)
	call	IF_READ_SECTOR
	movlw	0x00			; default is no error
	btfsc	STATUS,C		; set error to 0x01 in case of error
	movlw	0x01
	movff	WREG,USB_BUF+5		; store error number in USB buffer
; send answer 0x47 <sector number (4 byte, big endian)> <error (1 byte)>: internal flash sector read (error = 0 for success)
; data in USB buffer is set already
	movlw	0x06
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x48: get compact flash flag (!= 0 if executing command from compact flash)
USB_RECV_48:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x48 <compact flash flag>: compact flash flag (!= 0 if executing commands from compact flash)
	movlw	0x02
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x48
	movwf	POSTINC0
	movlw	0x00			; 0x00 for internal flash
	btfsc	BIT_CMD_CF
	movlw	0x01			; 0x01 for external flash
	movwf	POSTINC0
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x50 <0x00..0x07> <32 bytes>: load 1/8 of variables
USB_RECV_50:
; check length
	movf	USB_CNT,W
	xorlw	0x22
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x07) of parameter
	movf	PREINC0,W
	sublw	0x07
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; get pointer to variables
	lfsr	FSR1,CMD_VARS
; get offset of first byte into PRODH:PRODL
	movlw	32
	mulwf	POSTINC0
; get address of first byte
	movf	PRODL,W
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; copy 32 bytes of data
	movlw	32
	movwf	A1_REG
USB_RECV_50_LOOP:
	movff	POSTINC0,POSTINC1
	decfsz	A1_REG
	bra	USB_RECV_50_LOOP
; send answer 0x50 <0x00..0x07>: 1/8 of variables set
	movlw	0x02
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x51 <0x00..0x07>: read 1/8 of variables
USB_RECV_51:
; check length
	movf	USB_CNT,W
	xorlw	0x02
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x07) of parameter
	movf	PREINC0,W
	sublw	0x07
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; get pointer to variables
	lfsr	FSR1,CMD_VARS
; get offset of first byte into PRODH:PRODL
	movlw	32
	mulwf	POSTINC0
; get address of first byte
	movf	PRODL,W
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; first two bytes of USB buffer contain the correct bytes
; FSR0 points to third byte of USB buffer
; copy 32 bytes of data
	movlw	32
	movwf	A1_REG
USB_RECV_51_LOOP:
	movff	POSTINC1,POSTINC0
	decfsz	A1_REG
	bra	USB_RECV_51_LOOP
; send answer 0x51 <0x00..0x07> <32 bytes>: data read from read 1/8 of variables
	movlw	0x22
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x54 <0x00..0x03> <32 bytes>: load 1/4 of stack
USB_RECV_54:
; check length
	movf	USB_CNT,W
	xorlw	0x22
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x03) of parameter
	movf	PREINC0,W
	sublw	0x03
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; get pointer to stack
	lfsr	FSR1,CMD_STACK
; get offset of first byte into PRODH:PRODL
	movlw	32
	mulwf	POSTINC0
; get address of first byte
	movf	PRODL,W
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; copy 32 bytes of data
	movlw	32
	movwf	A1_REG
USB_RECV_54_LOOP:
	movff	POSTINC0,POSTINC1
	decfsz	A1_REG
	bra	USB_RECV_54_LOOP
; send answer 0x54 <0x00..0x07>: 1/4 of stack set
	movlw	0x02
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x55 <0x00..0x03>: read 1/4 of stack
USB_RECV_55:
; check length
	movf	USB_CNT,W
	xorlw	0x02
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range (0x00..0x03) of parameter
	movf	PREINC0,W
	sublw	0x03
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; get pointer to stack
	lfsr	FSR1,CMD_STACK
; get offset of first byte into PRODH:PRODL
	movlw	32
	mulwf	POSTINC0
; get address of first byte
	movf	PRODL,W
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; first two bytes of USB buffer contain the correct bytes
; FSR0 points to third byte of USB buffer
; copy 32 bytes of data
	movlw	32
	movwf	A1_REG
USB_RECV_55_LOOP:
	movff	POSTINC1,POSTINC0
	decfsz	A1_REG
	bra	USB_RECV_55_LOOP
; send answer 0x51 <0x00..0x03> <32 bytes>: data read from read 1/4 of stack
	movlw	0x22
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x56 <stack pointer (1 byte)>: set stack pointer
USB_RECV_56:
; check length
	movf	USB_CNT,W
	xorlw	0x02
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; set stack pointer
	movff	PREINC0,CMD_STACK_PTR
	bcf	CMD_STACK_PTR,7		; keep stack pointer in rage 0..127
; send answer 0x56 <stack pointer (1 byte)>: stack pointer set
	movff	CMD_STACK_PTR,INDF0
	movlw	0x02
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x57: get stack pointer
USB_RECV_57:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x57 <stack pointer (1 byte)>: value of stack pointer
	movff	CMD_STACK_PTR,PREINC0
	movlw	0x02
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x60 <pressed> <pressed_mask> <released> <released_mask>: get key state and set key state afterwards (all fields 1 byte)
USB_RECV_60:
; check length
	movf	USB_CNT,W
	xorlw	0x05
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; get data for new key state
	movff	PREINC0,A1_REG		; pressed -> A1_REG
	movff	PREINC0,B1_REG		; pressed_mask -> A1_REG
	movff	PREINC0,C1_REG		; released -> A1_REG
	movff	PREINC0,D1_REG		; released_mask -> A1_REG
; prepare for writing key state into answer buffer
	lfsr	FSR0,USB_BUF
	movlw	0x60
	movwf	POSTINC0
; disable timer 4 interrupt (keys are read there)
	bcf	PIE3,3
; write key states into answer buffer
	movff	KEYS,POSTINC0
	movff	KEYS_PRESSED,POSTINC0
	movff	KEYS_RELEASED,POSTINC0
; write new key state
	comf	B1_REG,w		; KEYS_PRESSED &= ~pressed_mask
	andwf	KEYS_PRESSED
	movf	B1_REG,w		; KEYS_PRESSED | = pressed_mask & pressed
	andwf	A1_REG,w
	iorwf	KEYS_PRESSED
	comf	D1_REG,w		; KEYS_RELEASED &= ~released_mask
	andwf	KEYS_RELEASED
	movf	D1_REG,w		; KEYS_RELEASED | = released_mask & released
	andwf	C1_REG,w
	iorwf	KEYS_RELEASED
; reenable timer 4 interrupt
	bsf	PIE3,3
; send answer 0x60 <state> <pressed> <released>: key state (all fields 1 byte)
	movlw	0x04
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x62 <pressed> <pressed_mask> <released> <released_mask>: get SNES0 key state and set SNES0 key state afterwards (all fields 2 bytes)
USB_RECV_62:
; check length
	movf	USB_CNT,W
	xorlw	0x09
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; get data for new key state
	movff	PREINC0,A2_REG+1	; pressed -> A2_REG
	movff	PREINC0,A2_REG
	movff	PREINC0,B2_REG+1	; pressed_mask -> B2_REG
	movff	PREINC0,B2_REG
	movff	PREINC0,C2_REG+1	; released -> C2_REG
	movff	PREINC0,C2_REG
	movff	PREINC0,D2_REG+1	; released_mask -> D2_REG
	movff	PREINC0,D2_REG
; prepare for writing key state into answer buffer
	lfsr	FSR0,USB_BUF
	movlw	0x62
	movwf	POSTINC0
; disable timer 4 interrupt (keys are read there)
	bcf	PIE3,3
; write key states into answer buffer
	movff	SNES0+1,POSTINC0
	movff	SNES0,POSTINC0
	movff	SNES0_PRESSED+1,POSTINC0
	movff	SNES0_PRESSED,POSTINC0
	movff	SNES0_RELEASED+1,POSTINC0
	movff	SNES0_RELEASED,POSTINC0
; write new key state
	comf	B2_REG+1,w		; SNES0_PRESSED &= ~pressed_mask
	andwf	SNES0_PRESSED+1
	comf	B2_REG,w
	andwf	SNES0_PRESSED
	movf	B2_REG+1,w		; SNES0_PRESSED | = pressed_mask & pressed
	andwf	A2_REG+1,w
	iorwf	SNES0_PRESSED+1
	movf	B2_REG,w
	andwf	A2_REG,w
	iorwf	SNES0_PRESSED
	comf	D2_REG+1,w		; SNES0_RELEASED &= ~released_mask
	andwf	SNES0_RELEASED+1
	comf	D2_REG,w
	andwf	SNES0_RELEASED
	movf	D2_REG+1,w		; SNES0_RELEASED | = released_mask & released
	andwf	C2_REG+1,w
	iorwf	SNES0_RELEASED+1
	movf	D2_REG,w
	andwf	C2_REG,w
	iorwf	SNES0_RELEASED
; reenable timer 4 interrupt
	bsf	PIE3,3
; send answer 0x62 <state> <pressed> <released>: SNES0 key state (all fields 2 byte)
	movlw	0x07
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x63 <pressed> <pressed_mask> <released> <released_mask>: get SNES1 key state and set SNES1 key state afterwards (all fields 2 bytes)
USB_RECV_63:
; check length
	movf	USB_CNT,W
	xorlw	0x09
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; get data for new key state
	movff	PREINC0,A2_REG+1	; pressed -> A2_REG
	movff	PREINC0,A2_REG
	movff	PREINC0,B2_REG+1	; pressed_mask -> B2_REG
	movff	PREINC0,B2_REG
	movff	PREINC0,C2_REG+1	; released -> C2_REG
	movff	PREINC0,C2_REG
	movff	PREINC0,D2_REG+1	; released_mask -> D2_REG
	movff	PREINC0,D2_REG
; prepare for writing key state into answer buffer
	lfsr	FSR0,USB_BUF
	movlw	0x63
	movwf	POSTINC0
; disable timer 4 interrupt (keys are read there)
	bcf	PIE3,3
; write key states into answer buffer
	movff	SNES1+1,POSTINC0
	movff	SNES1,POSTINC0
	movff	SNES1_PRESSED+1,POSTINC0
	movff	SNES1_PRESSED,POSTINC0
	movff	SNES1_RELEASED+1,POSTINC0
	movff	SNES1_RELEASED,POSTINC0
; write new key state
	comf	B2_REG+1,w		; SNES1_PRESSED &= ~pressed_mask
	andwf	SNES1_PRESSED+1
	comf	B2_REG,w
	andwf	SNES1_PRESSED
	movf	B2_REG+1,w		; SNES1_PRESSED | = pressed_mask & pressed
	andwf	A2_REG+1,w
	iorwf	SNES1_PRESSED+1
	movf	B2_REG,w
	andwf	A2_REG,w
	iorwf	SNES1_PRESSED
	comf	D2_REG+1,w		; SNES1_RELEASED &= ~released_mask
	andwf	SNES1_RELEASED+1
	comf	D2_REG,w
	andwf	SNES1_RELEASED
	movf	D2_REG+1,w		; SNES1_RELEASED | = released_mask & released
	andwf	C2_REG+1,w
	iorwf	SNES1_RELEASED+1
	movf	D2_REG,w
	andwf	C2_REG,w
	iorwf	SNES1_RELEASED
; reenable timer 4 interrupt
	bsf	PIE3,3
; send answer 0x63 <state> <pressed> <released>: SNES1 key state (all fields 2 byte)
	movlw	0x07
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x70 <sec> <min> <hour>: set time (all fields 1 byte)
USB_RECV_70:
; check length
	movf	USB_CNT,W
	xorlw	0x04
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; check range of parameters
	movf	PREINC0,W		; seconds
	sublw	59
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
	movf	PREINC0,W		; minutes
	sublw	59
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
	movf	PREINC0,W		; hours
	sublw	23
	btfss	STATUS,C
	bra	USB_RECV_GO_ON		; invalid parameter
; disable timer 4 interrupt (time is advanced there)
	bcf	PIE3,3
; write time into time varibales
	movff	POSTDEC0,TIME_HOUR
	movff	POSTDEC0,TIME_MIN
	movff	POSTDEC0,TIME_SEC
; reenable timer 4 interrupt
	bsf	PIE3,3
; send answer 0x70: time set
	movlw	0x01
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x71: get time (all fields 1 byte)
USB_RECV_71:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; disable timer 4 interrupt (time is advanced there)
	bcf	PIE3,3
; read time from time variables
	movff	TIME_SEC,PREINC0
	movff	TIME_MIN,PREINC0
	movff	TIME_HOUR,PREINC0
; reenable timer 4 interrupt
	bsf	PIE3,3
; send answer 0x71 <sec> <min> <hour>: time (all fields 1 byte)
	movlw	0x04
	movwf	USB_CNT
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x80: get version information
USB_RECV_80:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x80 <major> <minor> <revison> <year> <month> <day>: version information
	movlw	0x07
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x80
	movwf	POSTINC0
	movlw	VER_MAJ			; version
	movwf	POSTINC0
	movlw	VER_MIN
	movwf	POSTINC0
	movlw	VER_REV
	movwf	POSTINC0
	movlw	VER_YEAR		; date
	movwf	POSTINC0
	movlw	VER_MONTH
	movwf	POSTINC0
	movlw	VER_DAY
	movwf	POSTINC0
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x84: get ID
USB_RECV_84:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x84 <ID 0> ... <ID 15> <error>: ID read
	movlw	0x12
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x84
	movwf	POSTINC0
	call	EE_GET_ID		; ID
	movwf	POSTINC0		; error code
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; 0x85 <ID 0> ... <ID 15>: set ID
USB_RECV_85:
; check length
	movf	USB_CNT,W
	xorlw	0x11
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; set ID
	movf	POSTINC0,W
	call	EE_SET_ID
	movwf	A1_REG			; remember error code
; send answer 0x85 <error>: ID set (error = 0 for success)
	movlw	0x02
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x85
	movwf	POSTINC0
	movff	A1_REG,POSTINC0		; error code
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x88: get output style
USB_RECV_88:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x88 <output style> <error>: output style read (0x00=LED, 0x01=LATCH) (error = 0 for success)
	movlw	0x03
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x88
	movwf	POSTINC0
	call	EE_GET_OUT_STYLE	; output style
	movff	A1_REG,POSTINC0
	movwf	POSTINC0		; error code
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x89 <output style>: set output style (0x00=LED, 0x01=LATCH)
USB_RECV_89:
; check length
	movf	USB_CNT,W
	xorlw	0x02
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; set output style
	movf	POSTINC0,W
	movff	POSTINC0,A1_REG
	call	EE_SET_OUT_STYLE
	movwf	A1_REG			; remember error code
; send answer 0x89 <error>: output style set (error = 0 for success)
	movlw	0x02
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x89
	movwf	POSTINC0
	movff	A1_REG,POSTINC0		; error code
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x8A: get output timing
USB_RECV_8A:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x8A <timing 1> ... <timing 15> <error>: output timing read (error = 0 for success)
	movlw	0x11
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x8A
	movwf	POSTINC0
	call	EE_GET_OUT_TIMING	; output timing
	movwf	POSTINC0		; error code
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x8B <timing 1> ... <timing 15>: set output timing (0..50)
USB_RECV_8B:
; check length
	movf	USB_CNT,W
	xorlw	0x10
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; set output timing
	movf	POSTINC0,W
	call	EE_SET_OUT_TIMING
	movwf	A1_REG			; remember error code
; send answer 0x8B <error>: output timing set (error = 0 for success)
	movlw	0x02
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x8B
	movwf	POSTINC0
	movff	A1_REG,POSTINC0		; error code
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x8C: get output greyscale mapping
USB_RECV_8C:
; check length
	movf	USB_CNT,W
	xorlw	0x01
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; send answer 0x8C <mapping 1> ... <mapping 15> <error>: output greyscale mapping read (error = 0 for success)
	movlw	0x11
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x8C
	movwf	POSTINC0
	call	EE_GET_OUT_GREYMAP	; output greyscale mapping
	movwf	POSTINC0		; error code
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; message 0x8D <mapping 1> ... <mapping 15>: set output greyscale mapping (0..15)
USB_RECV_8D:
; check length
	movf	USB_CNT,W
	xorlw	0x10
	btfss	STATUS,Z
	bra	USB_RECV_GO_ON
; set output greyscale mapping
	movf	POSTINC0,W
	call	EE_SET_OUT_GREYMAP
	movwf	A1_REG			; remember error code
; send answer 0x8D <error>: output greyscale mapping set (error = 0 for success)
	movlw	0x02
	movwf	USB_CNT
	lfsr	FSR0,USB_BUF
	movlw	0x8D
	movwf	POSTINC0
	movff	A1_REG,POSTINC0		; error code
	bsf	USB_S1_REQ
; message processed
	bra	USB_RECV_GO_ON

; reenable reception
USB_RECV_GO_ON:
	movlw	b'00001100' 		; RXC1: RFWL=0, FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC1
	call	USB_WRITE_REG
	movlw	b'00000101' 		; RXC1: RFWL=0, FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC1
	call	USB_WRITE_REG
USB_RECV_END:

; send packets on data endpoint 1
	btfss	USB_S1_ACT		; jump if not currently sending
	bra	USB_SEND_NEW		
	movlw	USB_TXS1		; get transmit state from TXS1
	call	USB_READ_REG
	btfss	A1_REG,5		; jump if TX_DONE not set
	bra	USB_SEND_END
	btfsc	A1_REG,6		; jump if ACK_STAT set
	bra	USB_SEND_ACK
; transmission failed (no ACK), repeat it
	movlw	b'00010000'	 	; TXC1: RFF=1
	movwf	A1_REG
	movlw	USB_TXC1 
	call	USB_WRITE_REG
	movlw	b'00000011'		; TXC1: LAST=1, TX_EN=1, TOGGLE=?
	btfsc	USB_S1_TOGGLE
	movlw	b'00000111'
	movwf	A1_REG
	movlw	USB_TXC1	
	call	USB_WRITE_REG
	bra	USB_SEND_END
; transmission succeeded
USB_SEND_ACK:
	bcf	USB_S1_ACT		; no more sending
	btg	USB_S1_TOGGLE		; invert toggle bit
; start a new transmission
USB_SEND_NEW:
	btfss	USB_S1_REQ		; jump if nothing to send
	bra	USB_SEND_END
	bcf	USB_S1_REQ		; clear request bit
	movlw	b'00001000'		; TXC1: FLUSH=1
	movwf	A1_REG
	movlw	USB_TXC1
	call	USB_WRITE_REG
	lfsr	FSR0,USB_BUF		; load data into FIFO (TXD1)
USB_SEND_NEW_LOOP:
	movff	POSTINC0,A1_REG
	movlw	USB_TXD1
	call	USB_WRITE_REG
	decfsz	USB_CNT
	bra	USB_SEND_NEW_LOOP	
	movlw	b'00000011'		; TXC1: LAST=1, TX_EN=1, TOGGLE=?
	btfsc	USB_S1_TOGGLE
	movlw	b'00000111'
	movwf	A1_REG
	movlw	USB_TXC1	
	call	USB_WRITE_REG
	bsf	USB_S1_ACT		; now sending
USB_SEND_END:

; done
	return

;-------------------------------------------------------------------------------------------
; turn on usb data channels
;-------------------------------------------------------------------------------------------
USB_CONF_ON:
	movlw	b'00001000'		; TXC1: IGN_ISOMSK=0, TFWL=0, RRF=0, FLUSH=1, TOGGLE=0, LAST=0, TX_EN=0
	movwf	A1_REG
	movlw	USB_TXC1
	call	USB_WRITE_REG
	movlw	b'00010001'		; EPC1: STALL=0, ISO=0, EP_EN=1, ADDR=1
	movwf	A1_REG
	movlw	USB_EPC1
	call	USB_WRITE_REG
	movlw	b'00001100'		; RXC1: RFWL=0, IGN_ISO=0, FLUSH=1, IGN_SETUP=1, RX_EN=0
	movwf	A1_REG
	movlw	USB_RXC1
	call	USB_WRITE_REG
	movlw	b'00000101'		; RXC1: RFWL=0, IGN_ISO=0, FLUSH=0, IGN_SETUP=1, RX_EN=1
	movwf	A1_REG
	movlw	USB_RXC1
	call	USB_WRITE_REG
	movlw	b'00010001'		; EPC2: STALL=0, ISO=0, EP_EN=1, ADDR=1
	movwf	A1_REG
	movlw	USB_EPC2
	call	USB_WRITE_REG
	bcf	BIT_USB_CMD_STOP	; do not stop command execution
	bcf	BIT_USB_CMD_ST_N	; do not stop command execution after next command
	movlw	USB_WAIT_CF_NOT		; not wainting for CF
	movwf	USB_WAIT_CF
	return

;-------------------------------------------------------------------------------------------
; turn off usb data channels
;-------------------------------------------------------------------------------------------
USB_CONF_OFF:
	movlw	b'00000000'		; EPC1: STALL=0, ISO=0, EP_EN=0, ADDR=0
	movwf	A1_REG
	movlw	USB_EPC1
	call	USB_WRITE_REG
	movlw	b'00000000'		; EPC2: STALL=0, ISO=0, EP_EN=0, ADDR=0
	movwf	A1_REG
	movlw	USB_EPC2
	call	USB_WRITE_REG
	bcf	BIT_USB_CMD_STOP	; do not stop command execution
	bcf	BIT_USB_CMD_ST_N	; do not stop command execution after next command
	movlw	USB_WAIT_CF_NOT		; not wainting for CF
	movwf	USB_WAIT_CF
	return

;-------------------------------------------------------------------------------------------
; standard usb answers
;-------------------------------------------------------------------------------------------
USB_ANSWER_EMPTY:
	DB	low (USB_ANSWER_EMPTY_END - USB_ANSWER_EMPTY_START)
USB_ANSWER_EMPTY_START:
USB_ANSWER_EMPTY_END:

USB_ANSWER_00:
	DB	low (USB_ANSWER_00_END - USB_ANSWER_00_START)
USB_ANSWER_00_START:
	DB	0x00
USB_ANSWER_00_END:

USB_ANSWER_01:
	DB	low (USB_ANSWER_01_END - USB_ANSWER_01_START)
USB_ANSWER_01_START:
	DB	0x01
USB_ANSWER_01_END:


USB_ANSWER_00_00:
	DB	low (USB_ANSWER_00_00_END - USB_ANSWER_00_00_START)
USB_ANSWER_00_00_START:
	DB	0x00,0x00
USB_ANSWER_00_00_END:

;-------------------------------------------------------------------------------------------
; usb descriptors
;-------------------------------------------------------------------------------------------

USB_DESCR_DEV:
	DB	low (USB_DESCR_DEV_END - USB_DESCR_DEV_START)
USB_DESCR_DEV_START:
	DB	0x12,0x01	; bLength: size of descriptor, bType: device descriptor
	DB	0x10,0x01	; bcdUSB: USB 1.1.0
	DB	0x00,0x00	; bDeviceClass: no class, bDeviceSubClass: no subclass
	DB	0x00,0x08	; bDeviceProtocol: no protocol, bMaxPacketSize: 8 bytes per packet
	DB	0x66,0x66	; idVendor: 0x6666 prototype vendor id
	DB	0x6D,0x41	; idProduct: 0x416D 'Am' for ARCADEmini
	DB	0x01,0x00	; bcdDevice: 0x0001 1st revision of USB interface
	DB	0x01,0x02	; iManufacturer: index of manufacturer string, iProduct: index of product string
	DB	0x00,0x01	; iSerial: no serial number string,
USB_DESCR_DEV_END:		; bNumConfigurations: we only have one possible configuration


USB_DESCR_CONF:
	DB	low (USB_DESCR_CONF_END - USB_DESCR_CONF_START)
USB_DESCR_CONF_START:
; ConfigurationDescriptor
	DB	0x09,0x02	; bLength: size of descriptor
				; bType: configuration descriptor
	DB	0x20,0x00	; wTotalLength: total size of entire configuration
	DB	0x01,0x01	; bNumInterfaces: one interface
				; bConfigurationValue: configuration identifier (our configuration is called 0x01)
	DB	0x00,0x80	; iConfiguration: no configuration string
				; bmAttributes: not self powered, no remote wakeup
	DB	0xFA,0x09	; bMaxPower: power consumption max. 500mA
; InterfaceDescriptor		; bLength: size of descriptor
	DB	0x04,0x00	; bType: interface descriptor
				; bInterfaceNumber: this is the first interface
	DB	0x00,0x02	; bAlternateSetting: no alternate settings
				; bNumEndpoints: 2 endpoints (without endpoint 0)
	DB	0xFF,0xFF	; bInterfaceClass: we don't have a registered class code
				; bInterfaceSubClass: we don't have a registered subclass code
	DB	0xFF,0x00	; bInterfaceProtocol: we don't have a registered protocol code
				; iInterface: no interface string
; EndpointDescriptor1
	DB	0x07,0x05	; bLength: size of descriptor
				; bType: endpoint descriptor
	DB	0x01,0x02	; bEndpointAddress: endpoint 1, out
				; bmAttributes: bulk
	DB	0x40,0x00	; wMaxPacketSize: max. packet size 64
				; Higbyte max. packet size
	DB	0x00,0x07	; bInterval: ignored for bulk
; EndpointDescriptor2		; bLength: size of descriptor
	DB	0x05,0x81	; bType: endpoint descriptor
				; bEndpointAddress: endpoint 1, in
	DB	0x02,0x40	; bmAttributes: bulk
				; wMaxPacketSize: max. packet size 64
	DB	0x00,0x00	; Higbyte max. packet size
				; bInterval: ignored for bulk
USB_DESCR_CONF_END:

USB_DESCR_STR0:
	DB	low (USB_DESCR_STR0_END - USB_DESCR_STR0_START)
USB_DESCR_STR0_START:
	DB	0x04,0x03	; bLength: size of descriptor
				; bType: string descriptor
	DB	0x09,0x04	; wLangId: US English
USB_DESCR_STR0_END:

USB_DESCR_STR1:
	DB	low (USB_DESCR_STR1_END - USB_DESCR_STR1_START)
USB_DESCR_STR1_START:
	DB	0x36,0x03	; bLength: size of descriptor
				; bType: string descriptor
	DW	'S'
	DW	't'
	DW	'e'
	DW	'f'
	DW	'a'
	DW	'n'
	DW	' '
	DW	'S'
	DW	'c'
	DW	'h'
	DW	'u'
	DW	'e'
	DW	'r'
	DW	'm'
	DW	'a'
	DW	'n'
	DW	's'
	DW	' '
	DW	'S'
	DW	'o'
	DW	'f'
	DW	't'
	DW	'w'
	DW	'a'
	DW	'r'
	DW	'e'
USB_DESCR_STR1_END:

USB_DESCR_STR2:
	DB	low (USB_DESCR_STR2_END - USB_DESCR_STR2_START)
USB_DESCR_STR2_START:
	DB	0x1E,0x03	; bLength: size of descriptor
				; bType: string descriptor
	DW	'A'
	DW	'r'
	DW	'c'
	DW	'a'
	DW	'd'
	DW	'e'
	DW	'M'
	DW	'i'
	DW	'n'
	DW	'i'
	DW	' '
	IF VER_MAJ < 10
		DW	'0' + VER_MAJ
	ELSE
		DW	'0' + (VER_MAJ / 10)
		DW	'0' + (VER_MAJ % 10)
	ENDIF
	DW	'.'
	IF VER_MIN < 10
		DW	'0' + VER_MIN
	ELSE
		DW	'0' + (VER_MIN / 10)
		DW	'0' + (VER_MIN % 10)
	ENDIF
	DW	'.'
	IF VER_REV < 10
		DW	'0' + VER_REV
	ELSE
		DW	'0' + (VER_REV / 10)
		DW	'0' + (VER_REV % 10)
	ENDIF
USB_DESCR_STR2_END:



;===========================================================================================
; global variables
;===========================================================================================



;===========================================================================================
; global procedures
;===========================================================================================
	global		USB_INIT, USB_TASK



	end
