; ARCADEmini firmware
; part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
; Copyright (C) 2003-2006 Stefan Schuermans <1stein@schuermans.info>
; may be used for non-commercial purposes in unmodified form if:
; may be used for non-commercial purposes in modified form if:
;   - this copyright header is left unchanged
;   - and every modification is sent to the copyright holder in source code form
;   - and the source code of the modification may be used in every way be the copyright holder

	TITLE		"commands for ARCADEmini firmware"

	include		"P18F8520.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0, BITS1
	extern		A1_REG, B1_REG
	extern		A2_REG, B2_REG, C2_REG, D2_REG, E2_REG
	extern		A4_REG, B4_REG, C4_REG
	extern		DELAY
	extern		TIME_SEC100, TIME_SEC, TIME_MIN, TIME_HOUR
	extern		KEYS, KEYS_PRESSED, KEYS_RELEASED
	extern		SNES0, SNES1, SNES0_PRESSED, SNES1_PRESSED, SNES0_RELEASED, SNES1_RELEASED
	extern		BUFFER, BUFFER_OV
	extern		O_FRAME0, O_FRAME1
	extern		RANDOM



;===========================================================================================
; imported procedures
;===========================================================================================
	extern		CF_READ_SECTOR, CF_READ_SECTOR_PART
	extern		IF_READ_SECTOR



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

; command pointer
; current sector
CMD_SECTOR	res	4
; current offset
CMD_OFFSET	res	2

; CF wait mode
CMD_CF_WAIT	res	1
CMD_CF_WAIT_NOT		equ	0x00	; not waiting for CF
CMD_CF_WAIT_WANT_SEC	equ	0x01	; waiting for CF because wanting to load a sector
CMD_CF_WAIT_LOAD_SEC	equ	0x02	; waiting for CF to finish loading a sector
CMD_CF_WAIT_RD_CF	equ	0x03	; waiting for CF to finish loading a part of a sector for RD_CF command

; the stack pointer
; - stack is CMD_STACK
; - increments on push / decrements on pop
; - always contains the offset of the next value to write
; - always contains a value of 0..127
CMD_STACK_PTR	res	1



;===========================================================================================
ram_cmd		UDATA			; normal RAM for commands
; (IMPORTANT: must begin at a page boundary (address 0x??00))
;===========================================================================================

; variables for commands
; (IMPORTANT: must begin at a page boundary (address 0x??00))
CMD_VARS	res	256

; overlay buffer (one nibble for each pixel)
CMD_OVERLAY	res	260

; line buffer for CP_BLK[_TR[_C]]
CMD_LINE	res	26

; the stack
; - grows towards higher addresses
; - wraps around on overflow / underflow
; - stack pointer is CMD_STACK_PTR
CMD_STACK	res	128



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; commands initialization
;-------------------------------------------------------------------------------------------
CMD_INIT:
; clear command pointer
	clrf	CMD_SECTOR+0
	clrf	CMD_SECTOR+1
	clrf	CMD_SECTOR+2
	clrf	CMD_SECTOR+3
	clrf	CMD_OFFSET+0
	clrf	CMD_OFFSET+1
; not waiting for CF
	movlw	CMD_CF_WAIT_NOT
	movwf	CMD_CF_WAIT
; clear stack pointer
	clrf	CMD_STACK_PTR
; done
	return

;-------------------------------------------------------------------------------------------
; process a command
; input:	-
; output:	-
; changes:	PRODH, PRODL, PCLATU, PCLATH, A1_REG, WREG
;		(plus changes done by command:
;		 e.g. FSR0, A2_REG, A4_REG, BUFFER)
;-------------------------------------------------------------------------------------------
CMD_EXEC:
; load address of command buffer plus offset part of command pointer into FSR1
	movlw	low BUFFER
	addwf	CMD_OFFSET+0,W
	movwf	FSR1L
	movlw	high BUFFER
	addwfc	CMD_OFFSET+1,W
	movwf	FSR1H

; check if waiting for CF
	movf	CMD_CF_WAIT,W
	xorlw	CMD_CF_WAIT_NOT
	bz	CMD_EXEC_CF_NOWAIT
; end waiting for CF if not executing commands from CF
	btfsc	BIT_CMD_CF
	bra	CMD_EXEC_CF_WAIT_NOINTFLASH
	movlw	CMD_CF_WAIT_NOT
	movwf	CMD_CF_WAIT
	bra	CMD_EXEC_CF_NOWAIT
CMD_EXEC_CF_WAIT_NOINTFLASH:
; exit if compact flash is busy
	btfsc	BIT_CF_BUSY
	return

; waiting for CF because wanting to load a sector
	movlw	CMD_CF_WAIT_WANT_SEC
	cpfseq	CMD_CF_WAIT
	bra	CMD_EXEC_CF_WAIT_WANT_SEC_END
; no more waiting for CF
	movlw	CMD_CF_WAIT_NOT
	movwf	CMD_CF_WAIT
; try to load the sector now
	goto	CMD_LOAD_SECTOR
CMD_EXEC_CF_WAIT_WANT_SEC_END:

; waiting for CF to finish loading a sector
	movlw	CMD_CF_WAIT_LOAD_SEC
	cpfseq	CMD_CF_WAIT
	bra	CMD_EXEC_CF_WAIT_LOAD_SEC_END
; no more waiting for CF
	movlw	CMD_CF_WAIT_NOT
	movwf	CMD_CF_WAIT
; switch back to internal flash if error compact flash error bit is set
	btfss	BIT_CF_ERROR
	bra	CMD_EXEC_CF_WAIT_LOAD_SEC_END
	bcf	BIT_CMD_CF		; signal CF error to caller
	return
CMD_EXEC_CF_WAIT_LOAD_SEC_END:

; waiting for CF to finish loading a part of a sector for RD_CF command
	movlw	CMD_CF_WAIT_RD_CF
	cpfseq	CMD_CF_WAIT
	bra	CMD_EXEC_CF_WAIT_RD_CF_END
; no more waiting for CF
	movlw	CMD_CF_WAIT_NOT
	movwf	CMD_CF_WAIT
; put error code into A1_REG
	movlw	0		; success
	btfsc	BIT_CF_ERROR
	movlw	2		; error code for CF error
	movwf	A1_REG
; finish RD_CF command now
	goto	CMD__RD_CF_END
CMD_EXEC_CF_WAIT_RD_CF_END:

CMD_EXEC_CF_NOWAIT:

; exit if command execution is stopped by usb
	btfsc	BIT_USB_CMD_STOP	
	return

; exit if waiting for delay
	btfsc	BIT_DELAY_WAIT
	return

; jump to command procedure
	movlw	4			; read command and get index into command table
	mulwf	POSTINC1		; PRODH:PRODL := POSTINC1 * 4
	movlw	low CMD_TAB		; add address of CMD_TAB to PRODH:PRODL
	addwf	PRODL,W			; and store result in PCLATU:PCLATH:A1_REG
	movwf	A1_REG
	movlw	high CMD_TAB
	addwfc	PRODH,W
	movwf	PCLATH
	clrf	PCLATU
	movlw	upper CMD_TAB
	addwfc	PCLATU
	movf	A1_REG,W		; jump to PCLATU:PCLATH:A1_REG
	movwf	PCL

;-------------------------------------------------------------------------------------------
; command table
; - must contain 256 gotos to the command functions
; - some commands use the last bit(s) of the command number as paramater
;   so these commands use multiple command number
;-------------------------------------------------------------------------------------------
CMD_TAB:
	goto	CMD__JMP_NEXT_SEC	; 0x00
	goto	CMD__JMP_SEC		; 0x01
	goto	CMD__JMP_OFS		; 0x02 (bit0 is param)
	goto	CMD__JMP_OFS		; 0x03 (bit0 is param)
	goto	CMD__JMP_OFS_SEC	; 0x04 (bit0 is param)
	goto	CMD__JMP_OFS_SEC	; 0x05 (bit0 is param)
	goto	CMD__JMP_IND		; 0x06
	goto	CMD__NOP		; 0x07
	goto	CMD__JMP_EQ		; 0x09 (bit0 is param)
	goto	CMD__JMP_EQ		; 0x00 (bit0 is param)
	goto	CMD__JMP_EQ_C		; 0x0A (bit0 is param)
	goto	CMD__JMP_EQ_C		; 0x0B (bit0 is param)
	goto	CMD__JMP_NEQ		; 0x0C (bit0 is param)
	goto	CMD__JMP_NEQ		; 0x0D (bit0 is param)
	goto	CMD__JMP_NEQ_C		; 0x0E (bit0 is param)
	goto	CMD__JMP_NEQ_C		; 0x0F (bit0 is param)
	goto	CMD__JMP_GT		; 0x10 (bit0 is param)
	goto	CMD__JMP_GT		; 0x11 (bit0 is param)
	goto	CMD__JMP_GT_C		; 0x12 (bit0 is param)
	goto	CMD__JMP_GT_C		; 0x13 (bit0 is param)
	goto	CMD__JMP_NGT		; 0x14 (bit0 is param)
	goto	CMD__JMP_NGT		; 0x15 (bit0 is param)
	goto	CMD__JMP_NGT_C		; 0x16 (bit0 is param)
	goto	CMD__JMP_NGT_C		; 0x17 (bit0 is param)
	goto	CMD__JMP_LT		; 0x18 (bit0 is param)
	goto	CMD__JMP_LT		; 0x19 (bit0 is param)
	goto	CMD__JMP_LT_C		; 0x1A (bit0 is param)
	goto	CMD__JMP_LT_C		; 0x1B (bit0 is param)
	goto	CMD__JMP_NLT		; 0x1C (bit0 is param)
	goto	CMD__JMP_NLT		; 0x1D (bit0 is param)
	goto	CMD__JMP_NLT_C		; 0x1E (bit0 is param)
	goto	CMD__JMP_NLT_C		; 0x1F (bit0 is param)
	goto	CMD__JMP_GTS		; 0x20 (bit0 is param)
	goto	CMD__JMP_GTS		; 0x21 (bit0 is param)
	goto	CMD__JMP_GTS_C		; 0x22 (bit0 is param)
	goto	CMD__JMP_GTS_C		; 0x23 (bit0 is param)
	goto	CMD__JMP_NGTS		; 0x24 (bit0 is param)
	goto	CMD__JMP_NGTS		; 0x25 (bit0 is param)
	goto	CMD__JMP_NGTS_C		; 0x26 (bit0 is param)
	goto	CMD__JMP_NGTS_C		; 0x27 (bit0 is param)
	goto	CMD__JMP_LTS		; 0x28 (bit0 is param)
	goto	CMD__JMP_LTS		; 0x29 (bit0 is param)
	goto	CMD__JMP_LTS_C		; 0x2A (bit0 is param)
	goto	CMD__JMP_LTS_C		; 0x2B (bit0 is param)
	goto	CMD__JMP_NLTS		; 0x2C (bit0 is param)
	goto	CMD__JMP_NLTS		; 0x2D (bit0 is param)
	goto	CMD__JMP_NLTS_C		; 0x2E (bit0 is param)
	goto	CMD__JMP_NLTS_C		; 0x2F (bit0 is param)
	goto	CMD__JMP_AND_Z		; 0x30 (bit0 is param)
	goto	CMD__JMP_AND_Z		; 0x31 (bit0 is param)
	goto	CMD__JMP_AND_Z_C	; 0x32 (bit0 is param)
	goto	CMD__JMP_AND_Z_C	; 0x33 (bit0 is param)
	goto	CMD__JMP_AND_NZ		; 0x34 (bit0 is param)
	goto	CMD__JMP_AND_NZ		; 0x35 (bit0 is param)
	goto	CMD__JMP_AND_NZ_C	; 0x36 (bit0 is param)
	goto	CMD__JMP_AND_NZ_C	; 0x37 (bit0 is param)
	goto	CMD__JMP_OR_M		; 0x38 (bit0 is param)
	goto	CMD__JMP_OR_M		; 0x39 (bit0 is param)
	goto	CMD__JMP_OR_M_C		; 0x3A (bit0 is param)
	goto	CMD__JMP_OR_M_C		; 0x3B (bit0 is param)
	goto	CMD__JMP_OR_NM		; 0x3C (bit0 is param)
	goto	CMD__JMP_OR_NM		; 0x3D (bit0 is param)
	goto	CMD__JMP_OR_NM_C	; 0x3E (bit0 is param)
	goto	CMD__JMP_OR_NM_C	; 0x3F (bit0 is param)
	goto	CMD__NOP		; 0x40
	goto	CMD__NOP		; 0x41
	goto	CMD__NOP		; 0x42
	goto	CMD__NOP		; 0x43
	goto	CMD__NOP		; 0x44
	goto	CMD__NOP		; 0x45
	goto	CMD__NOP		; 0x46
	goto	CMD__NOP		; 0x47
	goto	CMD__NOP		; 0x48
	goto	CMD__NOP		; 0x49
	goto	CMD__NOP		; 0x4A
	goto	CMD__NOP		; 0x4B
	goto	CMD__NOP		; 0x4C
	goto	CMD__NOP		; 0x4D
	goto	CMD__NOP		; 0x4E
	goto	CMD__NOP		; 0x4F
	goto	CMD__MOV		; 0x50
	goto	CMD__MOV_C		; 0x51
	goto	CMD__MOV2		; 0x52
	goto	CMD__MOV2_C		; 0x53
	goto	CMD__MOV4		; 0x54
	goto	CMD__MOV4_C		; 0x55
	goto	CMD__NOP		; 0x56
	goto	CMD__NOP		; 0x57
	goto	CMD__RD_I		; 0x58
	goto	CMD__WR_I		; 0x59
	goto	CMD__NOP		; 0x5A
	goto	CMD__NOP		; 0x5B
	goto	CMD__RD_S		; 0x5C
	goto	CMD__RD_CF		; 0x5D
	goto	CMD__NOP		; 0x5E
	goto	CMD__NOP		; 0x5F
	goto	CMD__ADD		; 0x60
	goto	CMD__ADD_C		; 0x61
	goto	CMD__ADDC		; 0x62
	goto	CMD__ADDC_C		; 0x63
	goto	CMD__SUB		; 0x64
	goto	CMD__SUB_C		; 0x65
	goto	CMD__SUBB		; 0x66
	goto	CMD__SUBB_C		; 0x67
	goto	CMD__MUL		; 0x68
	goto	CMD__MUL_C		; 0x69
	goto	CMD__MULS		; 0x6A
	goto	CMD__MULS_C		; 0x6B
	goto	CMD__DIV		; 0x6C
	goto	CMD__DIVS		; 0x6D
	goto	CMD__NOP		; 0x6E
	goto	CMD__NOP		; 0x6F
	goto	CMD__AND		; 0x70
	goto	CMD__AND_C		; 0x71
	goto	CMD__OR			; 0x72
	goto	CMD__OR_C		; 0x73
	goto	CMD__XOR		; 0x74
	goto	CMD__XOR_C		; 0x75
	goto	CMD__NOP		; 0x76
	goto	CMD__NOP		; 0x77
	goto	CMD__SHL		; 0x78
	goto	CMD__SHL_C		; 0x79
	goto	CMD__SHR		; 0x7A
	goto	CMD__SHR_C		; 0x7B
	goto	CMD__ROL		; 0x7C
	goto	CMD__ROL_C		; 0x7D
	goto	CMD__ROR		; 0x7E
	goto	CMD__ROR_C		; 0x7F
	goto	CMD__PUSH		; 0x80
	goto	CMD__PUSH_C		; 0x81
	goto	CMD__POP		; 0x82
	goto	CMD__NOP		; 0x83
	goto	CMD__CALL		; 0x84 (bit0 is param)
	goto	CMD__CALL		; 0x85 (bit0 is param)
	goto	CMD__CALL_IND		; 0x86
	goto	CMD__RET		; 0x87
	goto	CMD__NOP		; 0x88
	goto	CMD__NOP		; 0x89
	goto	CMD__NOP		; 0x8A
	goto	CMD__NOP		; 0x8B
	goto	CMD__NOP		; 0x8C
	goto	CMD__NOP		; 0x8D
	goto	CMD__NOP		; 0x8E
	goto	CMD__NOP		; 0x8F
	goto	CMD__RD_PIX		; 0x90
	goto	CMD__RD_PIX_C		; 0x91
	goto	CMD__RD_OVP		; 0x92
	goto	CMD__RD_OVP_C		; 0x93
	goto	CMD__WR_PIX		; 0x94
	goto	CMD__WR_PIX_C		; 0x95
	goto	CMD__WR_OVP		; 0x96
	goto	CMD__WR_OVP_C		; 0x97
	goto	CMD__WR_PIX_CO		; 0x98
	goto	CMD__WR_PIX_CO_C	; 0x99
	goto	CMD__WR_OVP_CO		; 0x9A
	goto	CMD__WR_OVP_CO_C	; 0x9B
	goto	CMD__NOP		; 0x9C
	goto	CMD__NOP		; 0x9D
	goto	CMD__NOP		; 0x9E
	goto	CMD__NOP		; 0x9F
	goto	CMD__WR_PIX_LINE	; 0xA0
	goto	CMD__WR_PIX_LINE_C	; 0xA1
	goto	CMD__WR_OVP_LINE	; 0xA2
	goto	CMD__WR_OVP_LINE_C	; 0xA3
	goto	CMD__WR_PIX_QUART	; 0xA4
	goto	CMD__WR_PIX_QUART_C	; 0xA5
	goto	CMD__WR_OVP_QUART	; 0xA6
	goto	CMD__WR_OVP_QUART_C	; 0xA7
	goto	CMD__WR_PIX_CLR		; 0xA8
	goto	CMD__WR_OVP_CLR		; 0xA9
	goto	CMD__WR_PIX_SET		; 0xAA
	goto	CMD__WR_OVP_SET		; 0xAB
	goto	CMD__WR_PIX_BW		; 0xAC
	goto	CMD__WR_OVP_BW		; 0xAD
	goto	CMD__WR_PIX_SEL		; 0xAE
	goto	CMD__WR_PIX_SEL_C	; 0xAF
	goto	CMD__WR_OVP_SEL		; 0xB0
	goto	CMD__WR_OVP_SEL_C	; 0xB1
	goto	CMD__NOP		; 0xB2
	goto	CMD__NOP		; 0xB3
	goto	CMD__CP_BLK		; 0xB4
	goto	CMD__NOP		; 0xB5
	goto	CMD__CP_BLK_TR		; 0xB6
	goto	CMD__CP_BLK_TR_C	; 0xB7
	goto	CMD__NOP		; 0xB8
	goto	CMD__NOP		; 0xB9
	goto	CMD__NOP		; 0xBA
	goto	CMD__NOP		; 0xBB
	goto	CMD__NOP		; 0xBC
	goto	CMD__NOP		; 0xBD
	goto	CMD__NOP		; 0xBE
	goto	CMD__NOP		; 0xBF
	goto	CMD__DISPLAY		; 0xC0
	goto	CMD__DISPLAY_CP		; 0xC1
	goto	CMD__NOP		; 0xC2
	goto	CMD__NOP		; 0xC3
	goto	CMD__NOP		; 0xC4
	goto	CMD__NOP		; 0xC5
	goto	CMD__NOP		; 0xC6
	goto	CMD__NOP		; 0xC7
	goto	CMD__DELAY_DO		; 0xC8
	goto	CMD__DELAY_DO_C		; 0xC9
	goto	CMD__DELAY_START	; 0xCA
	goto	CMD__DELAY_START_C	; 0xCB
	goto	CMD__DELAY_CHECK	; 0xCD
	goto	CMD__DELAY_WAIT		; 0xCD
	goto	CMD__NOP		; 0xCE
	goto	CMD__NOP		; 0xCF
	goto	CMD__GET_KEYS		; 0xD0
	goto	CMD__NOP		; 0xD1
	goto	CMD__GET_SNES0		; 0xD2
	goto	CMD__GET_SNES1		; 0xD3
	goto	CMD__GET_TIME		; 0xD4
	goto	CMD__SET_TIME		; 0xD5
	goto	CMD__NOP		; 0xD6
	goto	CMD__NOP		; 0xD7
	goto	CMD__RND		; 0xD8
	goto	CMD__NOP		; 0xD9
	goto	CMD__NOP		; 0xDA
	goto	CMD__NOP		; 0xDB
	goto	CMD__VER		; 0xDC
	goto	CMD__NOP		; 0xDD
	goto	CMD__NOP		; 0xDE
	goto	CMD__NOP		; 0xDF
	goto	CMD__NOP		; 0xE0
	goto	CMD__NOP		; 0xE1
	goto	CMD__NOP		; 0xE2
	goto	CMD__NOP		; 0xE3
	goto	CMD__NOP		; 0xE4
	goto	CMD__NOP		; 0xE5
	goto	CMD__NOP		; 0xE6
	goto	CMD__NOP		; 0xE7
	goto	CMD__NOP		; 0xE8
	goto	CMD__NOP		; 0xE9
	goto	CMD__NOP		; 0xEA
	goto	CMD__NOP		; 0xEB
	goto	CMD__NOP		; 0xEC
	goto	CMD__NOP		; 0xED
	goto	CMD__NOP		; 0xEE
	goto	CMD__NOP		; 0xEF
	goto	CMD__NOP		; 0xF0
	goto	CMD__NOP		; 0xF1
	goto	CMD__NOP		; 0xF2
	goto	CMD__NOP		; 0xF3
	goto	CMD__NOP		; 0xF4
	goto	CMD__NOP		; 0xF5
	goto	CMD__NOP		; 0xF6
	goto	CMD__NOP		; 0xF7
	goto	CMD__NOP		; 0xF8
	goto	CMD__NOP		; 0xF9
	goto	CMD__NOP		; 0xFA
	goto	CMD__NOP		; 0xFB
	goto	CMD__NOP		; 0xFC
	goto	CMD__NOP		; 0xFD
	goto	CMD__NOP		; 0xFE
	goto	CMD__NOP		; 0xFF

;-------------------------------------------------------------------------------------------
; load a sector
; - loads sector number CMD_SECTOR from compact flash or internal flash
; - sets compact flash error bit on compact flash error
; - sets sector to 0 and offset to 0x200 (jump to sector 1) on internal flash error
; - can be used to implement JMP_ commmands
;-------------------------------------------------------------------------------------------
CMD_LOAD_SECTOR:
; check if in compact flash mode
	btfss	BIT_CMD_CF
	bra	CMD_LOAD_SECTOR_IF

; compact flash mode
; check if compact flash busy
	btfss	BIT_CF_BUSY		; jump if not busy
	bra	CMD_LOAD_SECTOR_CF_NOT_BUSY
	movlw	CMD_CF_WAIT_WANT_SEC	; wait for compact flash to become available
	movwf	CMD_CF_WAIT
	goto	CMD_FINISH_OFFSET_NOSTOP ; done for now
CMD_LOAD_SECTOR_CF_NOT_BUSY:
; read sector of compact flash
	movff	CMD_SECTOR+0,A4_REG+0	; sector number
	movff	CMD_SECTOR+1,A4_REG+1
	movff	CMD_SECTOR+2,A4_REG+2
	movff	CMD_SECTOR+3,A4_REG+3
	lfsr	FSR0,BUFFER		; address of buffer (512 bytes)
	call	CF_READ_SECTOR
	btfsc	STATUS,C
	bra	CMD_LOAD_SECTOR_CF_ERR	; jump in case of error
	movlw	CMD_CF_WAIT_LOAD_SEC	; wait for compact flash to load the sector
	movwf	CMD_CF_WAIT
; done
	goto	CMD_FINISH_OFFSET
CMD_LOAD_SECTOR_CF_ERR:
; switch back to internal flash
	bcf	BIT_CMD_CF
; done
	goto	CMD_FINISH_OFFSET

; internal flash mode
CMD_LOAD_SECTOR_IF:
; read sector of internal flash
	movff	CMD_SECTOR+0,A4_REG+0	; sector number
	movff	CMD_SECTOR+1,A4_REG+1
	movff	CMD_SECTOR+2,A4_REG+2
	movff	CMD_SECTOR+3,A4_REG+3
	lfsr	FSR0,BUFFER		; address of buffer (512 bytes)
	call	IF_READ_SECTOR
; done if no error
	btfss	STATUS,C
	goto	CMD_FINISH_OFFSET
; internal flash error (invalid sector number)
	clrf	CMD_SECTOR		; sector 0
	clrf	CMD_SECTOR+1
	clrf	CMD_SECTOR+2
	clrf	CMD_SECTOR+3
	movlw	0x02			; set command pointer just behind command buffer
	movwf	CMD_OFFSET+1		; (overflow buffer contains 0x00 = jump to start of next sector)
	clrf	CMD_OFFSET+0
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NEXT_SEC
; load the next sector
;-------------------------------------------------------------------------------------------
CMD__JMP_NEXT_SEC:
; increment sector number
	movlw	0
	incf	CMD_SECTOR+0
	addwfc	CMD_SECTOR+1
	addwfc	CMD_SECTOR+2
	addwfc	CMD_SECTOR+3
; set command pointer to offset 0
	clrf	CMD_OFFSET+0
	clrf	CMD_OFFSET+1
; load sector
	goto	CMD_LOAD_SECTOR

;-------------------------------------------------------------------------------------------
; JMP_SEC <sectors (4)>
; jump <sectors> sectors further
; - <sectors (4)> is stored in big endian format
;-------------------------------------------------------------------------------------------
CMD__JMP_SEC:
; get number of sectors
	movff	POSTINC1,A4_REG+3
	movff	POSTINC1,A4_REG+2
	movff	POSTINC1,A4_REG+1
	movff	POSTINC1,A4_REG+0
; add to sector number
	movf	A4_REG+0,W
	addwf	CMD_SECTOR+0
	movf	A4_REG+1,W
	addwfc	CMD_SECTOR+1
	movf	A4_REG+2,W
	addwfc	CMD_SECTOR+2
	movf	A4_REG+3,W
	addwfc	CMD_SECTOR+3
; set command pointer to offset 0
	clrf	CMD_OFFSET+0
	clrf	CMD_OFFSET+1
; load sector
	goto	CMD_LOAD_SECTOR

;-------------------------------------------------------------------------------------------
; JMP_OFS (bit0 is bit8 of offset) <offset>
; jump to offset
;-------------------------------------------------------------------------------------------
CMD__JMP_OFS:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	CMD_OFFSET+1
	movff	POSTINC1,CMD_OFFSET+0	; low byte of offset
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_OFS_SEC (bit0 is bit8 of offset) <offset> <sectors (4)>
; jump to offset and <sectors> sectors further
; - <sectors (4)> is stored in big endian format
;-------------------------------------------------------------------------------------------
CMD__JMP_OFS_SEC:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	CMD_OFFSET+1
	movff	POSTINC1,CMD_OFFSET+0	; low byte of offset
; get number of sectors
	movff	POSTINC1,A4_REG+3
	movff	POSTINC1,A4_REG+2
	movff	POSTINC1,A4_REG+1
	movff	POSTINC1,A4_REG+0
; add to sector number
	movf	A4_REG+0,W
	addwf	CMD_SECTOR+0
	movf	A4_REG+1,W
	addwfc	CMD_SECTOR+1
	movf	A4_REG+2,W
	addwfc	CMD_SECTOR+2
	movf	A4_REG+3,W
	addwfc	CMD_SECTOR+3
; load sector
	goto	CMD_LOAD_SECTOR

;-------------------------------------------------------------------------------------------
; JMP_IND <offset_h> <offset_l> <sectors_hh> <sectors_hl> <sectors_lh> <sectors_ll>
; jump indirect to offset and <sectors> sectors further
;-------------------------------------------------------------------------------------------
CMD__JMP_IND:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get offset
	movff	POSTINC1,FSR0L		; get offset of offset_h
	movf	INDF0,W			; get offset_h
	andlw	0x01			; (offset has only 9 bits)
	movwf	CMD_OFFSET+1
	movff	POSTINC1,FSR0L		; get offset of offset_l
	movff	INDF0,CMD_OFFSET+0	; get offset_l
; get number of sectors
	movff	POSTINC1,FSR0L		; get offset of sectors_hh
	movff	INDF0,A4_REG+3		; get sectors_hh
	movff	POSTINC1,FSR0L		; get offset of sectors_hl
	movff	INDF0,A4_REG+2		; get sectors_hl
	movff	POSTINC1,FSR0L		; get offset of sectors_lh
	movff	INDF0,A4_REG+1		; get sectors_lh
	movff	POSTINC1,FSR0L		; get offset of sectors_ll
	movff	INDF0,A4_REG+0		; get sectors_ll
; add to sector number
	movf	A4_REG+0,W
	addwf	CMD_SECTOR+0
	movf	A4_REG+1,W
	addwfc	CMD_SECTOR+1
	movf	A4_REG+2,W
	addwfc	CMD_SECTOR+2
	movf	A4_REG+3,W
	addwfc	CMD_SECTOR+3
; load sector
	goto	CMD_LOAD_SECTOR

;-------------------------------------------------------------------------------------------
; JMP_EQ (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if equal
; jump to offset if <param1> == <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_EQ:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
; xor bitwise with param2
	movff	POSTINC1,FSR0L		; get offset of param2
	xorwf	INDF0,W			; xor bitwise with param2
; exit if not equal (bitwise xor is not 0)
	btfss	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_EQ_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if equal to constant
; jump to offset if <param> == <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_EQ_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
; xor bitwise with constant
	xorwf	POSTINC1,W
; exit if not equal (bitwise xor is not 0)
	btfss	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NEQ (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if not equal
; jump to offset if <param1> != <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_NEQ:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
; xor bitwise with param2
	movff	POSTINC1,FSR0L		; get offset of param2
	xorwf	INDF0,W			; xor bitwise with param2
; exit if equal (bitwise xor is 0)
	btfsc	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NEQ_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if not equal to constant
; jump to offset if <param> != <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_NEQ_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
; xor bitwise with constant
	xorwf	POSTINC1,W
; exit if equal (bitwise xor is 0)
	btfsc	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_GT (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if greater than
; jump to offset if <param1> > <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_GT:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movff	INDF0,A1_REG		; get param1
; get param2
	movff	POSTINC1,FSR0L		; get offset of param2
	movf	INDF0,W			; get param2
; exit if less or equal
	cpfsgt	A1_REG
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_GT_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if greater than
; jump to offset if <param> > <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_GT_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movff	INDF0,A1_REG		; get param
; get constant
	movf	POSTINC1,W
; exit if less or equal
	cpfsgt	A1_REG
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NGT (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if not greater than
; jump to offset if <param1> <= <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_NGT:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movff	INDF0,A1_REG		; get param1
; get param2
	movff	POSTINC1,FSR0L		; get offset of param2
	movf	INDF0,W			; get param2
; exit if greater
	cpfsgt	A1_REG
	bra	CMD__JMP_NGT_JMP
	goto	CMD_FINISH
CMD__JMP_NGT_JMP:
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NGT_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if not greater than
; jump to offset if <param> <= <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_NGT_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movff	INDF0,A1_REG		; get param
; get constant
	movf	POSTINC1,W
; exit if greater
	cpfsgt	A1_REG
	bra	CMD__JMP_NGT_C_JMP
	goto	CMD_FINISH
CMD__JMP_NGT_C_JMP:
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_LT (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if less than
; jump to offset if <param1> < <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_LT:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movff	INDF0,A1_REG		; get param1
; get param2
	movff	POSTINC1,FSR0L		; get offset of param2
	movf	INDF0,W			; get param2
; exit if greater or equal
	cpfslt	A1_REG
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_LT_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if less than
; jump to offset if <param> < <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_LT_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movff	INDF0,A1_REG		; get param
; get constant
	movf	POSTINC1,W
; exit if greater or equal
	cpfslt	A1_REG
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NLT (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if not less than
; jump to offset if <param1> >= <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_NLT:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movff	INDF0,A1_REG		; get param1
; get param2
	movff	POSTINC1,FSR0L		; get offset of param2
	movf	INDF0,W			; get param2
; exit if less
	cpfslt	A1_REG
	bra	CMD__JMP_NLT_JMP
	goto	CMD_FINISH
CMD__JMP_NLT_JMP:
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NLT_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if not less than
; jump to offset if <param> >= <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_NLT_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movff	INDF0,A1_REG		; get param
; get constant
	movf	POSTINC1,W
; exit if less
	cpfslt	A1_REG
	bra	CMD__JMP_NLT_C_JMP
	goto	CMD_FINISH
CMD__JMP_NLT_C_JMP:
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_GTS (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if greater (signed) than
; jump to offset if <param1> > <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_GTS:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1 (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
	addlw	0x80			; add 0x80
	movwf	A1_REG
; get param2 (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param2
	movf	INDF0,W			; get param2
	addlw	0x80			; add 0x80
; exit if less or equal
	cpfsgt	A1_REG
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_GTS_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if greater (signed) than
; jump to offset if <param> > <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_GTS_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
	addlw	0x80			; add 0x80
	movwf	A1_REG
; get constant (and add 0x80 to transform to unsigned comparison)
	movf	POSTINC1,W		; get constant
	addlw	0x80			; add 0x80
; exit if less or equal
	cpfsgt	A1_REG
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NGTS (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if not greater (signed) than
; jump to offset if <param1> <= <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_NGTS:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1 (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
	addlw	0x80			; add 0x80
	movwf	A1_REG
; get param2 (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param2
	movf	INDF0,W			; get param2
	addlw	0x80			; add 0x80
; exit if greater
	cpfsgt	A1_REG
	bra	CMD__JMP_NGTS_JMP
	goto	CMD_FINISH
CMD__JMP_NGTS_JMP:
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NGTS_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if not greater (signed) than
; jump to offset if <param> <= <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_NGTS_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
	addlw	0x80			; add 0x80
	movwf	A1_REG
; get constant (and add 0x80 to transform to unsigned comparison)
	movf	POSTINC1,W		; get constant
	addlw	0x80			; add 0x80
; exit if greater
	cpfsgt	A1_REG
	bra	CMD__JMP_NGTS_C_JMP
	goto	CMD_FINISH
CMD__JMP_NGTS_C_JMP:
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_LTS (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if less (signed) than
; jump to offset if <param1> < <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_LTS:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1 (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
	addlw	0x80			; add 0x80
	movwf	A1_REG
; get param2 (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param2
	movf	INDF0,W			; get param2
	addlw	0x80			; add 0x80
; exit if greater or equal
	cpfslt	A1_REG
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_LTS_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if less (signed) than
; jump to offset if <param> < <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_LTS_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
	addlw	0x80			; add 0x80
	movwf	A1_REG
; get constant (and add 0x80 to transform to unsigned comparison)
	movf	POSTINC1,W		; get constant
	addlw	0x80			; add 0x80
; exit if greater or equal
	cpfslt	A1_REG
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NLTS (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if not less (signed) than
; jump to offset if <param1> >= <param2>
;-------------------------------------------------------------------------------------------
CMD__JMP_NLTS:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1 (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
	addlw	0x80			; add 0x80
	movwf	A1_REG
; get param2 (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param2
	movf	INDF0,W			; get param2
	addlw	0x80			; add 0x80
; exit if less
	cpfslt	A1_REG
	bra	CMD__JMP_NLTS_JMP
	goto	CMD_FINISH
CMD__JMP_NLTS_JMP:
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_NLTS_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if not less (signed) than
; jump to offset if <param> >= <constant>
;-------------------------------------------------------------------------------------------
CMD__JMP_NLTS_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param (and add 0x80 to transform to unsigned comparison)
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
	addlw	0x80			; add 0x80
	movwf	A1_REG
; get constant (and add 0x80 to transform to unsigned comparison)
	movf	POSTINC1,W		; get constant
	addlw	0x80			; add 0x80
; exit if less
	cpfslt	A1_REG
	bra	CMD__JMP_NLTS_C_JMP
	goto	CMD_FINISH
CMD__JMP_NLTS_C_JMP:
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_AND_Z (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if bitwise and is zero
; jump to offset if <param1> AND <param2> == 0
;-------------------------------------------------------------------------------------------
CMD__JMP_AND_Z:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
; and bitwise with param2
	movff	POSTINC1,FSR0L		; get offset of param2
	andwf	INDF0,W			; and bitwise with param2
; exit if bitwise and is not zero
	btfss	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_AND_Z_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if bitwise and is zero
; jump to offset if <param> AND <constant> == 0
;-------------------------------------------------------------------------------------------
CMD__JMP_AND_Z_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
; and bitwise with constant
	andwf	POSTINC1,W
; exit if bitwise and is not zero
	btfss	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_AND_NZ (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if bitwise and is not zero
; jump to offset if <param1> AND <param2> != 0
;-------------------------------------------------------------------------------------------
CMD__JMP_AND_NZ:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
; and bitwise with param2
	movff	POSTINC1,FSR0L		; get offset of param2
	andwf	INDF0,W			; and bitwise with param2
; exit if bitwise and is zero
	btfsc	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_AND_NZ_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if bitwise and is not zero
; jump to offset if <param> AND <constant> != 0
;-------------------------------------------------------------------------------------------
CMD__JMP_AND_NZ_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
; and bitwise with constant
	andwf	POSTINC1,W
; exit if bitwise and is zero
	btfsc	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_OR_M (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if bitwise or is maximum
; jump to offset if <param1> OR <param2> == 0xFF
;-------------------------------------------------------------------------------------------
CMD__JMP_OR_M:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
; or bitwise with param2
	movff	POSTINC1,FSR0L		; get offset of param2
	iorwf	INDF0,W			; or bitwise with param2
; exit if bitwise or is not maximum
	xorlw	0xFF
	btfss	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_OR_M_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if bitwise or is maximum
; jump to offset if <param> OR <constant> == 0xFF
;-------------------------------------------------------------------------------------------
CMD__JMP_OR_M_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
; or bitwise with constant
	iorwf	POSTINC1,W
; exit if bitwise or is not maximum
	xorlw	0xFF
	btfss	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_OR_NM (bit0 is bit8 of offset) <offset> <param1> <param2>
; jump if bitwise or is not maximum
; jump to offset if <param1> OR <param2> != 0xFF
;-------------------------------------------------------------------------------------------
CMD__JMP_OR_NM:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param1
	movff	POSTINC1,FSR0L		; get offset of param1
	movf	INDF0,W			; get param1
; or bitwise with param2
	movff	POSTINC1,FSR0L		; get offset of param2
	iorwf	INDF0,W			; or bitwise with param2
; exit if bitwise or is maximum
	xorlw	0xFF
	btfsc	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; JMP_OR_NM_C (bit0 is bit8 of offset) <offset> <param> <constant>
; jump if bitwise or is not maximum
; jump to offset if <param> OR <constant> != 0xFF
;-------------------------------------------------------------------------------------------
CMD__JMP_OR_NM_C:
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	A2_REG+1
	movff	POSTINC1,A2_REG		; low byte of offset
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get param
	movff	POSTINC1,FSR0L		; get offset of param
	movf	INDF0,W			; get param
; or bitwise with constant
	iorwf	POSTINC1,W
; exit if bitwise or is maximum
	xorlw	0xFF
	btfsc	STATUS,Z
	goto	CMD_FINISH
; set command pointer to offset
	movff	A2_REG+0,CMD_OFFSET+0
	movff	A2_REG+1,CMD_OFFSET+1
; done
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; MOV <source> <dest>
; move
; write contents of source into dest
;-------------------------------------------------------------------------------------------
CMD__MOV:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movf	INDF0,W			; get source
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MOV_C <constant> <dest>
; move constant
; write constant into dest
;-------------------------------------------------------------------------------------------
CMD__MOV_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MOV2 <source1> <dest1> <source2> <dest2>
; move 2
; write contents of source<x> into dest<x>
;-------------------------------------------------------------------------------------------
CMD__MOV2:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movf	INDF0,W			; get source1
; write to dest1
	movff	POSTINC1,FSR0L		; get offset of dest1
	movwf	INDF0			; write to dest1
; get source2
	movff	POSTINC1,FSR0L		; get offset of source2
	movf	INDF0,W			; get source2
; write to dest2
	movff	POSTINC1,FSR0L		; get offset of dest2
	movwf	INDF0			; write to dest2
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MOV2_C <constant1> <dest1> <constant2> <dest2>
; move 2 constants
; write constant<x> into dest<x>
;-------------------------------------------------------------------------------------------
CMD__MOV2_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant1
	movf	POSTINC1,W
; write to dest1
	movff	POSTINC1,FSR0L		; get offset of dest1
	movwf	INDF0			; write to dest1
; get constant2
	movf	POSTINC1,W
; write to dest2
	movff	POSTINC1,FSR0L		; get offset of dest2
	movwf	INDF0			; write to dest2
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MOV4 <source1> <dest1> ... <source4> <dest4>
; move 4
; write contents of source<x> into dest<x>
;-------------------------------------------------------------------------------------------
CMD__MOV4:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movf	INDF0,W			; get source1
; write to dest1
	movff	POSTINC1,FSR0L		; get offset of dest1
	movwf	INDF0			; write to dest1
; get source2
	movff	POSTINC1,FSR0L		; get offset of source2
	movf	INDF0,W			; get source2
; write to dest2
	movff	POSTINC1,FSR0L		; get offset of dest2
	movwf	INDF0			; write to dest2
; get source3
	movff	POSTINC1,FSR0L		; get offset of source3
	movf	INDF0,W			; get source3
; write to dest3
	movff	POSTINC1,FSR0L		; get offset of dest3
	movwf	INDF0			; write to dest3
; get source4
	movff	POSTINC1,FSR0L		; get offset of source4
	movf	INDF0,W			; get source4
; write to dest4
	movff	POSTINC1,FSR0L		; get offset of dest4
	movwf	INDF0			; write to dest4
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MOV4_C <constant1> <dest1> ... <constant4> <dest4>
; move 4 constants
; write constant<x> into dest<x>
;-------------------------------------------------------------------------------------------
CMD__MOV4_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant1
	movf	POSTINC1,W
; write to dest1
	movff	POSTINC1,FSR0L		; get offset of dest1
	movwf	INDF0			; write to dest1
; get constant2
	movf	POSTINC1,W
; write to dest2
	movff	POSTINC1,FSR0L		; get offset of dest2
	movwf	INDF0			; write to dest2
; get constant3
	movf	POSTINC1,W
; write to dest3
	movff	POSTINC1,FSR0L		; get offset of dest3
	movwf	INDF0			; write to dest3
; get constant4
	movf	POSTINC1,W
; write to dest5
	movff	POSTINC1,FSR0L		; get offset of dest4
	movwf	INDF0			; write to dest4
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; RD_I <addr> <dest>
; read indirect
; write contents of variable pointed to by addr into dest
;-------------------------------------------------------------------------------------------
CMD__RD_I:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get addr
	movff	POSTINC1,FSR0L		; get offset of addr
	movff	INDF0,FSR0L		; get addr
; get contents of variable at addr
	movf	INDF0,W
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; WR_I <source> <addr>
; write indirect
; write contents of source into variable pointed to by addr
;-------------------------------------------------------------------------------------------
CMD__WR_I:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movf	INDF0,W			; get source
; get addr
	movff	POSTINC1,FSR0L		; get offset of addr
	movff	INDF0,FSR0L		; get addr
; write to variable at addr
	movwf	INDF0
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; RD_S <offset_h> <offset_l> <dest>
; read from current sector
; write byte at offset of current sector into dest
;-------------------------------------------------------------------------------------------
CMD__RD_S:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get offset
	movff	POSTINC1,FSR0L		; get offset of offset_h variable
	movff	INDF0,A2_REG+1		; get offset_h
	movff	POSTINC1,FSR0L		; get offset of offset_l variable
	movff	INDF0,A2_REG		; get offset_l
; get byte at offset of current sector
	movff	A2_REG+1,FSR0H
	movff	A2_REG,FSR0L
	movff	INDF0,W
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; RD_CF <off_h> <off_l> <sec_hh> <sec_hl> <sec_lh> <sec_ll> <dest> <follow> <err>
; read from compact flash
; reads the byte at offset <ofs> in the sector <sec> sectors behind the current one
; of the compact flash and the <follow> following bytes
; and places them into <dest> and the <follow> following registers
; goto	CMD_FINISHs an error code in <err>: 0 = no error, 1 = invalid parameters, 2 = CF error
;-------------------------------------------------------------------------------------------
CMD__RD_CF: 
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; put offset into A2_REG
	movff	POSTINC1,FSR0L		; get offset of off_h
	movf	INDF0,W			; get off_h
	andlw	0x01			; (offset has only 9 bits)
	movwf	A2_REG+1
	movff	POSTINC1,FSR0L		; get offset of off_l
	movff	INDF0,A2_REG		; get off_l
; get number of sectors
	movff	POSTINC1,FSR0L		; get offset of sec_hh
	movff	INDF0,A4_REG+3		; get sec_hh
	movff	POSTINC1,FSR0L		; get offset of sec_hl
	movff	INDF0,A4_REG+2		; get sec_hl
	movff	POSTINC1,FSR0L		; get offset of sec_lh
	movff	INDF0,A4_REG+1		; get sec_lh
	movff	POSTINC1,FSR0L		; get offset of sec_ll
	movff	INDF0,A4_REG		; get sec_ll
; put absolute sector number A4_REG
	movf	CMD_SECTOR+0,W
	addwf	A4_REG+0
	movf	CMD_SECTOR+1,W
	addwfc	A4_REG+1
	movf	CMD_SECTOR+2,W
	addwfc	A4_REG+2
	movf	CMD_SECTOR+3,W
	addwfc	A4_REG+3
; put destination address into FSR0
	movff	POSTINC1,FSR0L		; get number of first destination register
; put length-1 into C2_REG
	movff	POSTINC1,C2_REG		; get number of following registers
	clrf	C2_REG+1
; check parameters
	movlw	1			; error code for parameter error into A1_REG
	movwf	A1_REG
	movf	A2_REG,W		; offset + length-1 < 512
	addwf	C2_REG,W
	movwf	D2_REG
	movf	A2_REG+1,W
	addwfc	C2_REG+1
	movwf	D2_REG+1
	movlw	2
	cpfslt	D2_REG+1
	bra	CMD__RD_CF_END
	movf	FSR0L,W			; first register + length-1 < 256
	addwf	C2_REG,W
	movwf	D2_REG
	movf	A2_REG+1,W
	addwfc	C2_REG+1
	movwf	D2_REG+1
	movlw	1
	cpfslt	D2_REG+1
	bra	CMD__RD_CF_END
; put length into C2_REG
	incf	C2_REG			; add 1 to get length
	btfsc	STATUS,C
	incf	C2_REG+1
; if CF is busy, do this command again later
	btfss	BIT_CF_BUSY
	bra	CMD__RD_CF_NOTBUSY
	movlw	9			; put command pointer back to begin of command
	subwf	FSR1L
	movlw	0
	subwfb	FSR1H
	goto	CMD_FINISH_NOSTOP
CMD__RD_CF_NOTBUSY:
; read from compact flash
	call	CF_READ_SECTOR_PART
; if no error, start waiting for command to complete
	btfsc	STATUS,C
	bra	CMD__RD_CF_ERR
	movlw	CMD_CF_WAIT_RD_CF
	movwf	CMD_CF_WAIT
	goto	CMD_FINISH_NOSTOP
; put error code into A1_REG
CMD__RD_CF_ERR:
	movlw	2			; CF error
	movwf	A1_REG
CMD__RD_CF_END:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; put offset into A2_REG
	movff	POSTINC1,FSR0L		; get offset of err
	movff	A1_REG,INDF0		; write error code to err
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; ADD <source1> <source2> <dest>
; add
; write sum of contents of source1 and contants of source2 into dest
;-------------------------------------------------------------------------------------------
CMD__ADD:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movf	INDF0,W			; get source1
; add source2
	movff	POSTINC1,FSR0L		; get offset of source2
	addwf	INDF0,W			; add source2
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; ADD_C <constant> <source> <dest>
; add constant
; write sum of constant and contents of source into dest
;-------------------------------------------------------------------------------------------
CMD__ADD_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
; add source
	movff	POSTINC1,FSR0L		; get offset of source
	addwf	INDF0,W			; add source
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; ADDC <source1> <source2> <carry_in> <dest> <carry_out>
; add (with carry)
; write sum of contents of source1 and contants of source2 into dest
; bit 0 of carry_in is also included in the sum
; carry_out is 0x01 if the result is >= 0x100, 0x00 otherwise
;-------------------------------------------------------------------------------------------
CMD__ADDC:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movff	INDF0,A1_REG		; get source1
; get source2
	movff	POSTINC1,FSR0L		; get offset of source2
	movf	INDF0,W			; get source2
; get carry_in
	movff	POSTINC1,FSR0L		; get offset of carry_in
	bcf	STATUS,C		; get carry_in
	btfsc	INDF0,0
	bsf	STATUS,C	
; calculate sum
	addwfc	A1_REG,W
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; write carry to carry_out
	movff	POSTINC1,FSR0L		; get offset of carry_out
	clrf	INDF0			; write carry to carry_out
	btfsc	STATUS,C
	incf	INDF0
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; ADDC_C <constant> <source> <carry_in> <dest> <carry_out>
; add constant (with carry)
; write sum of constant and contents of source into dest
; bit 0 of carry_in is also included in the sum
; carry_out is 0x01 if the result is >= 0x100, 0x00 otherwise
;-------------------------------------------------------------------------------------------
CMD__ADDC_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movff	POSTINC1,A1_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movf	INDF0,W			; get source
; get carry_in
	movff	POSTINC1,FSR0L		; get offset of carry_in
	bcf	STATUS,C		; get carry_in
	btfsc	INDF0,0
	bsf	STATUS,C	
; calculate sum
	addwfc	A1_REG,W
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; write carry to carry_out
	movff	POSTINC1,FSR0L		; get offset of carry_out
	clrf	INDF0			; write carry to carry_out
	btfsc	STATUS,C
	incf	INDF0
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SUB <source1> <source2> <dest>
; subtract
; write difference of contents of source1 minus contents of source2 into dest
;-------------------------------------------------------------------------------------------
CMD__SUB:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movff	INDF0,A1_REG		; get source1
; get source2
	movff	POSTINC1,FSR0L		; get offset of source2
	movf	INDF0,W			; get source2
; subtract source2 from source1
	subwf	A1_REG,W
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SUB_C <constant> <source> <dest>
; subtract constant
; write difference of constant minus contants of source into dest
;-------------------------------------------------------------------------------------------
CMD__SUB_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movff	POSTINC1,A1_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movf	INDF0,W			; get source
; subtract source2 from source1
	subwf	A1_REG,W
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SUBB <source1> <source2> <borrow_in> <dest> <borrow_out>
; subtract (with borrow)
; write difference of contents of source1 minus contents of source2 into dest
; bit 0 of borrow_in is also subtracted
; borrow_out is 0x01 if the result is < 0x00, 0x00 otherwise
;-------------------------------------------------------------------------------------------
CMD__SUBB:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movff	INDF0,A1_REG		; get source1
; get source2
	movff	POSTINC1,FSR0L		; get offset of source2
	movf	INDF0,W			; get source2
; get borrow_in
	movff	POSTINC1,FSR0L		; get offset of borrow_in
	bcf	STATUS,C		; get borrow_in
	btfss	INDF0,0
	bsf	STATUS,C	
; subtract source2 from source1 (with borrow)
	subwfb	A1_REG,W
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; write borrow to borrow_out
	movff	POSTINC1,FSR0L		; get offset of borrow_out
	clrf	INDF0			; write carry to borrow_out
	btfss	STATUS,C
	incf	INDF0
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SUBB_C <constant> <source> <borrow_in> <dest> <borrow_out>
; subtract from constant (with borrow)
; write difference of constant minus contants of source into dest
; bit 0 of borrow_in is also subtracted
; borrow_out is 0x01 if the result is < 0x00, 0x00 otherwise
;-------------------------------------------------------------------------------------------
CMD__SUBB_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movff	POSTINC1,A1_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movf	INDF0,W			; get source
; get borrow_in
	movff	POSTINC1,FSR0L		; get offset of borrow_in
	bcf	STATUS,C		; get borrow_in
	btfss	INDF0,0
	bsf	STATUS,C	
; subtract source from constant (with borrow)
	subwfb	A1_REG,W
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; write borrow to borrow_out
	movff	POSTINC1,FSR0L		; get offset of borrow_out
	clrf	INDF0			; write carry to borrow_out
	btfss	STATUS,C
	incf	INDF0
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MUL <source1> <source2> <dest_h> <dest_l>
; multiply
; write product of contents of source1 and contants of source2 into dest_h and dest_l
;-------------------------------------------------------------------------------------------
CMD__MUL:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movf	INDF0,W			; get source1
; multiply with source2
	movff	POSTINC1,FSR0L		; get offset of source2
	mulwf	INDF0			; multiply with source2
; write to dest_h
	movff	POSTINC1,FSR0L		; get offset of dest_h
	movff	PRODH,INDF0		; write to dest_h
; write to dest_l
	movff	POSTINC1,FSR0L		; get offset of dest_l
	movff	PRODL,INDF0		; write to dest_l
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MUL_C <constant> <source> <dest_h> <dest_l>
; multiply constant
; write product of constant and contents of source into dest_h and dest_l
;-------------------------------------------------------------------------------------------
CMD__MUL_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
; multiply with source
	movff	POSTINC1,FSR0L		; get offset of source
	mulwf	INDF0			; multiply with source
; write to dest_h
	movff	POSTINC1,FSR0L		; get offset of dest_h
	movff	PRODH,INDF0		; write to dest_h
; write to dest_l
	movff	POSTINC1,FSR0L		; get offset of dest_l
	movff	PRODL,INDF0		; write to dest_l
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MULS <source1> <source2> <dest_h> <dest_l>
; multiply (signed)
; write product of contents of source1 and contants of source2 into dest_h and dest_l
;-------------------------------------------------------------------------------------------
CMD__MULS:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movff	INDF0,A1_REG		; get source1
; get source1
	movff	POSTINC1,FSR0L		; get offset of source2
	movff	INDF0,B1_REG		; get source2
; multiply signed
	movf	A1_REG,W
	mulwf	B1_REG
	btfsc	B1_REG,7		; test sign bit of Arg2
	subwf	PRODH			; PRODH -= Arg1
	movf	B1_REG,W
	btfsc	A1_REG,7		; test sign bit of Arg1
	subwf	PRODH			; PRODH -= Arg2
; write to dest_h
	movff	POSTINC1,FSR0L		; get offset of dest_h
	movff	PRODH,INDF0		; write to dest_h
; write to dest_l
	movff	POSTINC1,FSR0L		; get offset of dest_l
	movff	PRODL,INDF0		; write to dest_l
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; MULS_C <constant> <source> <dest_h> <dest_l>
; multiply with constant (signed)
; write product of constant and contents of source into dest_h and dest_l
;-------------------------------------------------------------------------------------------
CMD__MULS_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movff	POSTINC1,A1_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,B1_REG		; get source
; multiply signed
	movf	A1_REG,W
	mulwf	B1_REG
	btfsc	B1_REG,7		; test sign bit of Arg2
	subwf	PRODH			; PRODH -= Arg1
	movf	B1_REG,W
	btfsc	A1_REG,7		; test sign bit of Arg1
	subwf	PRODH			; PRODH -= Arg2
; write to dest_h
	movff	POSTINC1,FSR0L		; get offset of dest_h
	movff	PRODH,INDF0		; write to dest_h
; write to dest_l
	movff	POSTINC1,FSR0L		; get offset of dest_l
	movff	PRODL,INDF0		; write to dest_l
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DIV <source_h> <source_l> <divisor> <quotient> <remainder>
; divide
; calculate quotient and remainder of division of source by divisor
;-------------------------------------------------------------------------------------------
CMD__DIV:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source into A2_REG
	movff	POSTINC1,FSR0L		; get offset of source_h
	movff	INDF0,A2_REG+1		; get source_h
	movff	POSTINC1,FSR0L		; get offset of source_l
	movff	INDF0,A2_REG		; get source_l
; get divisor into A1_REG
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get divisor
; divide: setup
	movff	A2_REG,A4_REG		; source into A4_REG
	movff	A2_REG+1,A4_REG+1
	clrf	A4_REG+2
	clrf	B4_REG			; divisor shifted left 16 bit into B4_REG
	clrf	B4_REG+1
	movff	A1_REG,B4_REG+2
	clrf	C4_REG			; current weight (1 << 16) into C4_REG
	clrf	C4_REG+1
	movlw	1
	movwf	C4_REG+2
	clrf  	C2_REG			; quotient (0) into C2_REG
	clrf  	C2_REG+1
; divide: check if subtraction can be executed
CMD_DIV_CHECK:
	movf	B4_REG+2,W
	cpfsgt	A4_REG+2
	bra	CMD_DIV_2
	bra	CMD_DIV_SUB		; byte 2: source > divisor
CMD_DIV_2:
	cpfseq	A4_REG+2
	bra	CMD_DIV_STEP		; byte 2: source < divisor
	movf	B4_REG+1,W
	cpfsgt	A4_REG+1
	bra	CMD_DIV_1
	bra	CMD_DIV_SUB		; byte 1: source > divisor
CMD_DIV_1:
	cpfseq	A4_REG+1
	bra	CMD_DIV_STEP		; byte 1: source < divisor
	movf	B4_REG,W
	cpfsgt	A4_REG
	bra	CMD_DIV_0
	bra	CMD_DIV_SUB		; byte 0: source > divisor
CMD_DIV_0:
	cpfseq	A4_REG
	bra	CMD_DIV_STEP		; byte 0: source < divisor
; divide: subtract
CMD_DIV_SUB:
	movf	B4_REG,W		; subtract divisor from source
	subwf	A4_REG
	movf	B4_REG+1,W
	subwfb	A4_REG+1
	movf	B4_REG+2,W
	subwfb	A4_REG+2
	movf	C4_REG,W		; add weight to quotient
	addwf	C2_REG
	movf	C4_REG+1
	addwfc	C2_REG+1
; divide: step
CMD_DIV_STEP:
	bcf	STATUS,C		; shift right divisor
	rrcf	B4_REG+2
	rrcf	B4_REG+1
	rrcf	B4_REG
	bcf	STATUS,C		; shift right weight
	rrcf	C4_REG+2
	rrcf	C4_REG+1
	rrcf	C4_REG
; divide: check if last turn
	movf	C4_REG+2,W
	iorwf	C4_REG+1,W
	iorwf	C4_REG,W
	btfss	STATUS,Z
	bra	CMD_DIV_CHECK		; weight != 0, next turn
; divide: get results
	movff	C2_REG,A1_REG		; low part of quotient into A1_REG
	movff	A4_REG,B1_REG		; low part of remainder into B1_REG
; write quotient
	movff	POSTINC1,FSR0L		; get offset of quotient
	movff	A1_REG,INDF0		; write quotient
; write remainder
	movff	POSTINC1,FSR0L		; get offset of remainder
	movff	B1_REG,INDF0		; write remainder
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DIVS <source_h> <source_l> <divisor> <quotient> <remainder>
; divide (signed)
; calculate quotient and remainder of division of source by divisor
;-------------------------------------------------------------------------------------------
CMD__DIVS:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source into A2_REG
	movff	POSTINC1,FSR0L		; get offset of source_h
	movff	INDF0,A2_REG+1		; get source_h
	movff	POSTINC1,FSR0L		; get offset of source_l
	movff	INDF0,A2_REG		; get source_l
; get divisor into A1_REG
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get divisor
; make parameters positive
	clrf	E2_REG			; use E2_REG to remember signs
	btfss	A2_REG+1,7
	bra	CMD_DIVS_S_P		; source is positive
	comf	A2_REG			; negate source
	comf	A2_REG+1
	incf	A2_REG
	movlw	0
	addwfc	A2_REG+1
	bsf	E2_REG,0		; remember neg. sign of source
CMD_DIVS_S_P:
	btfss	A1_REG,7
	bra	CMD_DIVS_D_P		; divisor is positive
	negf	A1_REG			; negate divisor
	bsf	E2_REG,1		; remember neg. sign of divisor
CMD_DIVS_D_P:
; divide: setup
	movff	A2_REG,A4_REG		; source into A4_REG
	movff	A2_REG+1,A4_REG+1
	clrf	A4_REG+2
	clrf	B4_REG			; divisor shifted left 16 bit into B4_REG
	clrf	B4_REG+1
	movff	A1_REG,B4_REG+2
	clrf	C4_REG			; current weight (1 << 16) into C4_REG
	clrf	C4_REG+1
	movlw	1
	movwf	C4_REG+2
	clrf  	C2_REG			; quotient (0) into C2_REG
	clrf  	C2_REG+1
; divide: check if subtraction can be executed
CMD_DIVS_CHECK:
	movf	B4_REG+2,W
	cpfsgt	A4_REG+2
	bra	CMD_DIVS_2
	bra	CMD_DIVS_SUB		; byte 2: source > divisor
CMD_DIVS_2:
	cpfseq	A4_REG+2
	bra	CMD_DIVS_STEP		; byte 2: source < divisor
	movf	B4_REG+1,W
	cpfsgt	A4_REG+1
	bra	CMD_DIVS_1
	bra	CMD_DIVS_SUB		; byte 1: source > divisor
CMD_DIVS_1:
	cpfseq	A4_REG+1
	bra	CMD_DIVS_STEP		; byte 1: source < divisor
	movf	B4_REG,W
	cpfsgt	A4_REG
	bra	CMD_DIVS_0
	bra	CMD_DIVS_SUB		; byte 0: source > divisor
CMD_DIVS_0:
	cpfseq	A4_REG
	bra	CMD_DIVS_STEP		; byte 0: source < divisor
; divide: subtract
CMD_DIVS_SUB:
	movf	B4_REG,W		; subtract divisor from source
	subwf	A4_REG
	movf	B4_REG+1,W
	subwfb	A4_REG+1
	movf	B4_REG+2,W
	subwfb	A4_REG+2
	movf	C4_REG,W		; add weight to quotient
	addwf	C2_REG
	movf	C4_REG+1
	addwfc	C2_REG+1
; divide: step
CMD_DIVS_STEP:
	bcf	STATUS,C		; shift right divisor
	rrcf	B4_REG+2
	rrcf	B4_REG+1
	rrcf	B4_REG
	bcf	STATUS,C		; shift right weight
	rrcf	C4_REG+2
	rrcf	C4_REG+1
	rrcf	C4_REG
; divide: check if last turn
	movf	C4_REG+2,W
	iorwf	C4_REG+1,W
	iorwf	C4_REG,W
	btfss	STATUS,Z
	bra	CMD_DIVS_CHECK		; weight != 0, next turn
; divide: get results
	movff	C2_REG,A1_REG		; low part of quotient into A1_REG
	movff	A4_REG,B1_REG		; low part of remainder into B1_REG
; adjust signs of results
	btfsc	E2_REG,0		; skip if source was positive
	negf	A1_REG			; negate quotient
	btfsc	E2_REG,0		; skip if source was positive
	negf	A2_REG			; negate remainder
	btfsc	E2_REG,1		; skip if divisor was positive
	negf	A1_REG			; negate quotient
; write quotient
	movff	POSTINC1,FSR0L		; get offset of quotient
	movff	A1_REG,INDF0		; write quotient
; write remainder
	movff	POSTINC1,FSR0L		; get offset of remainder
	movff	B1_REG,INDF0		; write remainder
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; AND <source1> <source2> <dest>
; bitwise and
; write bitwise and of contents of source1 and contants of source2 into dest
;-------------------------------------------------------------------------------------------
CMD__AND:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movf	INDF0,W			; get source1
; and bitwise with source2
	movff	POSTINC1,FSR0L		; get offset of source2
	andwf	INDF0,W			; and bitwise with source2
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; AND_C <constant> <source> <dest>
; and bitwise with constant
; write bitwise and of constant and contents of source into dest
;-------------------------------------------------------------------------------------------
CMD__AND_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
; and bitwise with source
	movff	POSTINC1,FSR0L		; get offset of source
	andwf	INDF0,W			; and bitwise with source
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; OR <source1> <source2> <dest>
; bitwise or
; write bitwise or of contents of source1 and contants of source2 into dest
;-------------------------------------------------------------------------------------------
CMD__OR:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movf	INDF0,W			; get source1
; or bitwise with source2
	movff	POSTINC1,FSR0L		; get offset of source2
	iorwf	INDF0,W			; or bitwise with source2
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; OR_C <constant> <source> <dest>
; or bitwise with constant
; write bitwise or of constant and contents of source into dest
;-------------------------------------------------------------------------------------------
CMD__OR_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
; or bitwise with source
	movff	POSTINC1,FSR0L		; get offset of source
	iorwf	INDF0,W			; or bitwise with source
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; XOR <source1> <source2> <dest>
; bitwise xor
; write bitwise xor of contents of source1 and contants of source2 into dest
;-------------------------------------------------------------------------------------------
CMD__XOR:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source1
	movff	POSTINC1,FSR0L		; get offset of source1
	movf	INDF0,W			; get source1
; xor bitwise with source2
	movff	POSTINC1,FSR0L		; get offset of source2
	xorwf	INDF0,W			; xor bitwise with source2
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; XOR_C <constant> <source> <dest>
; xor bitwise with constant
; write bitwise xor of constant and contents of source into dest
;-------------------------------------------------------------------------------------------
CMD__XOR_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
; xor bitwise with source
	movff	POSTINC1,FSR0L		; get offset of source
	xorwf	INDF0,W			; xor bitwise with source
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SHL <bits> <source> <dest>
; shift left
; shift left contents of source <bits> bits and write result into dest
;-------------------------------------------------------------------------------------------
CMD__SHL:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get bits
	movff	POSTINC1,FSR0L		; get offset of bits
	movf	INDF0,W			; get bits
	andlw	0x07			; shifting more than 7 bits makes no sense
	movwf	A2_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; shift left
	movf	A2_REG			; do not shift if A2_REG is 0
	bz	CMD__SHL_LOOP_END
CMD__SHL_LOOP:
	bcf	STATUS,C		; shift A2_REG times
	rlcf	A1_REG
	decfsz	A2_REG
	bra	CMD__SHL_LOOP
CMD__SHL_LOOP_END:
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SHL_C <constant> <source> <dest>
; shift left constant number of bits
; shift left contents of source <constant> bits and write result into dest
;-------------------------------------------------------------------------------------------
CMD__SHL_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
	andlw	0x07			; shifting more than 7 bits makes no sense
	movwf	A2_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; shift left
	movf	A2_REG			; do not shift if A2_REG is 0
	bz	CMD__SHL_C_LOOP_END
CMD__SHL_C_LOOP:
	bcf	STATUS,C		; shift A2_REG times
	rlcf	A1_REG
	decfsz	A2_REG
	bra	CMD__SHL_C_LOOP
CMD__SHL_C_LOOP_END:
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SHR <bits> <source> <dest>
; shift right
; shift right contents of source <bits> bits and write result into dest
;-------------------------------------------------------------------------------------------
CMD__SHR:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get bits
	movff	POSTINC1,FSR0L		; get offset of bits
	movf	INDF0,W			; get bits
	andlw	0x07			; shifting more than 7 bits makes no sense
	movwf	A2_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; shift right
	movf	A2_REG			; do not shift if A2_REG is 0
	bz	CMD__SHR_LOOP_END
CMD__SHR_LOOP:
	bcf	STATUS,C		; shift A2_REG times
	rrcf	A1_REG
	decfsz	A2_REG
	bra	CMD__SHR_LOOP
CMD__SHR_LOOP_END:
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SHR_C <constant> <source> <dest>
; shift right constant number of bits
; shift right contents of source <constant> bits and write result into dest
;-------------------------------------------------------------------------------------------
CMD__SHR_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
	andlw	0x07			; shifting more than 7 bits makes no sense
	movwf	A2_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; shift right
	movf	A2_REG			; do not shift if A2_REG is 0
	bz	CMD__SHR_C_LOOP_END
CMD__SHR_C_LOOP:
	bcf	STATUS,C		; shift A2_REG times
	rrcf	A1_REG
	decfsz	A2_REG
	bra	CMD__SHR_C_LOOP
CMD__SHR_C_LOOP_END:
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; ROL <bits> <source> <dest>
; rotate left
; rotate left contents of source <bits> bits and write result into dest
;-------------------------------------------------------------------------------------------
CMD__ROL:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get bits
	movff	POSTINC1,FSR0L		; get offset of bits
	movf	INDF0,W			; get bits
	andlw	0x07			; rotating more than 7 bits makes no sense
	movwf	A2_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; rotate left
	movf	A2_REG			; do not rotate if A2_REG is 0
	bz	CMD__ROL_LOOP_END
CMD__ROL_LOOP:
	rlncf	A1_REG			; rotate A2_REG times
	decfsz	A2_REG
	bra	CMD__ROL_LOOP
CMD__ROL_LOOP_END:
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; ROL_C <constant> <source> <dest>
; rotate left constant number of bits
; rotate left contents of source <constant> bits and write result into dest
;-------------------------------------------------------------------------------------------
CMD__ROL_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
	andlw	0x07			; rotating more than 7 bits makes no sense
	movwf	A2_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; rotate left
	movf	A2_REG			; do not rotate if A2_REG is 0
	bz	CMD__ROL_C_LOOP_END
CMD__ROL_C_LOOP:
	rlncf	A1_REG			; rotate A2_REG times
	decfsz	A2_REG
	bra	CMD__ROL_C_LOOP
CMD__ROL_C_LOOP_END:
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; ROR <bits> <source> <dest>
; rotate right
; rotate right contents of source <bits> bits and write result into dest
;-------------------------------------------------------------------------------------------
CMD__ROR:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get bits
	movff	POSTINC1,FSR0L		; get offset of bits
	movf	INDF0,W			; get bits
	andlw	0x07			; rotating more than 7 bits makes no sense
	movwf	A2_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; rotate right
	movf	A2_REG			; do not shift if A2_REG is 0
	bz	CMD__ROR_LOOP_END
CMD__ROR_LOOP:
	rrncf	A1_REG			; rotate A2_REG times
	decfsz	A2_REG
	bra	CMD__ROR_LOOP
CMD__ROR_LOOP_END:
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; ROR_C <constant> <source> <dest>
; rotate right constant number of bits
; rotate right contents of source <constant> bits and write result into dest
;-------------------------------------------------------------------------------------------
CMD__ROR_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get constant
	movf	POSTINC1,W
	andlw	0x07			; rotating more than 7 bits makes no sense
	movwf	A2_REG
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; rotate right
	movf	A2_REG			; do not rotate if A2_REG is 0
	bz	CMD__ROR_C_LOOP_END
CMD__ROR_C_LOOP:
	rrncf	A1_REG			; rotate A2_REG times
	decfsz	A2_REG
	bra	CMD__ROR_C_LOOP
CMD__ROR_C_LOOP_END:
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; PUSH <source>
; push
; places the value of <source> onto the stack
;-------------------------------------------------------------------------------------------
CMD__PUSH:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; place value onto stack
	lfsr	FSR0,CMD_STACK		; get base address of stack
	movf	CMD_STACK_PTR,W		; add stack pointer
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	A1_REG,INDF0		; write value to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	bcf	CMD_STACK_PTR,7		; (keep stack pointer in range 0..127)
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; PUSH_C <constant>
; push constant
; places <constant> onto the stack
;-------------------------------------------------------------------------------------------
CMD__PUSH_C:
; get constant
	movff	POSTINC1,A1_REG
; place value onto stack
	lfsr	FSR0,CMD_STACK		; get base address of stack
	movf	CMD_STACK_PTR,W		; add stack pointer
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	A1_REG,INDF0		; write value to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	bcf	CMD_STACK_PTR,7		; (keep stack pointer in range 0..127)
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; POP <dest>
; pop
; takes a value from the stack and places it into <dest>
;-------------------------------------------------------------------------------------------
CMD__POP:
; take value from stack
	decf	CMD_STACK_PTR		; decrement stack pointer
	bcf	CMD_STACK_PTR,7		; (keep stack pointer in range 0..127)
	lfsr	FSR0,CMD_STACK		; get base address of stack
	movf	CMD_STACK_PTR,W		; add stack pointer
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movf	INDF0,W			; get value from stack location
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; write value to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movwf	INDF0			; get dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; CALL (bit0 is bit8 of offset) <offset> <sectors (4)>
; call code at offset and <sectors> sectors further
; - first the return offset and sector are placed onto the stack
;   in the following order: ofs_h ofs_l sec_hh sec_hl sec_lh sec_ll
; - then a jump to the specified location is executed
; - <sectors (4)> is stored in big endian format
;-------------------------------------------------------------------------------------------
CMD__CALL:
; get return code location
	movlw	6			; advance offset to next command
	addwf	CMD_OFFSET+0
	movlw	0
	addwfc	CMD_OFFSET+1
	movff	CMD_SECTOR+0,A4_REG+0	; copy current absolute sector number
	movff	CMD_SECTOR+1,A4_REG+1
	movff	CMD_SECTOR+2,A4_REG+2
	movff	CMD_SECTOR+3,A4_REG+3
	movlw	0x01			; if offset >= 0x200, ...
	cpfsgt	CMD_OFFSET+1
	bra	CMD__CALL_SEC
	clrf	CMD_OFFSET+0		; offset := 0
	clrf	CMD_OFFSET+1
	incf	A4_REG+0		; sector++
	movlw	0
	addwfc	A4_REG+1
	addwfc	A4_REG+2
	addwfc	A4_REG+3
CMD__CALL_SEC:
; store offset behind command and absolute sector number onto stack
	lfsr	FSR0,CMD_STACK		; get base address of stack
	movf	CMD_STACK_PTR,W		; add stack pointer
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	CMD_OFFSET+1,POSTINC0	; write ofs_h to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_0
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_0:
	movff	CMD_OFFSET+0,POSTINC0	; write ofs_l to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_1
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_1:
	movff	A4_REG+3,POSTINC0	; write sec_hh to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_2
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_2:
	movff	A4_REG+2,POSTINC0	; write sec_hl to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_3
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_3:
	movff	A4_REG+1,POSTINC0	; write sec_lh to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_4
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_4:
	movff	A4_REG+0,INDF0		; write sec_ll to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	bcf	CMD_STACK_PTR,7		; (keep stack pointer in range 0..127)
; get offset
	movf	POSTDEC1,W		; go back to command byte
	movf	POSTINC1,W		; bit 0 is high byte of offset
	andlw	0x01
	movwf	CMD_OFFSET+1
	movff	POSTINC1,CMD_OFFSET+0	; low byte of offset
; get number of sectors
	movff	POSTINC1,A4_REG+3
	movff	POSTINC1,A4_REG+2
	movff	POSTINC1,A4_REG+1
	movff	POSTINC1,A4_REG
; add to sector number
	movf	A4_REG,w
	addwf	CMD_SECTOR
	movf	A4_REG+1,w
	addwfc	CMD_SECTOR+1
	movf	A4_REG+2,w
	addwfc	CMD_SECTOR+2
	movf	A4_REG+3,w
	addwfc	CMD_SECTOR+3
; load sector
	goto	CMD_LOAD_SECTOR

;-------------------------------------------------------------------------------------------
; CALL_IND <offset_h> <offset_l> <sectors_hh> <sectors_hl> <sectors_lh> <sectors_ll>
; call code at offset and <sectors> sectors further
; - first the return offset and sector are placed onto the stack
;   in the following order: ofs_h ofs_l sec_hh sec_hl sec_lh sec_ll
; - then a jump to the specified location is executed
;-------------------------------------------------------------------------------------------
CMD__CALL_IND:
; get return code location
	movlw	6			; advance offset to next command
	addwf	CMD_OFFSET+0
	movlw	0
	addwfc	CMD_OFFSET+1
	movff	CMD_SECTOR+0,A4_REG+0	; copy current absolute sector number
	movff	CMD_SECTOR+1,A4_REG+1
	movff	CMD_SECTOR+2,A4_REG+2
	movff	CMD_SECTOR+3,A4_REG+3
	movlw	0x01			; if offset >= 0x200, ...
	cpfsgt	CMD_OFFSET+1
	bra	CMD__CALL_IND_SEC
	clrf	CMD_OFFSET+0		; offset := 0
	clrf	CMD_OFFSET+1
	incf	A4_REG+0		; sector++
	movlw	0
	addwfc	A4_REG+1
	addwfc	A4_REG+2
	addwfc	A4_REG+3
CMD__CALL_IND_SEC:
; store offset behind command and absolute sector number onto stack
	lfsr	FSR0,CMD_STACK		; get base address of stack
	movf	CMD_STACK_PTR,W		; add stack pointer
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	CMD_OFFSET+1,POSTINC0	; write ofs_h to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_IND_0
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_IND_0:
	movff	CMD_OFFSET+0,POSTINC0	; write ofs_l to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_IND_1
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_IND_1:
	movff	A4_REG+3,POSTINC0	; write sec_hh to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_IND_2
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_IND_2:
	movff	A4_REG+2,POSTINC0	; write sec_hl to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_IND_3
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_IND_3:
	movff	A4_REG+1,POSTINC0	; write sec_lh to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_IND_4
	lfsr	FSR0,CMD_STACK		; ..., go back to base address of stack
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_IND_4:
	movff	A4_REG+0,INDF0		; write sec_ll to stack location
	incf	CMD_STACK_PTR		; increment stack pointer
	bcf	CMD_STACK_PTR,7		; (keep stack pointer in range 0..127)
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get offset
	movff	POSTINC1,FSR0L		; get offset of offset_h
	movf	INDF0,W			; get offset_h
	andlw	0x01			; (offset has only 9 bits)
	movwf	CMD_OFFSET+1
	movff	POSTINC1,FSR0L		; get offset of offset_l
	movff	INDF0,CMD_OFFSET+0	; get offset_l
; get number of sectors
	movff	POSTINC1,FSR0L		; get offset of sectors_hh
	movff	INDF0,A4_REG+3		; get sectors_hh
	movff	POSTINC1,FSR0L		; get offset of sectors_hl
	movff	INDF0,A4_REG+2		; get sectors_hl
	movff	POSTINC1,FSR0L		; get offset of sectors_lh
	movff	INDF0,A4_REG+1		; get sectors_lh
	movff	POSTINC1,FSR0L		; get offset of sectors_ll
	movff	INDF0,A4_REG		; get sectors_ll
; add to sector number
	movf	A4_REG,w
	addwf	CMD_SECTOR
	movf	A4_REG+1,w
	addwfc	CMD_SECTOR+1
	movf	A4_REG+2,w
	addwfc	CMD_SECTOR+2
	movf	A4_REG+3,w
	addwfc	CMD_SECTOR+3
; load sector
	goto	CMD_LOAD_SECTOR

;-------------------------------------------------------------------------------------------
; RET
; goto	CMD_FINISH from procedure
; - first the offset and the sector is taken from the stack
;   in the following order: sec_ll sec_lh sec_hl sec_hh ofs_l ofs_h
; - then a jump to this location is executed
;-------------------------------------------------------------------------------------------
CMD__RET:
; get offset and absolute sector number from stack
	decf	CMD_STACK_PTR		; decrement stack pointer
	bcf	CMD_STACK_PTR,7		; (keep stack pointer in range 0..127)
	lfsr	FSR0,CMD_STACK		; get base address of stack
	movf	CMD_STACK_PTR,W		; add stack pointer
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movff	POSTDEC0,CMD_SECTOR+0	; get sec_ll from stack location
	decf	CMD_STACK_PTR		; decrement stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_RET_0
	lfsr	FSR0,CMD_STACK+127	; ..., go back to end of stack (base address + 127)
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_RET_0:
	movff	POSTDEC0,CMD_SECTOR+1	; get sec_lh from stack location
	decf	CMD_STACK_PTR		; decrement stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_RET_1
	lfsr	FSR0,CMD_STACK+127	; ..., go back to end of stack (base address + 127)
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_RET_1:
	movff	POSTDEC0,CMD_SECTOR+2	; get sec_hl from stack location
	decf	CMD_STACK_PTR		; decrement stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_RET_2
	lfsr	FSR0,CMD_STACK+127	; ..., go back to end of stack (base address + 127)
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_RET_2:
	movff	POSTDEC0,CMD_SECTOR+3	; get sec_hh from stack location
	decf	CMD_STACK_PTR		; decrement stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_RET_3
	lfsr	FSR0,CMD_STACK+127	; ..., go back to end of stack (base address + 127)
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_RET_3:
	movff	POSTDEC0,CMD_OFFSET+0	; get ofs_l from stack location
	decf	CMD_STACK_PTR		; decrement stack pointer
	btfss	CMD_STACK_PTR,7		; if stack overflow, ...
	bra	CMD__CALL_RET_4
	lfsr	FSR0,CMD_STACK+127	; ..., go back to end of stack (base address + 127)
	bcf	CMD_STACK_PTR,7		; ..., keep stack pointer in range 0..127
CMD__CALL_RET_4:
	movf	INDF0,W			; get ofs_h from stack location
	andlw	0x01			; offset has only got 9 bits
	movwf	CMD_OFFSET+1
; load sector
	goto	CMD_LOAD_SECTOR

;-------------------------------------------------------------------------------------------
; RD_PIX <var_x> <var_y> <dest>
; read pixel
; write the value of the pixel at position var_x, var_y into dest
;-------------------------------------------------------------------------------------------
CMD__RD_PIX:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_x
	movff	POSTINC1,FSR0L		; get offset of var_x
	movff	INDF0,A2_REG		; get var_x
; get var_y
	movff	POSTINC1,FSR0L		; get offset of var_y
	movff	INDF0,A2_REG+1		; get var_y
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_RD_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_RD

;-------------------------------------------------------------------------------------------
; RD_PIX_C <con_x> <con_y> <dest>
; read constant pixel
; write the value of the pixel at position con_x, con_y into dest
;-------------------------------------------------------------------------------------------
CMD__RD_PIX_C:
; get con_x
	movff	POSTINC1,A2_REG
; get con_y
	movff	POSTINC1,A2_REG+1
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_RD_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_RD

;-------------------------------------------------------------------------------------------
; RD_OVP <var_x> <var_y> <dest>
; read overlay pixel
; write the value of the overlay pixel at position var_x, var_y into dest
;-------------------------------------------------------------------------------------------
CMD__RD_OVP:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_x
	movff	POSTINC1,FSR0L		; get offset of var_x
	movff	INDF0,A2_REG		; get var_x
; get var_y
	movff	POSTINC1,FSR0L		; get offset of var_y
	movff	INDF0,A2_REG+1		; get var_y
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_RD

;-------------------------------------------------------------------------------------------
; RD_OVP_C <con_x> <con_y> <dest>
; read constant overlay pixel
; write the value of the overlay pixel at position con_x, con_y into dest
;-------------------------------------------------------------------------------------------
CMD__RD_OVP_C:
; get con_x
	movff	POSTINC1,A2_REG
; get con_y
	movff	POSTINC1,A2_REG+1
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_RD

;-------------------------------------------------------------------------------------------
; second part of RD_xxx[_C]
; input:	A2_REG		x coordinate
;		A2_REG+1	y coordinate
;	 	FSR0		base address of pixel buffer
;-------------------------------------------------------------------------------------------
CMD_RD:
; default pixel value is 0x00
	clrf	A1_REG
; check range of coordinates
	movlw	26			; x < 26 ?
	cpfslt	A2_REG
	bra	CMD_RD_ERR
	movlw	20			; y < 20 ?
	cpfslt	A2_REG+1
	bra	CMD_RD_ERR
; get offset of pixel
	movlw	26			; PROD := 26 * y + x
	mulwf	A2_REG+1
	movf	A2_REG,W
	addwf	PRODL
	movlw	0
	addwfc	PRODH
	bcf	STATUS,C		; PROD := PROD / 2
	rrcf	PRODH
	rrcf	PRODL
; get address of pixel
	movf	PRODL,W			; FSR0 := FSR0 + PROD
	addwf	FSR0L
	movf	PRODH,W
	addwfc	FSR0H
; get value of pixel
	movf	INDF0,W
	btfss	A2_REG,0		; last bit of x coordinate not set: we want upper half of byte
	swapf	INDF0,W
	andlw	0x0F			; pixel value is only a nibble
	movwf	A1_REG
; store pixel value in dest
CMD_RD_ERR:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; write to dest
	movff	POSTINC1,FSR0L		; get offset of dest
	movff	A1_REG,INDF0		; write to dest
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; WR_PIX <var_x> <var_y> <source>
; write pixel
; write the value of source into the pixel at position var_x, var_y
;-------------------------------------------------------------------------------------------
CMD__WR_PIX:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_x
	movff	POSTINC1,FSR0L		; get offset of var_x
	movff	INDF0,A2_REG		; get var_x
; get var_y
	movff	POSTINC1,FSR0L		; get offset of var_y
	movff	INDF0,A2_REG+1		; get var_y
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	goto	CMD_FINISH
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR

;-------------------------------------------------------------------------------------------
; WR_PIX_C <con_x> <con_y> <source>
; write constant pixel
; write the value of source into the pixel at position con_x, con_y
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get con_x
	movff	POSTINC1,A2_REG
; get con_y
	movff	POSTINC1,A2_REG+1
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	goto	CMD_FINISH
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR

;-------------------------------------------------------------------------------------------
; WR_OVP <var_x> <var_y> <source>
; write overlay pixel
; write the value of source into the overlay pixel at position var_x, var_y
;-------------------------------------------------------------------------------------------
CMD__WR_OVP:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_x
	movff	POSTINC1,FSR0L		; get offset of var_x
	movff	INDF0,A2_REG		; get var_x
; get var_y
	movff	POSTINC1,FSR0L		; get offset of var_y
	movff	INDF0,A2_REG+1		; get var_y
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_WR

;-------------------------------------------------------------------------------------------
; WR_OVP_C <con_x> <con_y> <source>
; write constant overlay pixel
; write the value of source into the overlay pixel at position con_x, con_y
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get con_x
	movff	POSTINC1,A2_REG
; get con_y
	movff	POSTINC1,A2_REG+1
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_WR

;-------------------------------------------------------------------------------------------
; WR_PIX_CO <var_x> <var_y> <constant>
; set pixel to color
; write const_source into the pixel at position var_x, var_y
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_CO:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_x
	movff	POSTINC1,FSR0L		; get offset of var_x
	movff	INDF0,A2_REG		; get var_x
; get var_y
	movff	POSTINC1,FSR0L		; get offset of var_y
	movff	INDF0,A2_REG+1		; get var_y
; get constant
	movff	POSTINC1,A1_REG
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	goto	CMD_FINISH
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR

;-------------------------------------------------------------------------------------------
; WR_PIX_CO_C <con_x> <con_y> <constant>
; set constant pixel to color
; write constant into the pixel at position con_x, con_y
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_CO_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get con_x
	movff	POSTINC1,A2_REG
; get con_y
	movff	POSTINC1,A2_REG+1
; get constant
	movff	POSTINC1,A1_REG
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	goto	CMD_FINISH
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR

;-------------------------------------------------------------------------------------------
; WR_OVP_CO <var_x> <var_y> <constant>
; set overlay pixel to color
; write constant into the overlay pixel at position var_x, var_y
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_CO:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_x
	movff	POSTINC1,FSR0L		; get offset of var_x
	movff	INDF0,A2_REG		; get var_x
; get var_y
	movff	POSTINC1,FSR0L		; get offset of var_y
	movff	INDF0,A2_REG+1		; get var_y
; get constant
	movff	POSTINC1,A1_REG
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_WR

;-------------------------------------------------------------------------------------------
; WR_OVP_CO_C <con_x> <con_y> <constant>
; set constant overlay pixel to color
; write constant into the overlay pixel at position con_x, con_y
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_CO_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get con_x
	movff	POSTINC1,A2_REG
; get con_y
	movff	POSTINC1,A2_REG+1
; get constant
	movff	POSTINC1,A1_REG
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_WR

;-------------------------------------------------------------------------------------------
; second part of WR_xxx[_C]
; input:	A2_REG		x coordinate
;		A2_REG+1	y coordinate
;		A1_REG		pixel value
; 		FSR0		base address of pixel buffer
;-------------------------------------------------------------------------------------------
CMD_WR:
; check range of coordinates
	movlw	26			; x < 26 ?
	cpfslt	A2_REG
	goto	CMD_FINISH
	movlw	20			; y < 20 ?
	cpfslt	A2_REG+1
	goto	CMD_FINISH
; get offset of pixel
	movlw	26			; PROD := 26 * y + x
	mulwf	A2_REG+1
	movf	A2_REG,W
	addwf	PRODL
	movlw	0
	addwfc	PRODH
	bcf	STATUS,C		; PROD := PROD / 2
	rrcf	PRODH
	rrcf	PRODL
; get address of first pixel
	movf	PRODL,W			; FSR0 := FSR0 + PROD
	addwf	FSR0L
	movf	PRODH,W
	addwfc	FSR0H
; even or odd pixel?
	btfss	A2_REG,0		; last bit of x coordinate not set: upper half of byte
	bra	CMD_WR_ODD
; set value of even overlay pixel
	movlw	0xF0			; clear pixel
	andwf	INDF0
	movf	A1_REG,W		; get new pixel value
	andlw	0x0F
	iorwf	INDF0			; write new pixel value
; done
	goto	CMD_FINISH
; set value of odd pixel
CMD_WR_ODD:
	movlw	0x0F			; clear pixel
	andwf	INDF0
	swapf	A1_REG,W		; get new pixel value
	andlw	0xF0
	iorwf	INDF0			; write new pixel value
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; WR_PIX_LINE <var_y> <pixel_0_1> <pixel_2_3> ... <pixel_24_25>
; write pixel line
; two pixels are packed in one byte - e.g. pixel_0_1=0xF0 turns on pixel 0 and off pixel 1
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_LINE:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_y
	movff	POSTINC1,FSR0L		; get offset of var_y
	movff	INDF0,A2_REG+1		; get var_y
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_WR_LINE_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_LINE

;-------------------------------------------------------------------------------------------
; WR_PIX_LINE_C <con_y> <pixel_0_1> <pixel_2_3> ... <pixel_24_25>
; write constant pixel line
; two pixels are packed in one byte - e.g. pixel_0_1=0xF0 turns on pixel 0 and off pixel 1
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_LINE_C:
; get con_y
	movff	POSTINC1,A2_REG+1
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_WR_LINE_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_LINE

;-------------------------------------------------------------------------------------------
; WR_OVP_LINE <var_y> <ov-pixel_0_1> <ov-pixel_2_3> ... <ov-pixel_24_25>
; write overlay pixel line
; two overlay pixels are packed in one byte - e.g. overlay pixel_0_1=0xF0 turns on overlay pixel 0 and off overlay pixel 1
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_LINE:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_y
	movff	POSTINC1,FSR0L		; get offset of var_y
	movff	INDF0,A2_REG+1		; get var_y
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_WR_LINE

;-------------------------------------------------------------------------------------------
; WR_OVP_LINE_C <con_y> <ov-pixel_0_1> <ov-pixel_2_3> ... <ov-pixel_24_25>
; write constant overlay pixel line
; two overlay pixels are packed in one byte - e.g. overlay pixel_0_1=0xF0 turns on overlay pixel 0 and off overlay pixel 1
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_LINE_C:
; get con_y
	movff	POSTINC1,A2_REG+1
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_WR_LINE

;-------------------------------------------------------------------------------------------
; second part of WR_xxx_LINE[_C]
; input:	A2_REG+1	y coordinate
;		FSR0		base address of pixel buffer
;-------------------------------------------------------------------------------------------
CMD_WR_LINE:
; check range of coordinates
	movlw	20			; y < 20 ?
	cpfslt	A2_REG+1
	bra	CMD_WR_LINE_ERR
; get offset of overlay pixel
	movlw	13			; PROD := 13 * y
	mulwf	A2_REG+1
; get address of first pixel
	movf	PRODL,W			; FSR0 := FSR0 + PROD
	addwf	FSR0L
	movf	PRODH,W
	addwfc	FSR0H
; set value of next 26 pixels (13 bytes)
	movff	POSTINC1,POSTINC0	; byte 0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0	; byte 5
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0	; byte 10
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0	; byte 12
; done
	goto	CMD_FINISH
CMD_WR_LINE_ERR:
; jump over parameters of command (13 bytes)
	movlw	13
	addwf	FSR1L
	movlw	0
	addwfc	FSR1H
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; WR_PIX_QUART <var_quart> <pixel_l0_0_1> <pixel_l0_2_3> ... <pixel_l4_24_25>
; write pixel quarter
; two pixels are packed in one byte
; - e.g. pixel_l0_0_1=0xF0 turns on line 0, pixel 0 and off pixel 1 (lines rel. to quarter)
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_QUART:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_quart
	movff	POSTINC1,FSR0L		; get offset of var_quart
	movff	INDF0,A1_REG		; get var_quart
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_WR_QUART_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_QUART

;-------------------------------------------------------------------------------------------
; WR_PIX_QUART_C <con_quart> <pixel_l0_0_1> <pixel_l0_2_3> ... <pixel_l4_24_25>
; write constant pixel quarter
; two pixels are packed in one byte
; - e.g. pixel_l0_0_1=0xF0 turns on line 0, pixel 0 and off pixel 1 (lines rel. to quarter)
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_QUART_C:
; get con_quart
	movff	POSTINC1,A1_REG
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_WR_QUART_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_QUART

;-------------------------------------------------------------------------------------------
; WR_OVP_QUART <var_quart> <ov-pixel_l0_0_1> <ov-pixel_l0_2_3> ... <ov-pixel_l4_24_25>
; write overlay pixel quarter
; two overlay pixels are packed in one byte
; - e.g. overlay pixel_l0_0_1=0xF0 turns on line 0, overlay pixel 0 and off overlay pixel 1 (lines rel. to quarter)
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_QUART:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get var_quart
	movff	POSTINC1,FSR0L		; get offset of var_quart
	movff	INDF0,A1_REG		; get var_quart
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_WR_QUART

;-------------------------------------------------------------------------------------------
; WR_OVP_QUART_C <con_quart> <ov-pixel_l0_0_1> <ov-pixel_l0_2_3> ... <ov-pixel_l4_24_25>
; write constant overlay pixel quarter
; two overlay pixels are packed in one byte
; - e.g. overlay pixel_l0_0_1=0xF0 turns on line 0, overlay pixel 0 and off overlay pixel 1
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_QUART_C:
; get con_quart
	movff	POSTINC1,A1_REG
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY	; FSR0 := base address
; jump to second part
	bra	CMD_WR_QUART

;-------------------------------------------------------------------------------------------
; second part of WR_xxx_QUART[_C]
; input:	A1_REG		number of quarter
;		FSR0		base address of pixel buffer
;-------------------------------------------------------------------------------------------
CMD_WR_QUART:
; check range of coordinates
	movlw	4			; quart < 4 ?
	cpfslt	A1_REG
	bra	CMD_WR_QUART_ERR
; get offset of overlay pixel
	movlw	65			; PROD := 65 * quart
	mulwf	A1_REG
; get address of first pixel
	movf	PRODL,W			; FSR0 := FSR0 + PROD
	addwf	FSR0L
	movf	PRODH,W
	addwfc	FSR0H
; set value of next 130 pixels (65 bytes)
	movlw	4			; write 4 times 16 byte
	movwf	A1_REG
CMD_WR_QUART_LOOP:
	movff	POSTINC1,POSTINC0	; byte 0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0	; byte 5
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0	; byte 10
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0
	movff	POSTINC1,POSTINC0	; byte 15
	decfsz	A1_REG
	bra	CMD_WR_QUART_LOOP
	movff	POSTINC1,POSTINC0	; last byte
; done
	goto	CMD_FINISH
CMD_WR_QUART_ERR:
; jump over parameters of command (65 bytes)
	movlw	65
	addwf	FSR1L
	movlw	0
	addwfc	FSR1H
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; WR_PIX_CLR
; clear pixels
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_CLR:
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	goto	CMD_FINISH
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_CLR

;-------------------------------------------------------------------------------------------
; WR_OVP_CLR
; clear overlay pixels
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_CLR:
; get address overlay buffer
	lfsr	FSR0,CMD_OVERLAY
; jump to second part
	bra	CMD_WR_CLR

;-------------------------------------------------------------------------------------------
; second part of WR_xxx_CLR
; input:	FSR0		address of pixel buffer
;-------------------------------------------------------------------------------------------
CMD_WR_CLR:
; clear 260 bytes
	movlw	16			; clear 16 times 16 bytes
	movwf	A1_REG
CMD_WR_CLR_LOOP:
	clrf	POSTINC0		; byte 0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0		; byte 5
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0		; byte 10
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0		; byte 15
	decfsz	A1_REG
	bra	CMD_WR_CLR_LOOP
	clrf	POSTINC0		; clear last 4 bytes
	clrf	POSTINC0
	clrf	POSTINC0
	clrf	POSTINC0
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; WR_PIX_SET
; set pixels
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_SET:
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	goto	CMD_FINISH
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_SET

;-------------------------------------------------------------------------------------------
; WR_OVP_SET
; set overlay pixels
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_SET:
; get address overlay buffer
	lfsr	FSR0,CMD_OVERLAY
; jump to second part
	bra	CMD_WR_SET

;-------------------------------------------------------------------------------------------
; second part of WR_xxx_SET
; input:	FSR0		address of pixel buffer
;-------------------------------------------------------------------------------------------
CMD_WR_SET:
; set 260 bytes
	movlw	16			; set 16 times 16 bytes
	movwf	A1_REG
CMD_WR_SET_LOOP:
	setf	POSTINC0		; byte 0
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0		; byte 5
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0		; byte 10
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0		; byte 15
	decfsz	A1_REG
	bra	CMD_WR_SET_LOOP
	setf	POSTINC0		; set last 4 bytes
	setf	POSTINC0
	setf	POSTINC0
	setf	POSTINC0
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; WR_PIX_BW <bw0> ... <bw64>
; write frame of black/white pixels
; bw0,7 = pixel 0,0 / bw0,6 = pixel 1,0 / ... / bw64,0 = pixel 25,19
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_BW:
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_WR_BW_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_BW

;-------------------------------------------------------------------------------------------
; WR_OVP_BW <bw0> ... <bw64>
; write frame of black/white overlay pixels
; bw0,7 = pixel 0,0 / bw0,6 = pixel 1,0 / ... / bw64,0 = pixel 25,19
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_BW:
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY
; jump to second part
	bra	CMD_WR_BW

;-------------------------------------------------------------------------------------------
; second part of WR_xxx_BW
; input:	FSR0		address of pixel buffer
;-------------------------------------------------------------------------------------------
CMD_WR_BW:
; write pixels
	movlw	65			; 65 times 8 pixels
	movwf	A1_REG
CMD_WR_BW_LOOP:
	movlw	0x00			; bits 7, 6 of byte
	btfsc	INDF1,7
	iorlw	0xF0
	btfsc	INDF1,6
	iorlw	0x0F
	movwf	POSTINC0
	movlw	0x00			; bits 5, 4 of byte
	btfsc	INDF1,5
	iorlw	0xF0
	btfsc	INDF1,4
	iorlw	0x0F
	movwf	POSTINC0
	movlw	0x00			; bits 3, 2 of byte
	btfsc	INDF1,3
	iorlw	0xF0
	btfsc	INDF1,2
	iorlw	0x0F
	movwf	POSTINC0
	movlw	0x00			; bits 1, 0 of byte
	btfsc	INDF1,1
	iorlw	0xF0
	btfsc	POSTINC1,0		; move to next byte
	iorlw	0x0F
	movwf	POSTINC0
	decfsz	A1_REG
	bra	CMD_WR_BW_LOOP
; done
	goto	CMD_FINISH
CMD_WR_BW_ERR:
; jump over parameters of command (65 bytes)
	movlw	65
	addwf	FSR1L
	movlw	0
	addwfc	FSR1H
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; WR_PIX_SEL <source> <bw0> ... <bw64>
; set selected pixels of frame to source
; - bw selects pixels
; - pixels with bw=1 are set to source
; bw0,7 = pixel 0,0 / bw0,6 = pixel 1,0 / ... / bw64,0 = pixel 25,19
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_SEL:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_WR_SEL_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_SEL

;-------------------------------------------------------------------------------------------
; WR_PIX_SEL_C <constant> <bw0> ... <bw64>
; set selected pixels of frame to constant
; - bw selects pixels
; - pixels with bw=1 are set to constant
; bw0,7 = pixel 0,0 / bw0,6 = pixel 1,0 / ... / bw64,0 = pixel 25,19
;-------------------------------------------------------------------------------------------
CMD__WR_PIX_SEL_C:
; get constant
	movff	POSTINC1,A1_REG
; get address of output buffer not owned by interrupt
	btfsc	BIT_O_BUF_SW		; error if output buffer switch is in progress
	bra	CMD_WR_SEL_ERR
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
; jump to second part
	bra	CMD_WR_SEL

;-------------------------------------------------------------------------------------------
; WR_OVP_SEL <source> <bw0> ... <bw64>
; set selected ovaerlay pixels of frame to source
; - bw selects pixels
; - pixels with bw=1 are set to source
; bw0,0 = pixel 0,0 / bw0,1 = pixel 1,0 / ... / bw64,7 = pixel 25,19
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_SEL:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get source
	movff	POSTINC1,FSR0L		; get offset of source
	movff	INDF0,A1_REG		; get source
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY
; jump to second part
	bra	CMD_WR_SEL

;-------------------------------------------------------------------------------------------
; WR_OVP_SEL_C <constant> <bw0> ... <bw64>
; set selected overlay pixels of frame to constant
; - bw selects pixels
; - pixels with bw=1 are set to constant
; bw0,0 = pixel 0,0 / bw0,1 = pixel 1,0 / ... / bw64,7 = pixel 25,19
;-------------------------------------------------------------------------------------------
CMD__WR_OVP_SEL_C:
; get constant
	movff	POSTINC1,A1_REG
; get address of overlay buffer
	lfsr	FSR0,CMD_OVERLAY
; jump to second part
	bra	CMD_WR_SEL

;-------------------------------------------------------------------------------------------
; second part of WR_xxx_SEL[_C]
; input:	A1_REG		requested pixel value
;		FSR0		address of pixel buffer
;-------------------------------------------------------------------------------------------
CMD_WR_SEL:
; get pixel value for even and odd pixels
	swapf	A1_REG,W		; even pixels: value into upper nibble
	andlw	0xF0
	movwf	A2_REG
	movf	A1_REG,W		; odd pixels: value into lower nibble
	andlw	0x0F
	movwf	A2_REG+1
; write pixels
	movlw	65			; 65 times 8 pixels
	movwf	A1_REG
CMD_WR_SEL_LOOP:
	movf	INDF0,W			; bits 7, 6 of byte
	btfss	INDF1,7
	bra	CMD_WR_SEL_7
	andlw	0x0F
	iorwf	A2_REG,W
CMD_WR_SEL_7:
	btfss	INDF1,6
	bra	CMD_WR_SEL_6
	andlw	0xF0
	iorwf	A2_REG+1,W
CMD_WR_SEL_6:
	movwf	POSTINC0
	movf	INDF0,W			; bits 5, 4 of byte
	btfss	INDF1,5
	bra	CMD_WR_SEL_5
	andlw	0x0F
	iorwf	A2_REG,W
CMD_WR_SEL_5:
	btfss	INDF1,4
	bra	CMD_WR_SEL_4
	andlw	0xF0
	iorwf	A2_REG+1,W
CMD_WR_SEL_4:
	movwf	POSTINC0
	movf	INDF0,W			; bits 3, 2 of byte
	btfss	INDF1,3
	bra	CMD_WR_SEL_3
	andlw	0x0F
	iorwf	A2_REG,W
CMD_WR_SEL_3:
	btfss	INDF1,2
	bra	CMD_WR_SEL_2
	andlw	0xF0
	iorwf	A2_REG+1,W
CMD_WR_SEL_2:
	movwf	POSTINC0
	movf	INDF0,W			; bits 1, 0 of byte
	btfss	INDF1,1
	bra	CMD_WR_SEL_1
	andlw	0x0F
	iorwf	A2_REG,W
CMD_WR_SEL_1:
	btfss	POSTINC1,0
	bra	CMD_WR_SEL_0
	andlw	0xF0
	iorwf	A2_REG+1,W
CMD_WR_SEL_0:
	movwf	POSTINC0
	decfsz	A1_REG			; loop
	bra	CMD_WR_SEL_LOOP
; done
	goto	CMD_FINISH
CMD_WR_SEL_ERR:
; jump over parameters of command (65 bytes)
	movlw	65
	addwf	FSR1L
	movlw	0
	addwfc	FSR1H
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; CP_BLK <flags> <src_x> <src_y> <dest_x> <dest_y> <size_x> <size_y>
; copy a block of pixels
; flags,7: 0=copy from pixel buffer, 1=copy from overlay buffer
; flags,6: 0=copy to pixel buffer, 1=copy to overlay buffer
; flags,5: 0=src_x is a variable, 1=src_x is a constant
; flags,4: 0=src_y is a variable, 1=src_y is a constant
; flags,3: 0=dest_x is a variable, 1=dest_x is a constant
; flags,2: 0=dest_y is a variable, 1=dest_y is a constant
; flags,1: 0=size_x is a variable, 1=size_x is a constant
; flags,0: 0=size_y is a variable, 1=size_y is a constant
; if the two sections overlap, the result is undefined
;-------------------------------------------------------------------------------------------
CMD__CP_BLK:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get flags
	movff	POSTINC1,A1_REG
; no transparent color
	setf	B1_REG
; jump to second part
	bra	CMD_CP_BLK

;-------------------------------------------------------------------------------------------
; CP_BLK_TR <flags> <transp> <src_x> <src_y> <dest_x> <dest_y> <size_x> <size_y>
; copy a block of pixels with a transparent color
;  - if the variable transp is not in the range 0x00..0x0F no color is transparent
; flags,7: 0=copy from pixel buffer, 1=copy from overlay buffer
; flags,6: 0=copy to pixel buffer, 1=copy to overlay buffer
; flags,5: 0=src_x is a variable, 1=src_x is a constant
; flags,4: 0=src_y is a variable, 1=src_y is a constant
; flags,3: 0=dest_x is a variable, 1=dest_x is a constant
; flags,2: 0=dest_y is a variable, 1=dest_y is a constant
; flags,1: 0=size_x is a variable, 1=size_x is a constant
; flags,0: 0=size_y is a variable, 1=size_y is a constant
; if the two sections overlap, the result is undefined
;-------------------------------------------------------------------------------------------
CMD__CP_BLK_TR:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get flags
	movff	POSTINC1,A1_REG
; get transparent color
	movff	POSTINC1,FSR0L		; get offset of transp
	movff	INDF0,B1_REG		; get transp
; jump to second part
	bra	CMD_CP_BLK

;-------------------------------------------------------------------------------------------
; CP_BLK_TR_C <flags> <transp> <src_x> <src_y> <dest_x> <dest_y> <size_x> <size_y>
; copy a block of pixels with a constant transparent color
;  - if the constant transp is not in the range 0x00..0x0F no color is transparent
; flags,7: 0=copy from pixel buffer, 1=copy from overlay buffer
; flags,6: 0=copy to pixel buffer, 1=copy to overlay buffer
; flags,5: 0=src_x is a variable, 1=src_x is a constant
; flags,4: 0=src_y is a variable, 1=src_y is a constant
; flags,3: 0=dest_x is a variable, 1=dest_x is a constant
; flags,2: 0=dest_y is a variable, 1=dest_y is a constant
; flags,1: 0=size_x is a variable, 1=size_x is a constant
; flags,0: 0=size_y is a variable, 1=size_y is a constant
; if the two sections overlap, the result is undefined
;-------------------------------------------------------------------------------------------
CMD__CP_BLK_TR_C:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get flags
	movff	POSTINC1,A1_REG
; get transparent color
	movff	POSTINC1,B1_REG
; jump to second part
	bra	CMD_CP_BLK

;-------------------------------------------------------------------------------------------
; second part of CP_BLK[_TR[_C]]
; input:	FSR0		pointer to CMD_VARS
;		A1_REG		flags
;		B1_REG		transparent color or >= 0x10 if none
;-------------------------------------------------------------------------------------------
CMD_CP_BLK:
; get rest of parameters: src_x/y, dest_x/y, size_x/y
	movf	POSTINC1,W		; src_x into A2_REG
	movwf	FSR0L
	btfss	A1_REG,5
	movf	INDF0,W
	movwf	A2_REG
	movf	POSTINC1,W		; src_y into A2_REG+1
	movwf	FSR0L
	btfss	A1_REG,4
	movf	INDF0,W
	movwf	A2_REG+1
	movf	POSTINC1,W		; dest_x into B2_REG
	movwf	FSR0L
	btfss	A1_REG,3
	movf	INDF0,W
	movwf	B2_REG
	movf	POSTINC1,W		; dest_y into B2_REG+1
	movwf	FSR0L
	btfss	A1_REG,2
	movf	INDF0,W
	movwf	B2_REG+1
	movf	POSTINC1,W		; size_x into C2_REG
	movwf	FSR0L
	btfss	A1_REG,1
	movf	INDF0,W
	movwf	C2_REG
	movf	POSTINC1,W		; size_y into C2_REG+1
	movwf	FSR0L
	btfss	A1_REG,0
	movf	INDF0,W
	movwf	C2_REG+1
; check coordinates
	movf	C2_REG,W		; size_x > 0
	btfsc	STATUS,Z
	goto	CMD_FINISH
	movf	C2_REG+1,W		; size_y > 0
	btfsc	STATUS,Z
	goto	CMD_FINISH
	movf	A2_REG,W		; src_x + size_x <= 26
	addwf	C2_REG,W
	btfsc	STATUS,C		; overflow -> coordinate error
	goto	CMD_FINISH
	sublw	26			; sum too large -> coordinate error
	btfss	STATUS,C
	goto	CMD_FINISH
	movf	A2_REG+1,W		; src_y + size_y <= 20
	addwf	C2_REG+1,W
	btfsc	STATUS,C		; overflow -> coordinate error
	goto	CMD_FINISH
	sublw	20			; sum too large -> coordinate error
	btfss	STATUS,C
	goto	CMD_FINISH
	movf	B2_REG,W		; dest_x + size_x <= 26
	addwf	C2_REG,W
	btfsc	STATUS,C		; overflow -> coordinate error
	goto	CMD_FINISH
	sublw	26			; sum too large -> coordinate error
	btfss	STATUS,C
	goto	CMD_FINISH
	movf	B2_REG+1,W		; dest_y + size_y <= 20
	addwf	C2_REG+1,W
	btfsc	STATUS,C		; overflow -> coordinate error
	goto	CMD_FINISH
	sublw	20			; sum too large -> coordinate error
	btfss	STATUS,C
	goto	CMD_FINISH
; save offset part of command pointer
; - value in FSR1 will be destroyed
	movlw	low BUFFER
	subwf	FSR1L,W
	movwf	CMD_OFFSET+0
	movlw	high BUFFER
	subwfb	FSR1H,W
	movwf	CMD_OFFSET+1
; get address of src and dest buffer
	btfsc	BIT_O_BUF_SW		; get address of output buffer not owned by interrupt
	goto	CMD_FINISH		; error if output buffer switch is in progress
	lfsr	FSR0,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR0,O_FRAME1
	movff	FSR0L,FSR1L		; copy address into FSR1
	movff	FSR0H,FSR1H
	btfsc	A1_REG,7		; perhaps put address of overlay buffer into src pointer
	lfsr	FSR0,CMD_OVERLAY
	btfsc	A1_REG,6		; perhaps put address of overlay buffer into dest pointer
	lfsr	FSR1,CMD_OVERLAY
; get pointers to first src and dest line
	movlw	13			; src_offset = src_y * 13
	mulwf	A2_REG+1
	movf	PRODL,W			; add to src pointer
	addwf	FSR0L
	movf	PRODH,W
	addwfc	FSR0H
	movlw	13			; dest_offset = dest_y * 13
	mulwf	B2_REG+1
	movf	PRODL,W			; add to dest pointer
	addwf	FSR1L
	movf	PRODH,W
	addwfc	FSR1H
; get pixel masks
	clrf	A4_REG+0		; clear src mask
	clrf	A4_REG+1
	clrf	A4_REG+2
	clrf	A4_REG+3
	clrf	B4_REG+0		; clear dest mask
	clrf	B4_REG+1
	clrf	B4_REG+2
	clrf	B4_REG+3
CMD_CP_BLK_MSK1:
	bsf	STATUS,C		; shift a 1 into src mask
	rlcf	A4_REG+0
	rlcf	A4_REG+1
	rlcf	A4_REG+2
	rlcf	A4_REG+3
	bsf	STATUS,C		; shift a 1 into dest mask
	rlcf	B4_REG+0
	rlcf	B4_REG+1
	rlcf	B4_REG+2
	rlcf	B4_REG+3
	decfsz	C2_REG			; loop size_x times
	bra	CMD_CP_BLK_MSK1
CMD_CP_BLK_MSK2:
	movf	A2_REG			; do not shift 0s if src_x == 0
	bz	CMD_CP_BLK_MSK_2END
	bcf	STATUS,C		; shift a 0 into src mask
	rlcf	A4_REG+0
	rlcf	A4_REG+1
	rlcf	A4_REG+2
	rlcf	A4_REG+3
	decfsz	A2_REG			; loop src_x times
	bra	CMD_CP_BLK_MSK2
CMD_CP_BLK_MSK_2END:
CMD_CP_BLK_MSK3:
	movf	B2_REG			; do not shift 0s if dest_x == 0
	bz	CMD_CP_BLK_MSK_3END
	bcf	STATUS,C		; shift a 0 into dest mask
	rlcf	B4_REG+0
	rlcf	B4_REG+1
	rlcf	B4_REG+2
	rlcf	B4_REG+3
	decfsz	B2_REG			; loop dest_x times
	bra	CMD_CP_BLK_MSK3
CMD_CP_BLK_MSK_3END:
; save dest pointer in D2_REG
CMD_CP_BLK_LOOP:
	movff	FSR1L,D2_REG
	movff	FSR1H,D2_REG+1
; read line
	lfsr	FSR1,CMD_LINE		; get address of line buffer
	swapf	INDF0,W			; pixel 0
	btfsc	A4_REG+0,0
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 1
	btfsc	A4_REG+0,1
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 2
	btfsc	A4_REG+0,2
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 3
	btfsc	A4_REG+0,3
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 4
	btfsc	A4_REG+0,4
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 5
	btfsc	A4_REG+0,5
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 6
	btfsc	A4_REG+0,6
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 7
	btfsc	A4_REG+0,7
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 8
	btfsc	A4_REG+1,0
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 9
	btfsc	A4_REG+1,1
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 10
	btfsc	A4_REG+1,2
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 11
	btfsc	A4_REG+1,3
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 12
	btfsc	A4_REG+1,4
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 13
	btfsc	A4_REG+1,5
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 14
	btfsc	A4_REG+1,6
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 15
	btfsc	A4_REG+1,7
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 16
	btfsc	A4_REG+2,0
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 17
	btfsc	A4_REG+2,1
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 18
	btfsc	A4_REG+2,2
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 19
	btfsc	A4_REG+2,3
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 20
	btfsc	A4_REG+2,4
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 21
	btfsc	A4_REG+2,5
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 22
	btfsc	A4_REG+2,6
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 23
	btfsc	A4_REG+2,7
	movwf	POSTINC1
	swapf	INDF0,W			; pixel 24
	btfsc	A4_REG+3,0
	movwf	POSTINC1
	movf	POSTINC0,W		; pixel 25
	btfsc	A4_REG+3,1
	movwf	POSTINC1
; zero high nibbles of line buffer
	movlw	0x0F			; load mask
	lfsr	FSR1,CMD_LINE		; get address of line buffer
	andwf	POSTINC1		; pixel 0
	andwf	POSTINC1		; pixel 1
	andwf	POSTINC1		; pixel 2
	andwf	POSTINC1		; pixel 3
	andwf	POSTINC1		; pixel 4
	andwf	POSTINC1		; pixel 5
	andwf	POSTINC1		; pixel 6
	andwf	POSTINC1		; pixel 7
	andwf	POSTINC1		; pixel 8
	andwf	POSTINC1		; pixel 9
	andwf	POSTINC1		; pixel 10
	andwf	POSTINC1		; pixel 11
	andwf	POSTINC1		; pixel 12
	andwf	POSTINC1		; pixel 13
	andwf	POSTINC1		; pixel 14
	andwf	POSTINC1		; pixel 15
	andwf	POSTINC1		; pixel 16
	andwf	POSTINC1		; pixel 17
	andwf	POSTINC1		; pixel 18
	andwf	POSTINC1		; pixel 19
	andwf	POSTINC1		; pixel 20
	andwf	POSTINC1		; pixel 21
	andwf	POSTINC1		; pixel 22
	andwf	POSTINC1		; pixel 23
	andwf	POSTINC1		; pixel 24
	andwf	POSTINC1		; pixel 25
; calculate dest mask for this line
	movff	B4_REG+0,C4_REG+0	; use full dest mask by default
	movff	B4_REG+1,C4_REG+1
	movff	B4_REG+2,C4_REG+2
	movff	B4_REG+3,C4_REG+3
	movlw	0x10			; skip rest of this if no transparent color
	cpfslt	B1_REG
	bra	CMD_CP_BLK_NO_TR
	lfsr	FSR1,CMD_LINE		; get address of line buffer
	movf	POSTINC0,W		; pixel 0 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+0,0
	movf	POSTINC0,W		; pixel 1 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+0,1
	movf	POSTINC0,W		; pixel 2 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+0,2
	movf	POSTINC0,W		; pixel 3 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+0,3
	movf	POSTINC0,W		; pixel 4 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+0,4
	movf	POSTINC0,W		; pixel 5 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+0,5
	movf	POSTINC0,W		; pixel 6 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+0,6
	movf	POSTINC0,W		; pixel 7 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+0,7
	movf	POSTINC0,W		; pixel 8 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+1,0
	movf	POSTINC0,W		; pixel 9 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+1,1
	movf	POSTINC0,W		; pixel 10 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+1,2
	movf	POSTINC0,W		; pixel 11 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+1,3
	movf	POSTINC0,W		; pixel 12 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+1,4
	movf	POSTINC0,W		; pixel 13 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+1,5
	movf	POSTINC0,W		; pixel 14 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+1,6
	movf	POSTINC0,W		; pixel 15 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+1,7
	movf	POSTINC0,W		; pixel 16 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+2,0
	movf	POSTINC0,W		; pixel 17 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+2,1
	movf	POSTINC0,W		; pixel 18 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+2,2
	movf	POSTINC0,W		; pixel 19 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+2,3
	movf	POSTINC0,W		; pixel 20 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+2,4
	movf	POSTINC0,W		; pixel 21 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+2,5
	movf	POSTINC0,W		; pixel 22 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+2,6
	movf	POSTINC0,W		; pixel 23 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+2,7
	movf	POSTINC0,W		; pixel 24 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+3,0
	movf	POSTINC0,W		; pixel 25 transparent
	xorwf	B1_REG,W
	btfsc	STATUS,Z
	bcf	C4_REG+3,1
CMD_CP_BLK_NO_TR:
; restore dest pointer from D2_REG
	movff	D2_REG,FSR1L
	movff	D2_REG+1,FSR1H
; save src pointer in D2_REG
	movff	FSR0L,D2_REG
	movff	FSR0H,D2_REG+1
; write line
	lfsr	FSR0,CMD_LINE		; get address of line buffer
	btfss	C4_REG+0,0		; pixel 0
	bra	CMD_CP_BLK_W0
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W0:
	btfss	C4_REG+0,1		; pixel 1
	bra	CMD_CP_BLK_W1
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W1:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+0,2		; pixel 2
	bra	CMD_CP_BLK_W2
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W2:
	btfss	C4_REG+0,3		; pixel 3
	bra	CMD_CP_BLK_W3
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W3:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+0,4		; pixel 4
	bra	CMD_CP_BLK_W4
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W4:
	btfss	C4_REG+0,5		; pixel 5
	bra	CMD_CP_BLK_W5
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W5:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+0,6		; pixel 6
	bra	CMD_CP_BLK_W6
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W6:
	btfss	C4_REG+0,7		; pixel 7
	bra	CMD_CP_BLK_W7
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W7:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+1,0		; pixel 8
	bra	CMD_CP_BLK_W8
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W8:
	btfss	C4_REG+1,1		; pixel 9
	bra	CMD_CP_BLK_W9
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W9:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+1,2		; pixel 10
	bra	CMD_CP_BLK_W10
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W10:
	btfss	C4_REG+1,3		; pixel 11
	bra	CMD_CP_BLK_W11
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W11:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+1,4		; pixel 12
	bra	CMD_CP_BLK_W12
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W12:
	btfss	C4_REG+1,5		; pixel 13
	bra	CMD_CP_BLK_W13
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W13:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+1,6		; pixel 14
	bra	CMD_CP_BLK_W14
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W14:
	btfss	C4_REG+1,7		; pixel 15
	bra	CMD_CP_BLK_W15
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W15:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+2,0		; pixel 16
	bra	CMD_CP_BLK_W16
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W16:
	btfss	C4_REG+2,1		; pixel 17
	bra	CMD_CP_BLK_W17
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W17:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+2,2		; pixel 18
	bra	CMD_CP_BLK_W18
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W18:
	btfss	C4_REG+2,3		; pixel 19
	bra	CMD_CP_BLK_W19
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W19:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+2,4		; pixel 20
	bra	CMD_CP_BLK_W20
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W20:
	btfss	C4_REG+2,5		; pixel 21
	bra	CMD_CP_BLK_W21
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W21:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+2,6		; pixel 22
	bra	CMD_CP_BLK_W22
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W22:
	btfss	C4_REG+2,7		; pixel 23
	bra	CMD_CP_BLK_W23
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W23:
	movf	POSTINC1		; next address in dest buffer
	btfss	C4_REG+3,0		; pixel 24
	bra	CMD_CP_BLK_W24
	movlw	0x0F
	andwf	INDF1
	swapf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W24:
	btfss	C4_REG+3,1		; pixel 25
	bra	CMD_CP_BLK_W25
	movlw	0xF0
	andwf	INDF1
	movf	POSTINC0,W
	iorwf	INDF1
CMD_CP_BLK_W25:
	movf	POSTINC1		; next address in dest buffer
; restore src pointer from D2_REG
	movff	D2_REG,FSR0L
	movff	D2_REG+1,FSR0H
; loop size_y times
	decfsz	C2_REG+1
	bra	CMD_CP_BLK_LOOP
; done
; - new offset par of command pointer is already in CMD_OFFSET
; - value in FSR1 has been destroyed here
	goto	CMD_FINISH_OFFSET

;-------------------------------------------------------------------------------------------
; DISPLAY
; display frame
;-------------------------------------------------------------------------------------------
CMD__DISPLAY:
; request output buffer switch
	bsf	BIT_O_BUF_SW
; wait for buffer switch to complete
CMD__DISPLAY_WAIT:
	btfsc	BIT_O_BUF_SW
	bra	CMD__DISPLAY_WAIT
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DISPLAY_CP
; display frame and copy it into new drawing buffer
;-------------------------------------------------------------------------------------------
CMD__DISPLAY_CP:
; request output buffer switch
	bsf	BIT_O_BUF_SW
; wait for buffer switch to complete
CMD__DISPLAY_CP_WAIT:
	btfsc	BIT_O_BUF_SW
	bra	CMD__DISPLAY_CP_WAIT
; save FSR1 in A2_REG
	movff	FSR1L,A2_REG
	movff	FSR1H,A2_REG+1
; get pointer to buffer owned by interrupt into FSR0
	lfsr	FSR0,O_FRAME0
	btfsc	BIT_O_INT_BUF		; skip if buffer 0 is owned by interrupt
	lfsr	FSR0,O_FRAME1
; get pointer to buffer not owned by interrupt into FSR1
	lfsr	FSR1,O_FRAME0
	btfss	BIT_O_INT_BUF		; skip if buffer 0 is not owned by interrupt
	lfsr	FSR1,O_FRAME1
; copy buffer owned by inteerupt into other one
	movlw	65			; copy 65 times 4 bytes
	movwf	A1_REG
CMD__DISPLAY_CP_LOOP:
	movff	POSTINC0,POSTINC1	; copy 4 bytes
	movff	POSTINC0,POSTINC1
	movff	POSTINC0,POSTINC1
	movff	POSTINC0,POSTINC1
	decfsz	A1_REG			; loop
	bra	CMD__DISPLAY_CP_LOOP
; restore FSR1 from A2_REG
	movff	A2_REG,FSR1L
	movff	A2_REG+1,FSR1H
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DELAY_DO <delay_h> <delay_l>
; wait <delay> milliseconds
;-------------------------------------------------------------------------------------------
CMD__DELAY_DO:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; disable old delay
	bcf	BIT_DELAY		; no delay active
	bcf	BIT_DELAY_WAIT		; do not wait for delay
; get <delay> from variables and write it into delay counter
	movff	POSTINC1,FSR0L		; high part
	movff	INDF0,DELAY+1
	movff	POSTINC1,FSR0L		; low part
	movff	INDF0,DELAY
; exit, if delay is 0
	movf	DELAY,W
	iorwf	DELAY+1,W
	btfsc	STATUS,Z		; goto	CMD_FINISH if delay is 0
	goto	CMD_FINISH
; enable new delay
	bsf	BIT_DELAY		; delay is active
	bsf	BIT_DELAY_WAIT		; wait for delay
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DELAY_DO_C <cons_delay_h> <cons_delay_l>
; wait <delay> milliseconds
;-------------------------------------------------------------------------------------------
CMD__DELAY_DO_C:
; disable old delay
	bcf	BIT_DELAY		; no delay active
	bcf	BIT_DELAY_WAIT		; do not wait for delay
; get <cons_delay> from constants and write it into delay counter
	movff	POSTINC1,DELAY+1	; high part
	movff	POSTINC1,DELAY		; low part
; exit, if delay is 0
	movf	DELAY,W
	iorwf	DELAY+1,W
	btfsc	STATUS,Z		; goto	CMD_FINISH if delay is 0
	goto	CMD_FINISH
; enable new delay
	bsf	BIT_DELAY		; delay is active
	bsf	BIT_DELAY_WAIT		; wait for delay
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DELAY_START <delay_h> <delay_l>
; start background delay of <delay> milliseconds
; - background delay can be checked with DELAY_CHECK
; - background delay can be waited for with DELAY_WAIT
;-------------------------------------------------------------------------------------------
CMD__DELAY_START:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; disable old delay
	bcf	BIT_DELAY		; no delay active
	bcf	BIT_DELAY_WAIT		; do not wait for delay
; get <delay> from variables and write it into delay counter
	movff	POSTINC1,FSR0L		; high part
	movff	INDF0,DELAY+1
	movff	POSTINC1,FSR0L		; low part
	movff	INDF0,DELAY
; exit, if delay is 0
	movf	DELAY,W
	iorwf	DELAY+1,W
	btfsc	STATUS,Z		; goto	CMD_FINISH if delay is 0
	goto	CMD_FINISH
; enable new delay
	bsf	BIT_DELAY		; delay is active
	bcf	BIT_DELAY_WAIT		; do not wait for delay
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DELAY_START_C <cons_delay_h> <cons_delay_l>
; start background delay of <delay> milliseconds
; - background delay can be checked with DELAY_CHECK
; - background delay can be waited for with DELAY_WAIT
;-------------------------------------------------------------------------------------------
CMD__DELAY_START_C:
; disable old delay
	bcf	BIT_DELAY		; no delay active
	bcf	BIT_DELAY_WAIT		; do not wait for delay
; get <cons_delay> from constants and write it into delay counter
	movff	POSTINC1,DELAY+1	; high part
	movff	POSTINC1,DELAY		; low part
; exit, if delay is 0
	movf	DELAY,W
	iorwf	DELAY+1,W
	btfsc	STATUS,Z		; goto	CMD_FINISH if delay is 0
	goto	CMD_FINISH
; enable new delay
	bsf	BIT_DELAY		; delay is active
	bcf	BIT_DELAY_WAIT		; do not wait for delay
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DELAY_CHECK <check result>
; check if background delay elapsed (goto	CMD_FINISHs 0xFF if delay is active, 0x00 if not)
; - background delay can be started with DELAY_START
; - background delay can be waited for with DELAY_WAIT
;-------------------------------------------------------------------------------------------
CMD__DELAY_CHECK:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; check if delay is active and write result into variable
	movff	POSTINC1,FSR0L		; get offset
	clrf	INDF0			; default: goto	CMD_FINISH delay not active
	btfsc	BIT_DELAY		; skip if delay not active
	decf	INDF0			; goto	CMD_FINISH delay active
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; DELAY_WAIT
; wait for background delay to elapse
; - background delay can be started with DELAY_START
; - background delay can be checked with DELAY_CHECK
;-------------------------------------------------------------------------------------------
CMD__DELAY_WAIT:
; wait for delay to elapse
	btfsc	BIT_DELAY		; if delay is active, ...
	bsf	BIT_DELAY_WAIT		; ... wait for delay
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; GET_KEYS <current> <pressed> <released>
; get keys
; - current: current (debounced) key state
; - pressed: keys that were pressed since last GET_KEYS
; - released: keys that were released since last GET_KEYS
;-------------------------------------------------------------------------------------------
CMD__GET_KEYS:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; disable timer 4 interrupt (keys are read there)
	bcf	PIE3,3
; write key states into variables
	movff	POSTINC1,FSR0L		; current
	movff	KEYS,INDF0
	movff	POSTINC1,FSR0L		; pressed
	movff	KEYS_PRESSED,INDF0
	movff	POSTINC1,FSR0L		; released
	movff	KEYS_RELEASED,INDF0
; no keys were pressed or released since last GET_KEYS
	clrf	KEYS_PRESSED
	clrf	KEYS_RELEASED
; reenable timer 4 interrupt
	bsf	PIE3,3
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; GET_SNES0 <current_h> <current_l> <pressed_h> <pressed_l> <released_h> <released_l>
; get keys of 1st SNES pad
; - current: current (debounced) key state
; - pressed: keys that were pressed since last GET_SNES0
; - released: keys that were released since last GET_SNES0
;-------------------------------------------------------------------------------------------
CMD__GET_SNES0:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; disable timer 4 interrupt (keys are read there)
	bcf	PIE3,3
; write key states into variables
	movff	POSTINC1,FSR0L		; current
	movff	SNES0+1,INDF0
	movff	POSTINC1,FSR0L
	movff	SNES0,INDF0
	movff	POSTINC1,FSR0L		; pressed
	movff	SNES0_PRESSED+1,INDF0
	movff	POSTINC1,FSR0L
	movff	SNES0_PRESSED,INDF0
	movff	POSTINC1,FSR0L		; released
	movff	SNES0_RELEASED+1,INDF0
	movff	POSTINC1,FSR0L
	movff	SNES0_RELEASED,INDF0
; no keys were pressed or released since last GET_SNES0
	clrf	SNES0_PRESSED
	clrf	SNES0_PRESSED+1
	clrf	SNES0_RELEASED
	clrf	SNES0_RELEASED+1
; reenable timer 4 interrupt
	bsf	PIE3,3
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; GET_SNES1 <current_h> <current_l> <pressed_h> <pressed_l> <released_h> <released_l>
; get keys of 2nd SNES pad
; - current: current (debounced) key state
; - pressed: keys that were pressed since last GET_SNES1
; - released: keys that were released since last GET_SNES1
;-------------------------------------------------------------------------------------------
CMD__GET_SNES1:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; disable timer 4 interrupt (keys are read there)
	bcf	PIE3,3
; write key states into variables
	movff	POSTINC1,FSR0L		; current
	movff	SNES1+1,INDF0
	movff	POSTINC1,FSR0L
	movff	SNES1,INDF0
	movff	POSTINC1,FSR0L		; pressed
	movff	SNES1_PRESSED+1,INDF0
	movff	POSTINC1,FSR0L
	movff	SNES1_PRESSED,INDF0
	movff	POSTINC1,FSR0L		; released
	movff	SNES1_RELEASED+1,INDF0
	movff	POSTINC1,FSR0L
	movff	SNES1_RELEASED,INDF0
; no keys were pressed or released since last GET_SNES1
	clrf	SNES1_PRESSED
	clrf	SNES1_PRESSED+1
	clrf	SNES1_RELEASED
	clrf	SNES1_RELEASED+1
; reenable timer 4 interrupt
	bsf	PIE3,3
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; GET_TIME <sec> <min> <hour>
; get time: seconds, minutes, hours
;-------------------------------------------------------------------------------------------
CMD__GET_TIME:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; disable timer 4 interrupt (time is advanced there)
	bcf	PIE3,3
; write time into variables
	movff	POSTINC1,FSR0L		; sec
	movff	TIME_SEC,INDF0
	movff	POSTINC1,FSR0L		; min
	movff	TIME_MIN,INDF0
	movff	POSTINC1,FSR0L		; hour
	movff	TIME_HOUR,INDF0
; reenable timer 4 interrupt
	bsf	PIE3,3
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; SET_TIME <sec> <min> <hour>
; set time: seconds (0..59), minutes (0..59), hours (0..23)
;-------------------------------------------------------------------------------------------
CMD__SET_TIME:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; get new time from variables
	movff	POSTINC1,FSR0L		; sec
	movff	INDF0,A4_REG+0
	movff	POSTINC1,FSR0L		; min
	movff	INDF0,A4_REG+1
	movff	POSTINC1,FSR0L		; hour
	movff	INDF0,A4_REG+2
; check time
	movlw	60			; sec < 60?
	cpfslt	A4_REG+0
	goto	CMD_FINISH
	movlw	60			; min < 60?
	cpfslt	A4_REG+1
	goto	CMD_FINISH
	movlw	24			; hour < 24?
	cpfslt	A4_REG+2
	goto	CMD_FINISH
; disable timer 4 interrupt (time is advanced there)
	bcf	PIE3,3
; write time into time varibales
	movff	A4_REG+0,TIME_SEC
	movff	A4_REG+1,TIME_MIN
	movff	A4_REG+2,TIME_HOUR
; reenable timer 4 interrupt
	bsf	PIE3,3
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; CMD_RND <dest>
; fill dest with pseudorandom value
;-------------------------------------------------------------------------------------------
CMD__RND:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; disable timer 4 interrupt (RANDOM is modified there)
	bcf	PIE3,3
; calculate next pseudorandom value
	rlncf	RANDOM,W
	btfsc	RANDOM,4
	xorlw	1
	btfsc	RANDOM,5
	xorlw	1
	btfsc	RANDOM,3
	xorlw	1
	movwf	RANDOM
; write pseudorandom value into dest
	movff	POSTINC1,FSR0L
	movff	RANDOM,INDF0
; reenable timer 4 interrupt
	bsf	PIE3,3
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; VER <major> <minor> <revision> <year> <month> <day>
; get version information: version major.minor.revision from (2000+year)-month-day
;-------------------------------------------------------------------------------------------
CMD__VER:
; initialize FSR0 to access CMD_VARS (starts at page boundary (address 0x??00))
	lfsr	FSR0,CMD_VARS
; write version information into variables
	movff	POSTINC1,FSR0L		; major
	movlw	VER_MAJ
	movwf	INDF0
	movff	POSTINC1,FSR0L		; minor
	movlw	VER_MIN
	movwf	INDF0
	movff	POSTINC1,FSR0L		; revision
	movlw	VER_REV
	movwf	INDF0
	movff	POSTINC1,FSR0L		; year
	movlw	VER_YEAR
	movwf	INDF0
	movff	POSTINC1,FSR0L		; month
	movlw	VER_MONTH
	movwf	INDF0
	movff	POSTINC1,FSR0L		; day
	movlw	VER_DAY
	movwf	INDF0
; done
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; NOP
; no operation
;-------------------------------------------------------------------------------------------
CMD__NOP:
	goto	CMD_FINISH

;-------------------------------------------------------------------------------------------
; finish command execution
;-------------------------------------------------------------------------------------------
CMD_FINISH:
; get offset part of command pointer from FSR1
	movlw	low BUFFER
	subwf	FSR1L,W
	movwf	CMD_OFFSET+0
	movlw	high BUFFER
	subwfb	FSR1H,W
	movwf	CMD_OFFSET+1
; jump here if CMD_OFFSET is already correct (e.g. after a jump instruction)
CMD_FINISH_OFFSET:
; stop command execution if usb requested it
	btfss	BIT_USB_CMD_ST_N	; if stop next bit is set
	bra	CMD_FINISH_USB_CMD_ST_N
	bcf	BIT_USB_CMD_ST_N	; clear stop next bit
	bsf	BIT_USB_CMD_STOP	; set stop bit
	bsf	BIT_USB_CMD_ST_D	; set stop done bit
  ; end delay if delay is active and we are waiting for it
  	btfss	BIT_DELAY
  	bra	CMD_FINISH_USB_CMD_ST_N	; jump if delay not active
  	btfss	BIT_DELAY_WAIT
  	bra	CMD_FINISH_USB_CMD_ST_N	; jump if not waiting for delay
  	bcf	BIT_DELAY		; end delay
  	bcf	BIT_DELAY_WAIT
CMD_FINISH_USB_CMD_ST_N:
; jump here if CMD_OFFSET is already correct (e.g. after a jump instruction)
; and no USB stop after next step shall be done (e.g. command is not yet completed)
CMD_FINISH_OFFSET_NOSTOP:
; done
	return

; jump here if no USB stop after next step shall be done (e.g. command is not yet completed)
CMD_FINISH_NOSTOP:
; get offset part of command pointer from FSR1
	movlw	low BUFFER
	subwf	FSR1L,W
	movwf	CMD_OFFSET+0
	movlw	high BUFFER
	subwfb	FSR1H,W
	movwf	CMD_OFFSET+1
; continue in main finish code
	bra	CMD_FINISH_OFFSET_NOSTOP



;===========================================================================================
; global variables
;===========================================================================================
	global		CMD_SECTOR, CMD_OFFSET, CMD_STACK_PTR
	global		CMD_VARS, CMD_OVERLAY, CMD_STACK



;===========================================================================================
; global procedures
;===========================================================================================
	global		CMD_INIT, CMD_EXEC



	end
